/* SPDX-License-Identifier: Apache-2.0 */
/* Copyright Contributors to the ODPi Egeria project. */
package org.odpi.openmetadata.opentypes;


import org.odpi.openmetadata.frameworks.openmetadata.enums.*;
import org.odpi.openmetadata.frameworks.openmetadata.types.OpenMetadataProperty;
import org.odpi.openmetadata.frameworks.openmetadata.types.OpenMetadataType;
import org.odpi.openmetadata.frameworks.surveyaction.properties.AnnotationStatus;
import org.odpi.openmetadata.repositoryservices.archiveutilities.OMRSArchiveBuilder;
import org.odpi.openmetadata.repositoryservices.archiveutilities.OMRSArchiveHelper;
import org.odpi.openmetadata.repositoryservices.connectors.stores.archivestore.properties.OpenMetadataArchiveType;
import org.odpi.openmetadata.repositoryservices.connectors.stores.metadatacollectionstore.properties.instances.InstanceStatus;
import org.odpi.openmetadata.repositoryservices.connectors.stores.metadatacollectionstore.properties.typedefs.*;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

/**
 * OpenMetadataTypesArchive1_2 provides the archive content for the open metadata types that were defined before Release 1.3.  These types
 * should not be changed or extended in this file.  New types, and patches to existing types should be added to OpenMetadataTypes.
 * <p>
 * Details of the open metadata types are documented on the wiki:
 * <a href="https://egeria.odpi.org/open-metadata-publication/website/open-metadata-types/">The Open Metadata Type System</a>
 * </p>
 * <p>
 * There are 8 areas, each covering a different topic area of metadata.  The module breaks down the process of creating
 * the models into the areas and then the individual models to simplify the maintenance of this class
 * </p>
 */
public class OpenMetadataTypesArchive1_2
{
    /*
     * This is the header information for the archive.
     */
    private static final String                  archiveGUID        = "bce3b0a0-662a-4f87-b8dc-844078a11a6e";
    private static final String                  archiveName        = "Open Metadata Types";
    private static final String                  archiveVersion     = "1.2";
    private static final String                  archiveDescription = "Standard types for open metadata repositories.";
    private static final OpenMetadataArchiveType archiveType        = OpenMetadataArchiveType.CONTENT_PACK;
    private static final String                  originatorName     = "ODPi Egeria";
    private static final String                  originatorLicense  = "Apache-2.0";
    private static final Date                    creationDate       = new Date(1577886131090L);

    /*
     * Specific values for initializing TypeDefs
     */
    private static final long   versionNumber = 1L;
    private static final String versionName   = "1.0";


    private final OMRSArchiveBuilder archiveBuilder;
    private final OMRSArchiveHelper  archiveHelper;

    /**
     * Default constructor sets up the archive builder.  This in turn sets up the header for the archive.
     */
    public OpenMetadataTypesArchive1_2()
    {
        this.archiveBuilder = new OMRSArchiveBuilder(archiveGUID,
                                                     archiveName,
                                                     archiveDescription,
                                                     archiveType,
                                                     archiveVersion,
                                                     originatorName,
                                                     originatorLicense,
                                                     creationDate,
                                                     null);

        this.archiveHelper = new OMRSArchiveHelper(archiveBuilder,
                                                   archiveGUID,
                                                   originatorName,
                                                   creationDate,
                                                   versionNumber,
                                                   versionName);
    }


    /**
     * Chained constructor sets up the archive builder.  This in turn sets up the header for the archive.
     *
     * @param archiveBuilder accumulator for types
     */
    public OpenMetadataTypesArchive1_2(OMRSArchiveBuilder archiveBuilder)
    {
        this.archiveBuilder = archiveBuilder;

        this.archiveHelper = new OMRSArchiveHelper(archiveBuilder,
                                                   archiveGUID,
                                                   originatorName,
                                                   creationDate,
                                                   versionNumber,
                                                   versionName);
    }


    /**
     * Add the types from this archive to the archive builder supplied in the
     * constructor.
     */
    public void getOriginalTypes()
    {
        this.addStandardPrimitiveDefs();
        this.addStandardCollectionDefs();
        this.addArea0Types();
        this.addArea1Types();
        this.addArea2Types();
        this.addArea3Types();
        this.addArea4Types();
        this.addArea5Types();
        this.addArea6Types();
    }


    /*
     * ========================================
     * Attribute types
     */


    /**
     * Add the standard primitive types to the archive builder.
     */
    public void addStandardPrimitiveDefs()
    {
        this.archiveBuilder.addPrimitiveDef(archiveHelper.getPrimitiveDef(PrimitiveDefCategory.OM_PRIMITIVE_TYPE_UNKNOWN));
        this.archiveBuilder.addPrimitiveDef(archiveHelper.getPrimitiveDef(PrimitiveDefCategory.OM_PRIMITIVE_TYPE_BOOLEAN));
        this.archiveBuilder.addPrimitiveDef(archiveHelper.getPrimitiveDef(PrimitiveDefCategory.OM_PRIMITIVE_TYPE_BYTE));
        this.archiveBuilder.addPrimitiveDef(archiveHelper.getPrimitiveDef(PrimitiveDefCategory.OM_PRIMITIVE_TYPE_CHAR));
        this.archiveBuilder.addPrimitiveDef(archiveHelper.getPrimitiveDef(PrimitiveDefCategory.OM_PRIMITIVE_TYPE_SHORT));
        this.archiveBuilder.addPrimitiveDef(archiveHelper.getPrimitiveDef(PrimitiveDefCategory.OM_PRIMITIVE_TYPE_INT));
        this.archiveBuilder.addPrimitiveDef(archiveHelper.getPrimitiveDef(PrimitiveDefCategory.OM_PRIMITIVE_TYPE_LONG));
        this.archiveBuilder.addPrimitiveDef(archiveHelper.getPrimitiveDef(PrimitiveDefCategory.OM_PRIMITIVE_TYPE_FLOAT));
        this.archiveBuilder.addPrimitiveDef(archiveHelper.getPrimitiveDef(PrimitiveDefCategory.OM_PRIMITIVE_TYPE_DOUBLE));
        this.archiveBuilder.addPrimitiveDef(archiveHelper.getPrimitiveDef(PrimitiveDefCategory.OM_PRIMITIVE_TYPE_BIGINTEGER));
        this.archiveBuilder.addPrimitiveDef(archiveHelper.getPrimitiveDef(PrimitiveDefCategory.OM_PRIMITIVE_TYPE_BIGDECIMAL));
        this.archiveBuilder.addPrimitiveDef(archiveHelper.getPrimitiveDef(PrimitiveDefCategory.OM_PRIMITIVE_TYPE_STRING));
        this.archiveBuilder.addPrimitiveDef(archiveHelper.getPrimitiveDef(PrimitiveDefCategory.OM_PRIMITIVE_TYPE_DATE));
    }


    /**
     * Add the standard collection types to the archive builder.
     */
    public void addStandardCollectionDefs()
    {
        this.archiveBuilder.addCollectionDef(getMapStringStringCollectionDef());
        this.archiveBuilder.addCollectionDef(getMapStringBooleanCollectionDef());
        this.archiveBuilder.addCollectionDef(getMapStringIntCollectionDef());
        this.archiveBuilder.addCollectionDef(getMapStringLongCollectionDef());
        this.archiveBuilder.addCollectionDef(getMapStringObjectCollectionDef());
        this.archiveBuilder.addCollectionDef(getArrayStringCollectionDef());
        this.archiveBuilder.addCollectionDef(getArrayIntCollectionDef());

    }


    /**
     * Defines the {@code map<string,string>} type.
     *
     * @return CollectionDef for this type
     */
    private CollectionDef getMapStringStringCollectionDef()
    {
        final String guid            = "005c7c14-ac84-4136-beed-959401b041f8";
        final String description     = "A map from String to String.";
        final String descriptionGUID = null;

        return archiveHelper.getMapCollectionDef(guid,
                                                 description,
                                                 descriptionGUID,
                                                 PrimitiveDefCategory.OM_PRIMITIVE_TYPE_STRING,
                                                 PrimitiveDefCategory.OM_PRIMITIVE_TYPE_STRING);
    }


    /**
     * Defines the {@code map<string,boolean>} type.
     *
     * @return CollectionDef for this type
     */
    private CollectionDef getMapStringBooleanCollectionDef()
    {
        final String guid            = "8fa603dd-c2c5-43fc-8ff4-92141f2414ab";
        final String description     = "A map from String to Boolean.";
        final String descriptionGUID = null;

        return archiveHelper.getMapCollectionDef(guid,
                                                 description,
                                                 descriptionGUID,
                                                 PrimitiveDefCategory.OM_PRIMITIVE_TYPE_STRING,
                                                 PrimitiveDefCategory.OM_PRIMITIVE_TYPE_BOOLEAN);
    }


    /**
     * Defines the {@code map<string,int>} type.
     *
     * @return CollectionDef for this type
     */
    private CollectionDef getMapStringIntCollectionDef()
    {
        final String guid            = "8fa603dd-c2c5-43fc-8ff4-92141f2414ac";
        final String description     = "A map from String to int.";
        final String descriptionGUID = null;

        return archiveHelper.getMapCollectionDef(guid,
                                                 description,
                                                 descriptionGUID,
                                                 PrimitiveDefCategory.OM_PRIMITIVE_TYPE_STRING,
                                                 PrimitiveDefCategory.OM_PRIMITIVE_TYPE_INT);
    }


    /**
     * Defines the {@code map<string,long>} type.
     *
     * @return CollectionDef for this type
     */
    private CollectionDef getMapStringLongCollectionDef()
    {
        final String guid            = "8fa603dd-c2c5-43fc-8ff4-92141f2414ae";
        final String description     = "A map from String to long.";
        final String descriptionGUID = null;

        return archiveHelper.getMapCollectionDef(guid,
                                                 description,
                                                 descriptionGUID,
                                                 PrimitiveDefCategory.OM_PRIMITIVE_TYPE_STRING,
                                                 PrimitiveDefCategory.OM_PRIMITIVE_TYPE_LONG);
    }


    /**
     * Defines the {@code map<string,object>} type.
     *
     * @return CollectionDef for this type
     */
    private CollectionDef getMapStringObjectCollectionDef()
    {
        final String guid            = "8fa603dd-c2c5-43fc-8ff4-92141f2414ad";
        final String description     = "A map from String to Object.";
        final String descriptionGUID = null;

        return archiveHelper.getMapCollectionDef(guid,
                                                 description,
                                                 descriptionGUID,
                                                 PrimitiveDefCategory.OM_PRIMITIVE_TYPE_STRING,
                                                 PrimitiveDefCategory.OM_PRIMITIVE_TYPE_UNKNOWN);
    }


    /**
     * Define the {@code array<string>} type.
     *
     * @return CollectionDef for this object
     */
    private CollectionDef getArrayStringCollectionDef()
    {
        final String guid            = "0428b5d3-f824-459c-b7f5-f8151de59707";
        final String description     = "An array of Strings.";
        final String descriptionGUID = null;

        return archiveHelper.getArrayCollectionDef(guid,
                                                   description,
                                                   descriptionGUID,
                                                   PrimitiveDefCategory.OM_PRIMITIVE_TYPE_STRING);
    }


    /**
     * Define the {@code array<int>} type.
     *
     * @return CollectionDef for this object
     */
    private CollectionDef getArrayIntCollectionDef()
    {
        final String guid            = "0103fe10-98b0-4910-8ee0-21d529f7ff6d";
        final String description     = "An array of integers.";
        final String descriptionGUID = null;

        return archiveHelper.getArrayCollectionDef(guid,
                                                   description,
                                                   descriptionGUID,
                                                   PrimitiveDefCategory.OM_PRIMITIVE_TYPE_INT);
    }


    /*
     * ========================================
     * AREA 0: common types and infrastructure
     */

    /**
     * Add the list of area 0 types
     */
    public void addArea0Types()
    {
        this.add0010BaseModel();
        this.add0015LinkedMediaTypes();
        this.add0017ExternalIdentifiers();
        this.add0019MoreInformation();
        this.add0020PropertyFacets();
        this.add0025Locations();
        this.add0030HostsAndPlatforms();
        this.add0035ComplexHosts();
        this.add0037SoftwareServerPlatforms();
        this.add0040SoftwareServers();
        this.add0042SoftwareServerCapabilities();
        this.add0045ServersAndAssets();
        this.add0050ApplicationsAndProcesses();
        this.add0055DataProcessingEngines();
        this.add0070NetworksAndGateways();
        this.add0090CloudPlatformsAndServices();
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0010 Base Model defines the core entities that have been inherited from the original Apache Atlas model.
     * It defines an initial set of asset types that need to be governed (more assets are defined in Area 2).
     */
    private void add0010BaseModel()
    {
        this.archiveBuilder.addEnumDef(getAssetOwnerTypeEnum());

        this.archiveBuilder.addEntityDef(getReferenceableEntity());
        this.archiveBuilder.addEntityDef(getAssetEntity());
        this.archiveBuilder.addEntityDef(getInfrastructureEntity());
        this.archiveBuilder.addEntityDef(getProcessEntity());
        this.archiveBuilder.addEntityDef(getDataSetEntity());
    }


    /**
     * The Referenceable entity is the superclass of all the governed open metadata entities.  It specifies that
     * these entities have a unique identifier called OpenMetadataProperty.QUALIFIED_NAME.name.
     *
     * @return Referenceable EntityDef
     */
    private EntityDef getReferenceableEntity()
    {
        /*
         * Build the Entity
         */
        final String guid            = OpenMetadataType.REFERENCEABLE.typeGUID;
        final String name            = OpenMetadataType.REFERENCEABLE.typeName;
        final String description     = OpenMetadataType.REFERENCEABLE.description;
        final String descriptionGUID = OpenMetadataType.REFERENCEABLE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.REFERENCEABLE.wikiURL;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                null,
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.QUALIFIED_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.QUALIFIED_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.QUALIFIED_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.ADDITIONAL_PROPERTIES.name;
        final String attribute2Description     = OpenMetadataProperty.ADDITIONAL_PROPERTIES.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.ADDITIONAL_PROPERTIES.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        property.setUnique(true);
        property.setValuesMaxCount(1);
        property.setValuesMinCount(1);
        property.setAttributeCardinality(AttributeCardinality.ONE_ONLY);
        properties.add(property);
        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute2Name,
                                                                    attribute2Description,
                                                                    attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    /**
     * The Asset entity is the root entity for the assets that open metadata and governance is governing.
     *
     * @return Asset EntityDef
     */
    private EntityDef getAssetEntity()
    {
        /*
         * Build the Entity
         */
        final String guid            = OpenMetadataType.ASSET.typeGUID;
        final String name            = OpenMetadataType.ASSET.typeName;
        final String description     = OpenMetadataType.ASSET.description;
        final String descriptionGUID = OpenMetadataType.ASSET.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.ASSET.wikiURL;
        
        final String superTypeName   = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NAME.name;
        final String attribute1Description     = OpenMetadataProperty.NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        /* Moved to Ownership classification */
        final String attribute3Name            = OpenMetadataProperty.OWNER.name;
        final String attribute3Description     = OpenMetadataProperty.OWNER.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.OWNER.descriptionGUID;
        /* Moved to Ownership classification */
        final String attribute4Name            = OpenMetadataProperty.OWNER_TYPE.name;
        final String attribute4Description     = OpenMetadataProperty.OWNER_TYPE.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.OWNER_TYPE.descriptionGUID;
        /* Moved to AssetZoneMembership classification */
        final String attribute5Name            = OpenMetadataProperty.ZONE_MEMBERSHIP.name;
        final String attribute5Description     = OpenMetadataProperty.ZONE_MEMBERSHIP.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.ZONE_MEMBERSHIP.descriptionGUID;
        /* Moved to LatestChange entity  */
        final String attribute6Name            = OpenMetadataProperty.LATEST_CHANGE.name;
        final String attribute6Description     = OpenMetadataProperty.LATEST_CHANGE.description;
        final String attribute6DescriptionGUID = OpenMetadataProperty.LATEST_CHANGE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("AssetOwnerType",
                                                         attribute4Name,
                                                         attribute4Description,
                                                         attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute5Name,
                                                                attribute5Description,
                                                                attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    /**
     * The Infrastructure entity describes an asset that is physical infrastructure or part of the software
     * platform that supports the data and process assets.
     *
     * @return Infrastructure EntityDef
     */
    private EntityDef getInfrastructureEntity()
    {
        /*
         * Build the Entity
         */
        final String guid            = OpenMetadataType.INFRASTRUCTURE.typeGUID;
        final String name            = OpenMetadataType.INFRASTRUCTURE.typeName;
        final String description     = OpenMetadataType.INFRASTRUCTURE.description;
        final String descriptionGUID = OpenMetadataType.INFRASTRUCTURE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.INFRASTRUCTURE.wikiURL;

        final String superTypeName   = OpenMetadataType.ASSET.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    /**
     * The Process entity describes a well-defined sequence of activities performed by people or software components.
     *
     * @return Process EntityDef
     */
    private EntityDef getProcessEntity()
    {
        /*
         * Build the Entity
         */
        final String guid            = OpenMetadataType.PROCESS.typeGUID;
        final String name            = OpenMetadataType.PROCESS.typeName;
        final String description     = OpenMetadataType.PROCESS.description;
        final String descriptionGUID = OpenMetadataType.PROCESS.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.PROCESS.wikiURL;

        final String superTypeName   = OpenMetadataType.ASSET.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.FORMULA.name;
        final String attribute2Description     = OpenMetadataProperty.FORMULA.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.FORMULA.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        ArrayList<InstanceStatus> validInstanceStatusList = new ArrayList<>();

        validInstanceStatusList.add(InstanceStatus.DRAFT);
        validInstanceStatusList.add(InstanceStatus.PROPOSED);
        validInstanceStatusList.add(InstanceStatus.APPROVED);
        validInstanceStatusList.add(InstanceStatus.ACTIVE);
        validInstanceStatusList.add(InstanceStatus.DELETED);

        entityDef.setValidInstanceStatusList(validInstanceStatusList);
        entityDef.setInitialStatus(InstanceStatus.DRAFT);

        return entityDef;
    }


    /**
     * The DataSet entity describes a collection of related data.
     *
     * @return DataSet EntityDef
     */
    private EntityDef getDataSetEntity()
    {
        /*
         * Build the Entity
         */
        final String guid            = OpenMetadataType.DATA_SET.typeGUID;
        final String name            = OpenMetadataType.DATA_SET.typeName;
        final String description     = OpenMetadataType.DATA_SET.description;
        final String descriptionGUID = OpenMetadataType.DATA_SET.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.DATA_SET.wikiURL;

        final String superTypeName   = OpenMetadataType.ASSET.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0015 Linked Media Types describe different media (like images and documents) that enhance the description
     * of an entity.  Media entities can be added to any Referenceable entities.
     */
    private void add0015LinkedMediaTypes()
    {
        this.archiveBuilder.addEnumDef(getMediaTypeEnum());
        this.archiveBuilder.addEnumDef(getMediaUsageEnum());

        this.archiveBuilder.addEntityDef(getExternalReferenceEntity());
        this.archiveBuilder.addEntityDef(getRelatedMediaEntity());

        this.archiveBuilder.addRelationshipDef(getExternalReferenceLinkRelationship());
        this.archiveBuilder.addRelationshipDef(getMediaReferenceRelationship());

    }


    private EnumDef getMediaTypeEnum()
    {
        final String guid            = "9548390c-69f5-4dc6-950d-6fdffb257b56";
        final String name            = "MediaType";
        final String description     = "Defines the type of media.";
        final String descriptionGUID = null;

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID);

        List<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef       elementDef;

        final int    element1Ordinal         = 0;
        final String element1Value           = "Image";
        final String element1Description     = "The media is an image.";
        final String element1DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element1Ordinal,
                                                     element1Value,
                                                     element1Description,
                                                     element1DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element2Ordinal         = 1;
        final String element2Value           = "Audio";
        final String element2Description     = "The media is an audio recording.";
        final String element2DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element2Ordinal,
                                                     element2Value,
                                                     element2Description,
                                                     element2DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element3Ordinal         = 2;
        final String element3Value           = "Document";
        final String element3Description     = "The media is a text document, probably rich text.";
        final String element3DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element3Ordinal,
                                                     element3Value,
                                                     element3Description,
                                                     element3DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element4Ordinal         = 3;
        final String element4Value           = "Video";
        final String element4Description     = "The media is a video recording.";
        final String element4DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element4Ordinal,
                                                     element4Value,
                                                     element4Description,
                                                     element4DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element5Ordinal         = 99;
        final String element5Value           = "Other";
        final String element5Description     = "Another type of media, probably not supported.";
        final String element5DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element5Ordinal,
                                                     element5Value,
                                                     element5Description,
                                                     element5DescriptionGUID);
        elementDefs.add(elementDef);

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EnumDef getMediaUsageEnum()
    {
        final String guid            = "c6861a72-7485-48c9-8040-876f6c342b61";
        final String name            = "MediaUsage";
        final String description     = "Defines how a related media reference should be used.";
        final String descriptionGUID = null;

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID);

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        final int    element1Ordinal         = 0;
        final String element1Value           = "Icon";
        final String element1Description     = "Provides a small image to represent the asset in tree views and graphs.";
        final String element1DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element1Ordinal,
                                                     element1Value,
                                                     element1Description,
                                                     element1DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element2Ordinal         = 1;
        final String element2Value           = "Thumbnail";
        final String element2Description     = "Provides a small image about the asset that can be used in lists.";
        final String element2DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element2Ordinal,
                                                     element2Value,
                                                     element2Description,
                                                     element2DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element3Ordinal         = 2;
        final String element3Value           = "Illustration";
        final String element3Description     = "Illustrates how the asset works or what it contains. It is complementary to the asset's description.";
        final String element3DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element3Ordinal,
                                                     element3Value,
                                                     element3Description,
                                                     element3DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element4Ordinal         = 3;
        final String element4Value           = "UsageGuidance";
        final String element4Description     = "Provides guidance to a person on how to use the asset.";
        final String element4DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element4Ordinal,
                                                     element4Value,
                                                     element4Description,
                                                     element4DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element5Ordinal         = 99;
        final String element5Value           = "Other";
        final String element5Description     = "Another usage.";
        final String element5DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element5Ordinal,
                                                     element5Value,
                                                     element5Description,
                                                     element5DescriptionGUID);
        elementDefs.add(elementDef);

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getExternalReferenceEntity()
    {
        final String guid            = "af536f20-062b-48ef-9c31-1ddd05b04c56";
        final String name            = "ExternalReference";
        final String description     = "A link to more information.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute2Name            = "url";
        final String attribute2Description     = "Location of the external reference.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "version";
        final String attribute3Description     = "Version number of the external reference.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute4Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute5Name            = "organization";
        final String attribute5Description     = "Organization supplying the reference.";
        final String attribute5DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getRelatedMediaEntity()
    {
        final String guid            = "747f8b86-fe7c-4c9b-ba75-979e093cc307";
        final String name            = "RelatedMedia";
        final String description     = "Images, video or sound media.";
        final String descriptionGUID = null;

        final String superTypeName = "ExternalReference";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "mediaUsage";
        final String attribute1Description     = "Type of recommended media usage.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "mediaType";
        final String attribute2Description     = "Type of media.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getArrayIntTypeDefAttribute(attribute1Name,
                                                             attribute1Description,
                                                             attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("MediaType",
                                                         attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getExternalReferenceLinkRelationship()
    {
        final String guid            = "7d818a67-ab45-481c-bc28-f6b1caf12f06";
        final String name            = "ExternalReferenceLink";
        final String description     = "Link to more information.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "relatedItem";
        final String                     end1AttributeDescription     = "Item that is referencing this work.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ExternalReference";
        final String                     end2AttributeName            = "externalReference";
        final String                     end2AttributeDescription     = "Link to more information from an external source.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "referenceId";
        final String attribute1Description     = "Local identifier for the reference.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getMediaReferenceRelationship()
    {
        final String guid            = "1353400f-b0ab-4ab9-ab09-3045dd8a7140";
        final String name            = "MediaReference";
        final String description     = "Link to related media such as images, videos and audio.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "consumingItem";
        final String                     end1AttributeDescription     = "Item that is referencing this work.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "RelatedMedia";
        final String                     end2AttributeName            = "relatedMedia";
        final String                     end2AttributeDescription     = "Link to external media.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "mediaId";
        final String attribute1Description     = "Local identifier for the media.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0017 External Identifiers define identifiers used to identify this entity in other systems.
     */
    private void add0017ExternalIdentifiers()
    {
        this.archiveBuilder.addEnumDef(getKeyPatternEnum());

        this.archiveBuilder.addEntityDef(getExternalIdEntity());

        this.archiveBuilder.addRelationshipDef(getExternalIdScopeRelationship());
        this.archiveBuilder.addRelationshipDef(getExternalIdLinkRelationship());
    }


    private EnumDef getKeyPatternEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(KeyPattern.getOpenTypeGUID(),
                                                        KeyPattern.getOpenTypeName(),
                                                        KeyPattern.getOpenTypeDescription(),
                                                        KeyPattern.getOpenTypeDescriptionGUID(),
                                                        KeyPattern.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (KeyPattern enumValue : KeyPattern.values())
        {
            elementDef = archiveHelper.getEnumElementDef(enumValue.getOrdinal(),
                                                         enumValue.getName(),
                                                         enumValue.getDescription(),
                                                         enumValue.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (enumValue.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getExternalIdEntity()
    {
        final String guid            = OpenMetadataType.EXTERNAL_ID.typeGUID;
        final String name            = OpenMetadataType.EXTERNAL_ID.typeName;
        final String description     = OpenMetadataType.EXTERNAL_ID.description;
        final String descriptionGUID = OpenMetadataType.EXTERNAL_ID.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.EXTERNAL_ID.wikiURL;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.IDENTIFIER.name;
        final String attribute1Description     = OpenMetadataProperty.IDENTIFIER.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.IDENTIFIER.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.KEY_PATTERN.name;
        final String attribute2Description     = OpenMetadataProperty.KEY_PATTERN.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.KEY_PATTERN.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(OpenMetadataProperty.KEY_PATTERN.type,
                                                         attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getExternalIdScopeRelationship()
    {
        final String guid            = OpenMetadataType.EXTERNAL_ID_SCOPE_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.EXTERNAL_ID_SCOPE_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.EXTERNAL_ID_SCOPE_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.EXTERNAL_ID_SCOPE_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.EXTERNAL_ID_SCOPE_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "scopedTo";
        final String                     end1AttributeDescription     = "Identifies where this external identifier is known.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.EXTERNAL_ID.typeName;
        final String                     end2AttributeName            = "managedResources";
        final String                     end2AttributeDescription     = "Link to details of a resource that this component manages.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);


        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getExternalIdLinkRelationship()
    {
        final String guid            = OpenMetadataType.EXTERNAL_ID_LINK_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.EXTERNAL_ID_LINK_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.EXTERNAL_ID_LINK_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.EXTERNAL_ID_LINK_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.EXTERNAL_ID_LINK_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "resources";
        final String                     end1AttributeDescription     = "Resource being identified.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.EXTERNAL_ID.typeName;
        final String                     end2AttributeName            = "alsoKnownAs";
        final String                     end2AttributeDescription     = "Identifier used in an external system.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.USAGE.name;
        final String attribute2Description     = OpenMetadataProperty.USAGE.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.USAGE.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute3Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0019 More information
     */
    private void add0019MoreInformation()
    {
        this.archiveBuilder.addRelationshipDef(getMoreInformationRelationship());
    }


    private RelationshipDef getMoreInformationRelationship()
    {
        final String guid            = OpenMetadataType.MORE_INFORMATION_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.MORE_INFORMATION_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.MORE_INFORMATION_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.MORE_INFORMATION_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.MORE_INFORMATION_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "describes";
        final String                     end1AttributeDescription     = "Describes this core element.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end2AttributeName            = "providesMoreInformation";
        final String                     end2AttributeDescription     = "Provides more information about this referenceable.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0020 Property Facets define blocks of properties that are unique to a particular vendor or service.
     */
    private void add0020PropertyFacets()
    {
        this.archiveBuilder.addEntityDef(getPropertyFacetEntity());

        this.archiveBuilder.addRelationshipDef(getReferenceableFacetRelationship());
    }


    private EntityDef getPropertyFacetEntity()
    {
        final String guid            = "6403a704-aad6-41c2-8e08-b9525c006f85";
        final String name            = "PropertyFacet";
        final String description     = "Additional properties that support a particular vendor or service.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "version";
        final String attribute1Description     = "Version of the property facet schema.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "properties";
        final String attribute3Description     = "Properties for the property facet.";
        final String attribute3DescriptionGUID = null;


        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute3Name,
                                                                    attribute3Description,
                                                                    attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getReferenceableFacetRelationship()
    {
        final String guid            = "58c87647-ada9-4c90-a3c3-a40ace46b1f7";
        final String name            = "ReferenceableFacet";
        final String description     = "Link between a property facet and the resource it relates to.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "relatedEntity";
        final String                     end1AttributeDescription     = "Identifies which element this property facet belongs to.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "PropertyFacet";
        final String                     end2AttributeName            = "facets";
        final String                     end2AttributeDescription     = "Additional properties from different sources.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute1Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0025 Locations define entities that describe physical, logical and cyber locations for Assets.
     */
    private void add0025Locations()
    {
        this.archiveBuilder.addEntityDef(getLocationEntity());

        this.archiveBuilder.addRelationshipDef(getNestedLocationRelationship());
        this.archiveBuilder.addRelationshipDef(getAdjacentLocationRelationship());
        this.archiveBuilder.addRelationshipDef(getAssetLocationRelationship());

        this.archiveBuilder.addClassificationDef(getMobileAssetClassification());
        this.archiveBuilder.addClassificationDef(getFixedLocationClassification());
        this.archiveBuilder.addClassificationDef(getSecureLocationClassification());
        this.archiveBuilder.addClassificationDef(getCyberLocationClassification());
    }


    private EntityDef getLocationEntity()
    {
        final String guid            = "3e09cb2b-5f15-4fd2-b004-fe0146ad8628";
        final String name            = "Location";
        final String description     = "A physical place, digital location or area.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getNestedLocationRelationship()
    {
        final String guid            = "f82a96c2-95a3-4223-88c0-9cbf2882b772";
        final String name            = "NestedLocation";
        final String description     = "Link between two locations to show one is nested inside another.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Location";
        final String                     end1AttributeName            = "groupingLocations";
        final String                     end1AttributeDescription     = "Location that is covering the broader area.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Location";
        final String                     end2AttributeName            = "nestedLocations";
        final String                     end2AttributeDescription     = "Location that is nested in this location.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getAdjacentLocationRelationship()
    {
        final String guid            = "017d0518-fc25-4e5e-985e-491d91e61e17";
        final String name            = "AdjacentLocation";
        final String description     = "Link between two locations that are next to one another.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Location";
        final String                     end1AttributeName            = "peerLocations";
        final String                     end1AttributeDescription     = "Location that is adjacent to this location.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Location";
        final String                     end2AttributeName            = "peerLocations";
        final String                     end2AttributeDescription     = "Location that is adjacent to this location.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getAssetLocationRelationship()
    {
        final String guid            = "bc236b62-d0e6-4c5c-93a1-3a35c3dba7b1";
        final String name            = "AssetLocation";
        final String description     = "Location of an Asset.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Location";
        final String                     end1AttributeName            = "knownLocations";
        final String                     end1AttributeDescription     = "Places where this asset is sited.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.ASSET.typeName;
        final String                     end2AttributeName            = "localAssets";
        final String                     end2AttributeDescription     = "Assets sited at this location.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private ClassificationDef getMobileAssetClassification()
    {
        final String guid            = "b25fb90d-8fa2-4aa9-b884-ff0a6351a697";
        final String name            = "MobileAsset";
        final String description     = "An asset not restricted to a single physical location.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.ASSET.typeName;

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getFixedLocationClassification()
    {
        final String guid            = "bc111963-80c7-444f-9715-946c03142dd2";
        final String name            = "FixedLocation";
        final String description     = "A location linked to a physical place.";
        final String descriptionGUID = null;

        final String linkedToEntity = "Location";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "coordinates";
        final String attribute1Description     = "Geographical coordinates of this location.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "address";
        final String attribute2Description     = "Postal address of this location.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "timezone";
        final String attribute3Description     = "Timezone for the location.";
        final String attribute3DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private ClassificationDef getSecureLocationClassification()
    {
        final String guid            = "e7b563c0-fcdd-4ba7-a046-eecf5c4638b8";
        final String name            = "SecureLocation";
        final String description     = "A location that protects the assets in its care.";
        final String descriptionGUID = null;

        final String linkedToEntity = "Location";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = "level";
        final String attribute2Description     = "Level of security at this location.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private ClassificationDef getCyberLocationClassification()
    {
        final String guid            = "f9ec3633-8ac8-480b-aa6d-5e674b9e1b17";
        final String name            = "CyberLocation";
        final String description     = "A digital location.";
        final String descriptionGUID = null;

        final String linkedToEntity = "Location";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "address";
        final String attribute1Description     = "Network address of the location.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0030 Hosts and Platforms describe the Software and Hardware platforms hosting Assets.
     */
    private void add0030HostsAndPlatforms()
    {
        this.archiveBuilder.addEnumDef(getEndiannessEnum());

        this.archiveBuilder.addEntityDef(getITInfrastructureEntity());
        this.archiveBuilder.addEntityDef(getHostEntity());
        this.archiveBuilder.addEntityDef(getOperatingPlatformEntity());

        this.archiveBuilder.addRelationshipDef(getHostLocationRelationship());
        this.archiveBuilder.addRelationshipDef(getHostOperatingPlatformRelationship());
    }


    private EnumDef getEndiannessEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(ByteOrdering.getOpenTypeGUID(),
                                                        ByteOrdering.getOpenTypeName(),
                                                        ByteOrdering.getOpenTypeDescription(),
                                                        ByteOrdering.getOpenTypeDescriptionGUID(),
                                                        ByteOrdering.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (ByteOrdering byteOrdering : ByteOrdering.values())
        {
            elementDef = archiveHelper.getEnumElementDef(byteOrdering.getOrdinal(),
                                                         byteOrdering.getName(),
                                                         byteOrdering.getDescription(),
                                                         byteOrdering.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (byteOrdering.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getITInfrastructureEntity()
    {
        final String guid            = OpenMetadataType.IT_INFRASTRUCTURE.typeGUID;
        final String name            = OpenMetadataType.IT_INFRASTRUCTURE.typeName;
        final String description     = OpenMetadataType.IT_INFRASTRUCTURE.description;
        final String descriptionGUID = OpenMetadataType.IT_INFRASTRUCTURE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.IT_INFRASTRUCTURE.wikiURL;

        final String superTypeName = OpenMetadataType.INFRASTRUCTURE.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    private EntityDef getHostEntity()
    {
        final String guid            = "1abd16db-5b8a-4fd9-aee5-205db3febe99";
        final String name            = "Host";
        final String description     = "Named IT infrastructure system that supports multiple software platforms and servers.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.IT_INFRASTRUCTURE.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getOperatingPlatformEntity()
    {
        EntityDef entityDef = archiveHelper.getDefaultEntityDef(OpenMetadataType.OPERATING_PLATFORM.typeGUID,
                                                                OpenMetadataType.OPERATING_PLATFORM.typeName,
                                                                this.archiveBuilder.getEntityDef(OpenMetadataType.REFERENCEABLE.typeName),
                                                                OpenMetadataType.OPERATING_PLATFORM.description,
                                                                OpenMetadataType.OPERATING_PLATFORM.descriptionGUID,
                                                                OpenMetadataType.OPERATING_PLATFORM.wikiURL);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NAME.name;
        final String attribute1Description     = OpenMetadataProperty.NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "operatingSystem";
        final String attribute3Description     = "Name of the operating system running on this operating platform.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "endianness";
        final String attribute4Description     = "Definition of byte ordering.";
        final String attribute4DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getHostLocationRelationship()
    {
        final String guid            = "f3066075-9611-4886-9244-32cc6eb07ea9";
        final String name            = "HostLocation";
        final String description     = "Defines the location of a host.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Host";
        final String                     end1AttributeName            = "localHosts";
        final String                     end1AttributeDescription     = "Host sited at this location.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Location";
        final String                     end2AttributeName            = "locations";
        final String                     end2AttributeDescription     = "Locations for this host.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getHostOperatingPlatformRelationship()
    {
        final String guid            = "b9179df5-6e23-4581-a8b0-2919e6322b12";
        final String name            = "HostOperatingPlatform";
        final String description     = "Identifies the operating platform for a host.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Host";
        final String                     end1AttributeName            = "hosts";
        final String                     end1AttributeDescription     = "Host supporting this operating platform.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "OperatingPlatform";
        final String                     end2AttributeName            = "operatingPlatform";
        final String                     end2AttributeDescription     = "Type of platform supported by this host.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);


        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0035 Complex Hosts describe virtualization and clustering options for Hosts.
     */
    private void add0035ComplexHosts()
    {
        this.archiveBuilder.addEntityDef(getHostClusterEntity());
        this.archiveBuilder.addEntityDef(getVirtualContainerEntity());

        this.archiveBuilder.addRelationshipDef(getHostClusterMemberRelationship());
        this.archiveBuilder.addRelationshipDef(getDeployedVirtualContainerRelationship());
    }


    private EntityDef getHostClusterEntity()
    {
        final String guid            = "9794f42f-4c9f-4fe6-be84-261f0a7de890";
        final String name            = "HostCluster";
        final String description     = "A group of hosts operating together to provide a scalable platform.";
        final String descriptionGUID = null;

        final String superTypeName = "Host";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getVirtualContainerEntity()
    {
        final String guid            = "e2393236-100f-4ac0-a5e6-ce4e96c521e7";
        final String name            = "VirtualContainer";
        final String description     = "Container-based virtual host.";
        final String descriptionGUID = null;

        final String superTypeName = "Host";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getHostClusterMemberRelationship()
    {
        final String guid            = "1a1c3933-a583-4b0c-9e42-c3691296a8e0";
        final String name            = "HostClusterMember";
        final String description     = "Identifies a host as a member of a host cluster.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "HostCluster";
        final String                     end1AttributeName            = "hostCluster";
        final String                     end1AttributeDescription     = "Cluster managing this host.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Host";
        final String                     end2AttributeName            = "managedHosts";
        final String                     end2AttributeDescription     = "Member of the host cluster.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);


        return relationshipDef;
    }


    private RelationshipDef getDeployedVirtualContainerRelationship()
    {
        final String guid            = "4b981d89-e356-4d9b-8f17-b3a8d5a86676";
        final String name            = "DeployedVirtualContainer";
        final String description     = "Identifies the real host where a virtual container is deployed to.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Host";
        final String                     end1AttributeName            = "hosts";
        final String                     end1AttributeDescription     = "Deployed host for this container.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "VirtualContainer";
        final String                     end2AttributeName            = "hostedContainers";
        final String                     end2AttributeDescription     = "Virtual containers deployed on this host.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);


        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0037 Software Server Platforms describe the structure of a software server platform.
     */
    private void add0037SoftwareServerPlatforms()
    {
        this.archiveBuilder.addEnumDef(getOperationalStatusEnum());

        this.archiveBuilder.addEntityDef(getSoftwareServerPlatformEntity());

        this.archiveBuilder.addRelationshipDef(getSoftwareServerPlatformDeploymentRelationship());
    }


    private EnumDef getOperationalStatusEnum()
    {
        final String guid            = OpenMetadataType.OPERATIONAL_STATUS_ENUM_TYPE_GUID;
        final String name            = OpenMetadataType.OPERATIONAL_STATUS_ENUM_TYPE_NAME;
        final String description     = "Defines whether a component is operational.";
        final String descriptionGUID = null;

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID);

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        final int    element1Ordinal         = 0;
        final String element1Value           = "Disabled";
        final String element1Description     = "The component is not operational.";
        final String element1DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element1Ordinal,
                                                     element1Value,
                                                     element1Description,
                                                     element1DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element2Ordinal         = 1;
        final String element2Value           = "Enabled";
        final String element2Description     = "The component is operational.";
        final String element2DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element2Ordinal,
                                                     element2Value,
                                                     element2Description,
                                                     element2DescriptionGUID);
        elementDefs.add(elementDef);

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }



    private EntityDef getSoftwareServerPlatformEntity()
    {
        final String guid            = OpenMetadataType.SOFTWARE_SERVER_PLATFORM.typeGUID;
        final String name            = OpenMetadataType.SOFTWARE_SERVER_PLATFORM.typeName;
        final String description     = OpenMetadataType.SOFTWARE_SERVER_PLATFORM.description;
        final String descriptionGUID = OpenMetadataType.SOFTWARE_SERVER_PLATFORM.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.SOFTWARE_SERVER_PLATFORM.wikiURL;

        final String superTypeName = OpenMetadataType.IT_INFRASTRUCTURE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "type";
        final String attribute1Description     = "Type of software server platform.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "version";
        final String attribute2Description     = "Version number of the software server platform.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute3Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;
        final String attribute4Name            = "userId";
        final String attribute4Description     = "Server platform's authentication name.";
        final String attribute4DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }



    private RelationshipDef getSoftwareServerPlatformDeploymentRelationship()
    {
        final String guid            = "b909eb3b-5205-4180-9f63-122a65b30738";
        final String name            = "SoftwareServerPlatformDeployment";
        final String description     = "Defines the host that a software server platform is deployed to.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Host";
        final String                     end1AttributeName            = "host";
        final String                     end1AttributeDescription     = "Supporting host.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SoftwareServerPlatform";
        final String                     end2AttributeName            = "deployedServerPlatforms";
        final String                     end2AttributeDescription     = "Software server platforms deployed on this host.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "deploymentTime";
        final String attribute1Description     = "Time that the software server platform was deployed to the host.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "deployer";
        final String attribute2Description     = "Person, organization or engine that deployed the software server platform.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "platformStatus";
        final String attribute3Description     = "The operational status of the software server platform on this host.";
        final String attribute3DescriptionGUID = null;

        property = archiveHelper.getDateTypeDefAttribute(attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(OpenMetadataType.OPERATIONAL_STATUS_ENUM_TYPE_NAME,
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0040 Software Servers describe the structure of a software server and its capabilities.
     */
    private void add0040SoftwareServers()
    {
        this.archiveBuilder.addEntityDef(getSoftwareServerEntity());
        this.archiveBuilder.addEntityDef(getEndpointEntity());

        this.archiveBuilder.addRelationshipDef(getSoftwareServerDeploymentRelationship());
        this.archiveBuilder.addRelationshipDef(getServerEndpointRelationship());
    }


    private EntityDef getSoftwareServerEntity()
    {
        final String guid            = OpenMetadataType.SOFTWARE_SERVER.typeGUID;
        final String name            = OpenMetadataType.SOFTWARE_SERVER.typeName;
        final String description     = OpenMetadataType.SOFTWARE_SERVER.description;
        final String descriptionGUID = OpenMetadataType.SOFTWARE_SERVER.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.SOFTWARE_SERVER.wikiURL;

        final String superTypeName = OpenMetadataType.IT_INFRASTRUCTURE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "type";
        final String attribute1Description     = "Type of software server.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "version";
        final String attribute2Description     = "Version number of the software server.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute3Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;
        final String attribute4Name            = "userId";
        final String attribute4Description     = "Server's authentication name.";
        final String attribute4DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getEndpointEntity()
    {
        final String guid            = "dbc20663-d705-4ff0-8424-80c262c6b8e7";
        final String name            = "Endpoint";
        final String description     = "Description of the network address and related information needed to call a software service.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NAME.name;
        final String attribute1Description     = OpenMetadataProperty.NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "networkAddress";
        final String attribute3Description     = "Name used to connect to the endpoint.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "protocol";
        final String attribute4Description     = "Name of the protocol used to connect to the endpoint.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "encryptionMethod";
        final String attribute5Description     = "Type of encryption used at the endpoint (if any).";
        final String attribute5DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getSoftwareServerDeploymentRelationship()
    {
        final String guid            = "d909eb3b-5205-4180-9f63-122a65b30738";
        final String name            = "SoftwareServerDeployment";
        final String description     = "Defines the platform that a software server is deployed to.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "SoftwareServerPlatform";
        final String                     end1AttributeName            = "hostingPlatform";
        final String                     end1AttributeDescription     = "Supporting platform for the software server.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SoftwareServer";
        final String                     end2AttributeName            = "deployedSoftwareServers";
        final String                     end2AttributeDescription     = "Software servers deployed on this platform.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "deploymentTime";
        final String attribute1Description     = "Time that the software server was deployed to the platform.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "deployer";
        final String attribute2Description     = "Person, organization or engine that deployed the software server.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "serverStatus";
        final String attribute3Description     = "The operational status of the software server on this platform.";
        final String attribute3DescriptionGUID = null;

        property = archiveHelper.getDateTypeDefAttribute(attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(OpenMetadataType.OPERATIONAL_STATUS_ENUM_TYPE_NAME,
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getServerEndpointRelationship()
    {
        final String guid            = "2b8bfab4-8023-4611-9833-82a0dc95f187";
        final String name            = "ServerEndpoint";
        final String description     = "Defines an endpoint associated with a server.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "SoftwareServer";
        final String                     end1AttributeName            = "servers";
        final String                     end1AttributeDescription     = "Server supporting this endpoint.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Endpoint";
        final String                     end2AttributeName            = "endpoints";
        final String                     end2AttributeDescription     = "Endpoints supported by this server.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);


        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0042 Software Server Capabilities describe the deployed capabilities on a software server.
     */
    private void add0042SoftwareServerCapabilities()
    {

        this.archiveBuilder.addEntityDef(getSoftwareServerCapabilityEntity());

        this.archiveBuilder.addRelationshipDef(getSoftwareServerSupportedCapabilityRelationship());
    }


    private EntityDef getSoftwareServerCapabilityEntity()
    {
        final String guid            = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.typeGUID;
        final String name            = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.typeName;
        final String description     = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.description;
        final String descriptionGUID = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.wikiURL;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute3Name            = "type";
        final String attribute3Description     = "Type of the software server capability.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "version";
        final String attribute4Description     = "Version number of the software server capability.";
        final String attribute4DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getSoftwareServerSupportedCapabilityRelationship()
    {
        final String guid            = "8b7d7da5-0668-4174-a43b-8f8c6c068dd0";
        final String name            = "SoftwareServerSupportedCapability";
        final String description     = "Identifies a software capability that is deployed to a software server.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "SoftwareServer";
        final String                     end1AttributeName            = "servers";
        final String                     end1AttributeDescription     = "Servers hosting this capability.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.typeName;
        final String                     end2AttributeName            = "capabilities";
        final String                     end2AttributeDescription     = "Capabilities deployed on this software server.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute3Name            = "serverCapabilityStatus";
        final String attribute3Description     = "The operational status of the software server capability on this software server.";
        final String attribute3DescriptionGUID = null;

        property = archiveHelper.getEnumTypeDefAttribute(OpenMetadataType.OPERATIONAL_STATUS_ENUM_TYPE_NAME,
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0045 Servers and Assets defines the relationships between SoftwareServers and Assets.
     */
    private void add0045ServersAndAssets()
    {
        this.archiveBuilder.addEnumDef(getServerAssetUseTypeEnum());

        this.archiveBuilder.addRelationshipDef(getServerAssetUseRelationship());
    }


    private EnumDef getServerAssetUseTypeEnum()
    {
        final String guid            = "09439481-9489-467c-9ae5-178a6e0b6b5a";
        final String name            = "ServerAssetUseType";
        final String description     = "Defines how a software server capability may use an asset.";
        final String descriptionGUID = null;

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID);

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        final int    element1Ordinal         = 0;
        final String element1Value           = "Owns";
        final String element1Description     = "The software server capability is accountable for the maintenance and protection of the asset.";
        final String element1DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element1Ordinal,
                                                     element1Value,
                                                     element1Description,
                                                     element1DescriptionGUID);
        elementDefs.add(elementDef);
        enumDef.setDefaultValue(elementDef);

        final int    element2Ordinal         = 1;
        final String element2Value           = "Governs";
        final String element2Description     = "The software server capability provides management or oversight of the asset.";
        final String element2DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element2Ordinal,
                                                     element2Value,
                                                     element2Description,
                                                     element2DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element3Ordinal         = 2;
        final String element3Value           = "Maintains";
        final String element3Description     = "The software server capability keeps the asset up-to-date.";
        final String element3DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element3Ordinal,
                                                     element3Value,
                                                     element3Description,
                                                     element3DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element4Ordinal         = 3;
        final String element4Value           = "Uses";
        final String element4Description     = "The software server capability consumes the content of the asset.";
        final String element4DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element4Ordinal,
                                                     element4Value,
                                                     element4Description,
                                                     element4DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element5Ordinal         = 99;
        final String element5Value           = "Other";
        final String element5Description     = "Another usage.";
        final String element5DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element5Ordinal,
                                                     element5Value,
                                                     element5Description,
                                                     element5DescriptionGUID);
        elementDefs.add(elementDef);

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private RelationshipDef getServerAssetUseRelationship()
    {
        final String guid            = "92b75926-8e9a-46c7-9d98-89009f622397";
        final String name            = "AssetServerUse";
        final String description     = "Defines that a server capability is using an asset.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.typeName;
        final String                     end1AttributeName            = "consumedBy";
        final String                     end1AttributeDescription     = "Capability consuming this asset.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.ASSET.typeName;
        final String                     end2AttributeName            = "consumedAsset";
        final String                     end2AttributeDescription     = "Asset that this software server capability is dependent on.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.USE_TYPE.name;
        final String attribute1Description     = OpenMetadataProperty.USE_TYPE.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.USE_TYPE.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getEnumTypeDefAttribute(OpenMetadataType.SERVER_ASSET_USE_TYPE_TYPE_NAME,
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0050ApplicationsAndProcesses()
    {
        this.archiveBuilder.addEntityDef(getApplicationEntity());

        this.archiveBuilder.addRelationshipDef(getRuntimeForProcessRelationship());

        this.archiveBuilder.addClassificationDef(getApplicationServerClassification());
        this.archiveBuilder.addClassificationDef(getWebserverClassification());
    }


    private EntityDef getApplicationEntity()
    {
        final String guid            = OpenMetadataType.APPLICATION.typeGUID;
        final String name            = OpenMetadataType.APPLICATION.typeName;
        final String description     = OpenMetadataType.APPLICATION.description;
        final String descriptionGUID = OpenMetadataType.APPLICATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.APPLICATION.wikiURL;

        final String superTypeName = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    private RelationshipDef getRuntimeForProcessRelationship()
    {
        final String guid            = "f6b5cf4f-7b88-47df-aeb0-d80d28ba1ec1";
        final String name            = "RuntimeForProcess";
        final String description     = "Identifies the deployed application that supports a specific automated process.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Application";
        final String                     end1AttributeName            = "implementingApplication";
        final String                     end1AttributeDescription     = "Application that contains the process implementation.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.PROCESS.typeName;
        final String                     end2AttributeName            = "implementedProcesses";
        final String                     end2AttributeDescription     = "Processes that are implemented by this application.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private ClassificationDef getApplicationServerClassification()
    {
        final String guid            = "19196efb-2706-47bf-8e51-e8ba5b36d033";
        final String name            = "ApplicationServer";
        final String description     = "A server that hosts applications.";
        final String descriptionGUID = null;

        final String linkedToEntity = "SoftwareServer";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getWebserverClassification()
    {
        final String guid            = "d13e1cc5-bb7e-41ec-8233-9647fbf92a19";
        final String name            = "Webserver";
        final String description     = "A server that supports HTTP-based application such as websites and REST services.";
        final String descriptionGUID = null;

        final String linkedToEntity = "SoftwareServer";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0055DataProcessingEngines()
    {
        this.archiveBuilder.addEntityDef(getEngineEntity());

        this.archiveBuilder.addClassificationDef(getWorkflowEngineClassification());
        this.archiveBuilder.addClassificationDef(getReportingEngineClassification());
        this.archiveBuilder.addClassificationDef(getAnalyticsEngineClassification());
        this.archiveBuilder.addClassificationDef(getDataMovementEngineClassification());
        this.archiveBuilder.addClassificationDef(getDataVirtualizationEngineClassification());
    }


    private EntityDef getEngineEntity()
    {
        final String guid            = OpenMetadataType.ENGINE.typeGUID;
        final String name            = OpenMetadataType.ENGINE.typeName;
        final String description     = OpenMetadataType.ENGINE.description;
        final String descriptionGUID = OpenMetadataType.ENGINE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.ENGINE.wikiURL;

        final String superTypeName = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    private ClassificationDef getWorkflowEngineClassification()
    {
        final String guid            = OpenMetadataType.WORKFLOW_ENGINE_CLASSIFICATION.typeGUID;
        final String name            = OpenMetadataType.WORKFLOW_ENGINE_CLASSIFICATION.typeName;
        final String description     = OpenMetadataType.WORKFLOW_ENGINE_CLASSIFICATION.description;
        final String descriptionGUID = OpenMetadataType.WORKFLOW_ENGINE_CLASSIFICATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.WORKFLOW_ENGINE_CLASSIFICATION.wikiURL;

        final String linkedToEntity = OpenMetadataType.ENGINE.typeName;

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  descriptionWiki,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getReportingEngineClassification()
    {
        final String guid            = OpenMetadataType.REPORTING_ENGINE_CLASSIFICATION.typeGUID;
        final String name            = OpenMetadataType.REPORTING_ENGINE_CLASSIFICATION.typeName;
        final String description     = OpenMetadataType.REPORTING_ENGINE_CLASSIFICATION.description;
        final String descriptionGUID = OpenMetadataType.REPORTING_ENGINE_CLASSIFICATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.REPORTING_ENGINE_CLASSIFICATION.wikiURL;

        final String linkedToEntity = OpenMetadataType.ENGINE.typeName;

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  descriptionWiki,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getAnalyticsEngineClassification()
    {
        final String guid            = OpenMetadataType.ANALYTICS_ENGINE.typeGUID;
        final String name            = OpenMetadataType.ANALYTICS_ENGINE.typeName;
        final String description     = OpenMetadataType.ANALYTICS_ENGINE.description;
        final String descriptionGUID = OpenMetadataType.ANALYTICS_ENGINE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.ANALYTICS_ENGINE.wikiURL;

        final String linkedToEntity = OpenMetadataType.ENGINE.typeName;

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  descriptionWiki,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getDataMovementEngineClassification()
    {
        final String guid            = OpenMetadataType.DATA_MOVEMENT_ENGINE.typeGUID;
        final String name            = OpenMetadataType.DATA_MOVEMENT_ENGINE.typeName;
        final String description     = OpenMetadataType.DATA_MOVEMENT_ENGINE.description;
        final String descriptionGUID = OpenMetadataType.DATA_MOVEMENT_ENGINE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.DATA_MOVEMENT_ENGINE.wikiURL;

        final String linkedToEntity = OpenMetadataType.ENGINE.typeName;

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  descriptionWiki,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getDataVirtualizationEngineClassification()
    {
        final String guid            = OpenMetadataType.DATA_VIRTUALIZATION_ENGINE.typeGUID;
        final String name            = OpenMetadataType.DATA_VIRTUALIZATION_ENGINE.typeName;
        final String description     = OpenMetadataType.DATA_VIRTUALIZATION_ENGINE.description;
        final String descriptionGUID = OpenMetadataType.DATA_VIRTUALIZATION_ENGINE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.DATA_VIRTUALIZATION_ENGINE.wikiURL;

        final String linkedToEntity = OpenMetadataType.ENGINE.typeName;

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  descriptionWiki,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0070 Networks and Gateways provide a very simple network model.
     */
    private void add0070NetworksAndGateways()
    {
        this.archiveBuilder.addEntityDef(getNetworkEntity());
        this.archiveBuilder.addEntityDef(getNetworkGatewayEntity());

        this.archiveBuilder.addRelationshipDef(getHostNetworkRelationship());
        this.archiveBuilder.addRelationshipDef(getNetworkGatewayLinkRelationship());
    }


    private EntityDef getNetworkEntity()
    {
        final String guid            = "e0430f59-f021-411a-9d81-883e1ff3f6f6";
        final String name            = "Network";
        final String description     = "Inter-connectivity for systems.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.IT_INFRASTRUCTURE.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getNetworkGatewayEntity()
    {
        final String guid            = "9bbae94d-e109-4c96-b072-4f97123f04fd";
        final String name            = "NetworkGateway";
        final String description     = "A connection point enabling network traffic to pass between two networks.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getHostNetworkRelationship()
    {
        final String guid            = "f2bd7401-c064-41ac-862c-e5bcdc98fa1e";
        final String name            = "HostNetwork";
        final String description     = "One of the hosts connected to a network.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Host";
        final String                     end1AttributeName            = "connectedHosts";
        final String                     end1AttributeDescription     = "Hosts connected to this network.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Network";
        final String                     end2AttributeName            = "networkConnections";
        final String                     end2AttributeDescription     = "Connections to different networks.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getNetworkGatewayLinkRelationship()
    {
        final String guid            = "5bece460-1fa6-41fb-a29f-fdaf65ec8ce3";
        final String name            = "NetworkGatewayLink";
        final String description     = "Link from a network to one of its network gateways.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "NetworkGateway";
        final String                     end1AttributeName            = "gateways";
        final String                     end1AttributeDescription     = "Gateways to other networks.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Network";
        final String                     end2AttributeName            = "networkConnections";
        final String                     end2AttributeDescription     = "Connections to different networks.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0090 Cloud Platforms and Services provides classifications for infrastructure to allow cloud platforms
     * and services to be identified.
     */
    private void add0090CloudPlatformsAndServices()
    {
        this.archiveBuilder.addClassificationDef(getCloudProviderClassification());
        this.archiveBuilder.addClassificationDef(getCloudPlatformClassification());
        this.archiveBuilder.addClassificationDef(getCloudTenantClassification());
        this.archiveBuilder.addClassificationDef(getCloudServiceClassification());
    }


    private ClassificationDef getCloudProviderClassification()
    {
        final String guid            = "a2bfdd08-d0a8-49db-bc97-7f2406281046";
        final String name            = "CloudProvider";
        final String description     = "A host supporting cloud services.";
        final String descriptionGUID = null;

        final String linkedToEntity = "Host";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "providerName";
        final String attribute1Description     = "Name of the cloud provider.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private ClassificationDef getCloudPlatformClassification()
    {
        final String guid            = "1b8f8511-e606-4f65-86d3-84891706ad12";
        final String name            = "CloudPlatform";
        final String description     = "A software server platform supporting cloud services.";
        final String descriptionGUID = null;

        final String linkedToEntity = "SoftwareServerPlatform";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "type";
        final String attribute1Description     = "Type of cloud platform.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private ClassificationDef getCloudTenantClassification()
    {
        final String guid            = "1b8f8522-e606-4f65-86d3-84891706ad12";
        final String name            = "CloudTenant";
        final String description     = "A software server supporting cloud services.";
        final String descriptionGUID = null;

        final String linkedToEntity = "SoftwareServer";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "tenantName";
        final String attribute1Description     = "Identifier of the tenant.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "type";
        final String attribute2Description     = "Description of the type of tenant.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private ClassificationDef getCloudServiceClassification()
    {
        final String guid            = "337e7b1a-ad4b-4818-aa3e-0ff3307b2fbe";
        final String name            = "CloudService";
        final String description     = "A service running on a cloud platform.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.typeName;

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "offeringName";
        final String attribute1Description     = "Commercial name of the service.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "type";
        final String attribute2Description     = "Description of the type of the service.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }

    /*
     * ========================================
     * AREA 1: collaboration
     */

    private void addArea1Types()
    {
        this.add0110Actors();
        this.add0112People();
        this.add0115Teams();
        this.add0117EngineProfiles();
        this.add0120Collections();
        this.add0130Projects();
        this.add0135Meetings();
        this.add0137Actions();
        this.add0140Communities();
        this.add0150Feedback();
        this.add0155CrowdSourcing();
        this.add0160Notes();
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0110 Actors describe the people and their relationships who are using the Assets.
     */
    private void add0110Actors()
    {
        this.archiveBuilder.addEnumDef(getContactMethodTypeEnum());

        this.archiveBuilder.addEntityDef(getActorProfileEntity());
        this.archiveBuilder.addEntityDef(getUserIdentityEntity());
        this.archiveBuilder.addEntityDef(getContactDetailsEntity());

        this.archiveBuilder.addRelationshipDef(getContactThroughRelationship());
        this.archiveBuilder.addRelationshipDef(getProfileIdentityRelationship());
    }


    private EnumDef getContactMethodTypeEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(ContactMethodType.getOpenTypeGUID(),
                                                        ContactMethodType.getOpenTypeName(),
                                                        ContactMethodType.getOpenTypeDescription(),
                                                        ContactMethodType.getOpenTypeDescriptionGUID(),
                                                        ContactMethodType.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (ContactMethodType enumValues : ContactMethodType.values())
        {
            elementDef = archiveHelper.getEnumElementDef(enumValues.getOrdinal(),
                                                         enumValues.getName(),
                                                         enumValues.getDescription(),
                                                         enumValues.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (enumValues.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getActorProfileEntity()
    {
        final String guid            = "5a2f38dc-d69d-4a6f-ad26-ac86f118fa35";
        final String name            = "ActorProfile";
        final String description     = "Description of a person, team or automated process that is working with data.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NAME.name;
        final String attribute1Description     = OpenMetadataProperty.NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getUserIdentityEntity()
    {
        final String guid            = "fbe95779-1f3c-4ac6-aa9d-24963ff16282";
        final String name            = "UserIdentity";
        final String description     = "Name of the security account for a person or automated process.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getContactDetailsEntity()
    {
        EntityDef entityDef = archiveHelper.getDefaultEntityDef(OpenMetadataType.CONTACT_DETAILS.typeGUID,
                                                                OpenMetadataType.CONTACT_DETAILS.typeName,
                                                                this.archiveBuilder.getEntityDef(OpenMetadataType.REFERENCEABLE.typeName),
                                                                OpenMetadataType.CONTACT_DETAILS.description,
                                                                OpenMetadataType.CONTACT_DETAILS.descriptionGUID,
                                                                OpenMetadataType.CONTACT_DETAILS.wikiURL);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.CONTACT_METHOD_TYPE.name;
        final String attribute1Description     = OpenMetadataProperty.CONTACT_METHOD_TYPE.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.CONTACT_METHOD_TYPE.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.CONTACT_METHOD_VALUE.name;
        final String attribute2Description     = OpenMetadataProperty.CONTACT_METHOD_VALUE.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.CONTACT_METHOD_VALUE.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.CONTACT_METHOD_SERVICE.name;
        final String attribute3Description     = OpenMetadataProperty.CONTACT_METHOD_SERVICE.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.CONTACT_METHOD_SERVICE.descriptionGUID;

        property = archiveHelper.getEnumTypeDefAttribute(ContactMethodType.getOpenTypeName(),
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getContactThroughRelationship()
    {
        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(OpenMetadataType.CONTACT_THROUGH_RELATIONSHIP.typeGUID,
                                                                                OpenMetadataType.CONTACT_THROUGH_RELATIONSHIP.typeName,
                                                                                null,
                                                                                OpenMetadataType.CONTACT_THROUGH_RELATIONSHIP.description,
                                                                                OpenMetadataType.CONTACT_THROUGH_RELATIONSHIP.descriptionGUID,
                                                                                OpenMetadataType.CONTACT_THROUGH_RELATIONSHIP.wikiURL,
                                                                                ClassificationPropagationRule.NONE);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.ACTOR_PROFILE_TYPE_NAME;
        final String                     end1AttributeName            = "contactDetails";
        final String                     end1AttributeDescription     = "Contact details owner.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.CONTACT_DETAILS.typeName;
        final String                     end2AttributeName            = "contacts";
        final String                     end2AttributeDescription     = "Contact information.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getProfileIdentityRelationship()
    {
        final String guid            = "01664609-e777-4079-b543-6baffe910ff1";
        final String name            = "ProfileIdentity";
        final String description     = "Correlates a user identity with an actor profile.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "ActorProfile";
        final String                     end1AttributeName            = "profile";
        final String                     end1AttributeDescription     = "Description of the person, organization or engine that uses this user identity.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "UserIdentity";
        final String                     end2AttributeName            = "userIdentities";
        final String                     end2AttributeDescription     = "Authentication identifiers in use by the owner of this profile.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0112 People describe the people, their roles and their peer network.
     */
    private void add0112People()
    {
        this.archiveBuilder.addEntityDef(getPersonEntity());
        this.archiveBuilder.addEntityDef(getContributionRecordEntity());
        this.archiveBuilder.addEntityDef(getPersonRoleEntity());

        this.archiveBuilder.addRelationshipDef(getPersonRoleAppointmentRelationship());
        this.archiveBuilder.addRelationshipDef(getPersonalContributionRelationship());
        this.archiveBuilder.addRelationshipDef(getPeerRelationship());
    }


    private EntityDef getPersonEntity()
    {
        final String guid            = "ac406bf8-e53e-49f1-9088-2af28bbbd285";
        final String name            = "Person";
        final String description     = "An individual.";
        final String descriptionGUID = null;

        final String superTypeName = "ActorProfile";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "fullName";
        final String attribute1Description     = "Full or official name of the individual (if different from known name).";
        final String attribute1DescriptionGUID = null;

        final String attribute2Name            = "jobTitle";
        final String attribute2Description     = "Role or level in the organization.";
        final String attribute2DescriptionGUID = null;

        final String attribute3Name            = OpenMetadataProperty.IS_PUBLIC.name;
        final String attribute3Description     = "Is the profile visible to other collaborators?";
        final String attribute3DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        property = archiveHelper.getBooleanTypeDefAttribute(attribute3Name,
                                                            attribute3Description,
                                                            attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getContributionRecordEntity()
    {
        final String guid            = "ac406bf8-e53e-49f1-9088-2af28cccd285";
        final String name            = "ContributionRecord";
        final String description     = "A record of the contribution of an individual.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;


        final String attribute1Name            = OpenMetadataProperty.IS_PUBLIC.name;
        final String attribute1Description     = "Is the contribution visible to other collaborators?";
        final String attribute1DescriptionGUID = null;

        final String attribute2Name            = "karmaPoints";
        final String attribute2Description     = "Points capturing a person's engagement with open metadata.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getBooleanTypeDefAttribute(attribute1Name,
                                                            attribute1Description,
                                                            attribute1DescriptionGUID);
        properties.add(property);

        property = archiveHelper.getLongTypeDefAttribute(attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }

    private EntityDef getPersonRoleEntity()
    {
        final String guid            = "ac406bf8-e53e-49f1-9088-2af28bcbd285";
        final String name            = "PersonRole";
        final String description     = "A role performed by one or more individuals.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NAME.name;
        final String attribute1Description     = OpenMetadataProperty.NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;

        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        final String attribute3Name            = "scope";
        final String attribute3Description     = "Breadth of responsibility.";
        final String attribute3DescriptionGUID = null;

        final String attribute4Name            = "headCount";
        final String attribute4Description     = "Number of people that can be appointed to the role.";
        final String attribute4DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        property = archiveHelper.getIntTypeDefAttribute(attribute4Name,
                                                        attribute4Description,
                                                        attribute4DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getPersonRoleAppointmentRelationship()
    {
        final String guid            = "4a316abe-bcce-4d11-ad5a-4bfb4079b80b";
        final String name            = "PersonRoleAppointment";
        final String description     = "Relationship identifying a person's roles.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Person";
        final String                     end1AttributeName            = "rolePerformers";
        final String                     end1AttributeDescription     = "A person performing this role.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "PersonRole";
        final String                     end2AttributeName            = "performsRoles";
        final String                     end2AttributeDescription     = "A role performed by this person.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.IS_PUBLIC.name;
        final String attribute1Description     = "Is the appointment visible to more than the role owner and appointee?";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getBooleanTypeDefAttribute(attribute1Name,
                                                            attribute1Description,
                                                            attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getPersonalContributionRelationship()
    {
        final String guid            = "4a316abe-eeee-4d11-ad5a-4bfb4079b80b";
        final String name            = "PersonalContribution";
        final String description     = "Relationship identifying a person's contribution record.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Person";
        final String                     end1AttributeName            = "contributor";
        final String                     end1AttributeDescription     = "The person behind the contribution.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ContributionRecord";
        final String                     end2AttributeName            = "contributionRecord";
        final String                     end2AttributeDescription     = "The record of activity by this person.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getPeerRelationship()
    {
        final String guid            = "4a316abe-bccd-4d11-ad5a-4bfb4079b80b";
        final String name            = "Peer";
        final String description     = "Relationship identifying a person's peer network.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Person";
        final String                     end1AttributeName            = "peers";
        final String                     end1AttributeDescription     = "List of this person's peer network.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Person";
        final String                     end2AttributeName            = "peers";
        final String                     end2AttributeDescription     = "List of this person's peer network.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0115 Teams describe the organization of people.
     */
    private void add0115Teams()
    {
        this.archiveBuilder.addEntityDef(getTeamEntity());
        this.archiveBuilder.addEntityDef(getTeamLeaderEntity());
        this.archiveBuilder.addEntityDef(getTeamMemberEntity());

        this.archiveBuilder.addRelationshipDef(getTeamLeadershipRelationship());
        this.archiveBuilder.addRelationshipDef(getTeamMembershipRelationship());
        this.archiveBuilder.addRelationshipDef(getTeamStructureRelationship());
    }


    private EntityDef getTeamEntity()
    {
        final String guid            = "36db26d5-aba2-439b-bc15-d62d373c5db6";
        final String name            = "Team";
        final String description     = "Group of people working together.";
        final String descriptionGUID = null;

        final String superTypeName = "ActorProfile";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "teamType";
        final String attribute1Description     = "Type of team, such as department.";
        final String attribute1DescriptionGUID = null;


        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }

    private EntityDef getTeamLeaderEntity()
    {
        final String guid            = "36db26d5-abb2-439b-bc15-d62d373c5db6";
        final String name            = "TeamLeader";
        final String description     = "Person leading a team.";
        final String descriptionGUID = null;

        final String superTypeName = "PersonRole";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }

    private EntityDef getTeamMemberEntity()
    {
        final String guid            = "46db26d5-abb2-538b-bc15-d62d373c5db6";
        final String name            = "TeamMember";
        final String description     = "Person assigned to a team.";
        final String descriptionGUID = null;

        final String superTypeName = "PersonRole";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }

    private RelationshipDef getTeamStructureRelationship()
    {
        final String guid            = "5ebc4fb2-b62a-4269-8f18-e9237a2229ca";
        final String name            = "TeamStructure";
        final String description     = "Relationship identifying a team hierarchy.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Team";
        final String                     end1AttributeName            = "superTeam";
        final String                     end1AttributeDescription     = "The aggregating team.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Team";
        final String                     end2AttributeName            = "subTeam";
        final String                     end2AttributeDescription     = "The teams where work is delegated to.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "delegationEscalationAuthority";
        final String attribute1Description     = "Can delegations and escalations flow on this relationship.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getBooleanTypeDefAttribute(attribute1Name,
                                                            attribute1Description,
                                                            attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getTeamLeadershipRelationship()
    {
        final String guid            = "5ebc4fb2-b62a-4269-8f18-e9237a2119ca";
        final String name            = "TeamLeadership";
        final String description     = "Relationship identifying the leaders of teams.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "TeamLeader";
        final String                     end1AttributeName            = "teamLeaders";
        final String                     end1AttributeDescription     = "The leaders of the team.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Team";
        final String                     end2AttributeName            = "leadsTeam";
        final String                     end2AttributeDescription     = "The team lead by this team leader.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "position";
        final String attribute1Description     = "Details of the type of leadership position, eg deputy.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getTeamMembershipRelationship()
    {
        final String guid            = "1ebc4fb2-b62a-4269-8f18-e9237a2119ca";
        final String name            = "TeamMembership";
        final String description     = "Relationship identifying the members of teams.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "TeamMember";
        final String                     end1AttributeName            = "teamMembers";
        final String                     end1AttributeDescription     = "The members of the team.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Team";
        final String                     end2AttributeName            = "memberOfTeam";
        final String                     end2AttributeDescription     = "The team that this person is a member of.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "position";
        final String attribute1Description     = "Details of the type of membership position, if any.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0117 IT profiles defines a special actor profile for an engine
     */
    private void add0117EngineProfiles()
    {
        this.archiveBuilder.addEntityDef(getITProfileEntity());
    }

    private EntityDef getITProfileEntity()
    {
        final String guid            = "81394f85-6008-465b-926e-b3fae4668937";
        final String name            = "ITProfile";
        final String description     = "Descriptive details about a processing engine or other IT infrastructure.";
        final String descriptionGUID = null;

        final String superTypeName = "ActorProfile";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0120 Collections defines how to group related Referenceables together
     */
    private void add0120Collections()
    {
        this.archiveBuilder.addEnumDef(getOrderByEnum());

        this.archiveBuilder.addEntityDef(getCollectionEntity());

        this.archiveBuilder.addRelationshipDef(getCollectionMembershipRelationship());
        this.archiveBuilder.addRelationshipDef(getResourceListRelationship());

        this.archiveBuilder.addClassificationDef(getFolderClassification());
        this.archiveBuilder.addClassificationDef(getSetClassification());
    }


    private EnumDef getOrderByEnum()
    {
        final String guid            = OrderBy.getOpenTypeGUID();
        final String name            = OrderBy.getOpenTypeName();
        final String description     = OrderBy.getOpenTypeDescription();
        final String descriptionGUID = OrderBy.getOpenTypeDescriptionGUID();

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID);

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (OrderBy orderBy : OrderBy.values())
        {
            elementDef = archiveHelper.getEnumElementDef(orderBy.getOrdinal(),
                                                         orderBy.getName(),
                                                         orderBy.getDescription(),
                                                         orderBy.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (orderBy.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getCollectionEntity()
    {
        final String guid            = OpenMetadataType.COLLECTION.typeGUID;
        final String name            = OpenMetadataType.COLLECTION.typeName;
        final String description     = OpenMetadataType.COLLECTION.description;
        final String descriptionGUID = OpenMetadataType.COLLECTION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.COLLECTION.wikiURL;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NAME.name;
        final String attribute1Description     = OpenMetadataProperty.NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getCollectionMembershipRelationship()
    {
        final String guid            = OpenMetadataType.COLLECTION_MEMBERSHIP_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.COLLECTION_MEMBERSHIP_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.COLLECTION_MEMBERSHIP_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.COLLECTION_MEMBERSHIP_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.COLLECTION_MEMBERSHIP_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.COLLECTION.typeName;
        final String                     end1AttributeName            = "foundInCollections";
        final String                     end1AttributeDescription     = "Collections that link to this element.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end2AttributeName            = "collectionMembers";
        final String                     end2AttributeDescription     = "Members of this collection.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.MEMBERSHIP_RATIONALE.name;
        final String attribute1Description     = OpenMetadataProperty.MEMBERSHIP_RATIONALE.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.MEMBERSHIP_RATIONALE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getResourceListRelationship()
    {
        final String guid            = OpenMetadataType.RESOURCE_LIST_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.RESOURCE_LIST_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.RESOURCE_LIST_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.RESOURCE_LIST_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.RESOURCE_LIST_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "resourceListAnchors";
        final String                     end1AttributeDescription     = "Referenceable objects that are using the linked to resource.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end2AttributeName            = "supportingResources";
        final String                     end2AttributeDescription     = "Resources identified as of interest to the anchor.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "resourceUse";
        final String attribute1Description     = "Identifier that describes the type of resource use.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "resourceUseDescription";
        final String attribute2Description     = "Description of how the resource is used, or why it is useful.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "resourceUseProperties";
        final String attribute3Description     = "Additional properties that explains how to use the resource.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "watchResource";
        final String attribute4Description     = "Indicator whether the anchor should receive notifications of changes to the resource.";
        final String attribute4DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute3Name,
                                                                    attribute3Description,
                                                                    attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getBooleanTypeDefAttribute(attribute4Name,
                                                            attribute4Description,
                                                            attribute4DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private ClassificationDef getFolderClassification()
    {
        final String guid            = OpenMetadataType.FOLDER.typeGUID;
        final String name            = OpenMetadataType.FOLDER.typeName;
        final String description     = OpenMetadataType.FOLDER.description;
        final String descriptionGUID = OpenMetadataType.FOLDER.descriptionGUID;

        final String linkedToEntity = OpenMetadataType.COLLECTION.typeName;

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "orderBy";
        final String attribute1Description     = "Definition for how elements in the collection should be ordered.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "orderPropertyName";
        final String attribute2Description     = "Name of property to use for ordering.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getEnumTypeDefAttribute("OrderBy",
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private ClassificationDef getSetClassification()
    {
        final String guid            = OpenMetadataType.RESULTS_SET.typeGUID;
        final String name            = OpenMetadataType.RESULTS_SET.typeName;
        final String description     = OpenMetadataType.RESULTS_SET.description;
        final String descriptionGUID = OpenMetadataType.RESULTS_SET.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.RESULTS_SET.wikiURL;

        final String linkedToEntity = OpenMetadataType.COLLECTION.typeName;

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  descriptionWiki,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0130 Projects describes the structure of a project and related collections.
     */
    private void add0130Projects()
    {
        this.archiveBuilder.addEntityDef(getProjectEntity());
        this.archiveBuilder.addEntityDef(getProjectManagerEntity());

        this.archiveBuilder.addRelationshipDef(getProjectHierarchyRelationship());
        this.archiveBuilder.addRelationshipDef(getProjectDependencyRelationship());
        this.archiveBuilder.addRelationshipDef(getProjectTeamRelationship());
        this.archiveBuilder.addRelationshipDef(getProjectScopeRelationship());
        this.archiveBuilder.addRelationshipDef(getProjectManagementRelationship());

        this.archiveBuilder.addClassificationDef(getTaskClassification());
        this.archiveBuilder.addClassificationDef(getCampaignClassification());
    }


    private EntityDef getProjectEntity()
    {
        final String guid            = OpenMetadataType.PROJECT.typeGUID;
        final String name            = OpenMetadataType.PROJECT.typeName;
        final String description     = OpenMetadataType.PROJECT.description;
        final String descriptionGUID = OpenMetadataType.PROJECT.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.PROJECT.wikiURL;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NAME.name;
        final String attribute1Description     = OpenMetadataProperty.NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "startDate";
        final String attribute3Description     = "Start date of the project.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "plannedEndDate";
        final String attribute4Description     = "Planned completion data for the project.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "status";
        final String attribute5Description     = "Short description on current status of the project.";
        final String attribute5DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute4Name,
                                                         attribute4Description,
                                                         attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getProjectManagerEntity()
    {
        final String guid            = OpenMetadataType.PROJECT_MANAGER.typeGUID;
        final String name            = OpenMetadataType.PROJECT_MANAGER.typeName;
        final String description     = OpenMetadataType.PROJECT_MANAGER.description;
        final String descriptionGUID = OpenMetadataType.PROJECT_MANAGER.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.PROJECT_MANAGER.wikiURL;

        final String superTypeName = "PersonRole";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    private RelationshipDef getProjectHierarchyRelationship()
    {
        final String guid            = OpenMetadataType.PROJECT_HIERARCHY_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.PROJECT_HIERARCHY_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.PROJECT_HIERARCHY_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.PROJECT_HIERARCHY_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.PROJECT_HIERARCHY_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.PROJECT.typeName;
        final String                     end1AttributeName            = "managingProject";
        final String                     end1AttributeDescription     = "Project that oversees this project.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.PROJECT.typeName;
        final String                     end2AttributeName            = "managedProject";
        final String                     end2AttributeDescription     = "Project that this project is responsible for managing.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getProjectDependencyRelationship()
    {
        final String guid            = OpenMetadataType.PROJECT_DEPENDENCY_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.PROJECT_DEPENDENCY_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.PROJECT_DEPENDENCY_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.PROJECT_DEPENDENCY_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.PROJECT_DEPENDENCY_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.PROJECT.typeName;
        final String                     end1AttributeName            = "dependentProject";
        final String                     end1AttributeDescription     = "Projects that are dependent on this project.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.PROJECT.typeName;
        final String                     end2AttributeName            = "dependsOnProjects";
        final String                     end2AttributeDescription     = "Projects that are delivering resources or outcomes needed by this project.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "dependencySummary";
        final String attribute1Description     = "Reasons for the project dependency.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getDateTypeDefAttribute(attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getProjectTeamRelationship()
    {
        final String guid            = OpenMetadataType.PROJECT_TEAM_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.PROJECT_TEAM_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.PROJECT_TEAM_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.PROJECT_TEAM_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.PROJECT_TEAM_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.PROJECT.typeName;
        final String                     end1AttributeName            = "projectFocus";
        final String                     end1AttributeDescription     = "Projects that a team is working on.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ActorProfile";
        final String                     end2AttributeName            = "supportingActors";
        final String                     end2AttributeDescription     = "People and teams supporting this project.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "teamRole";
        final String attribute1Description     = "Description of the role of the team in the project.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getDateTypeDefAttribute(attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getProjectScopeRelationship()
    {
        final String guid            = "bc63ac45-b4d0-4fba-b583-92859de77dd8";
        final String name            = "ProjectScope";
        final String description     = "The documentation, assets and definitions that are affected by the project.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.PROJECT.typeName;
        final String                     end1AttributeName            = "impactingProjects";
        final String                     end1AttributeDescription     = "The projects that are making changes to these elements.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end2AttributeName            = "projectScope";
        final String                     end2AttributeDescription     = "The elements that are being changed by this project.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "scopeDescription";
        final String attribute1Description     = "Description of how each item is being changed by the project.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getDateTypeDefAttribute(attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /**
     * Add Task classification
     *
     * @return classification def
     */
    private ClassificationDef getTaskClassification()
    {
        final String guid            = OpenMetadataType.TASK_CLASSIFICATION.typeGUID;
        final String name            = OpenMetadataType.TASK_CLASSIFICATION.typeName;
        final String description     = OpenMetadataType.TASK_CLASSIFICATION.description;
        final String descriptionGUID = OpenMetadataType.TASK_CLASSIFICATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.TASK_CLASSIFICATION.wikiURL;

        final String linkedToEntity = OpenMetadataType.PROJECT.typeName;

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  descriptionWiki,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private RelationshipDef getProjectManagementRelationship()
    {
        final String guid            = OpenMetadataType.PROJECT_MANAGEMENT_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.PROJECT_MANAGEMENT_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.PROJECT_MANAGEMENT_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.PROJECT_MANAGEMENT_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.PROJECT_MANAGEMENT_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.PROJECT.typeName;
        final String                     end1AttributeName            = "projectsManaged";
        final String                     end1AttributeDescription     = "The projects that are being managed by this project manager.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.PROJECT_MANAGER.typeName;
        final String                     end2AttributeName            = "projectManagers";
        final String                     end2AttributeDescription     = "The roles for managing this project.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }

    private ClassificationDef getCampaignClassification()
    {
        final String guid            = OpenMetadataType.CAMPAIGN_CLASSIFICATION.typeGUID;
        final String name            = OpenMetadataType.CAMPAIGN_CLASSIFICATION.typeName;
        final String description     = OpenMetadataType.CAMPAIGN_CLASSIFICATION.description;
        final String descriptionGUID = OpenMetadataType.CAMPAIGN_CLASSIFICATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.CAMPAIGN_CLASSIFICATION.wikiURL;

        final String linkedToEntity = OpenMetadataType.COLLECTION.typeName;

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  descriptionWiki,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0135 Meetings defines how to record meetings and todos.
     */
    private void add0135Meetings()
    {
        this.archiveBuilder.addEntityDef(getMeetingEntity());

        this.archiveBuilder.addRelationshipDef(getMeetingsRelationship());
    }


    private EntityDef getMeetingEntity()
    {
        final String guid            = "6bf90c79-32f4-47ad-959c-8fff723fe744";
        final String name            = "Meeting";
        final String description     = "Two or more people come together to discuss a topic, agree and action or exchange information.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "title";
        final String attribute1Description     = "Title of the meeting.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "startTime";
        final String attribute2Description     = "Start time of the meeting.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "endTime";
        final String attribute3Description     = "End time of the meeting.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "objective";
        final String attribute4Description     = "Reason for the meeting and intended outcome.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "minutes";
        final String attribute5Description     = "Description of what happened at the meeting.";
        final String attribute5DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute5Name,
                                                                attribute5Description,
                                                                attribute5DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }

    private RelationshipDef getMeetingsRelationship()
    {
        final String guid            = "a05f918e-e7e2-419d-8016-5b37406df63a";
        final String name            = "Meetings";
        final String description     = "A meeting about a specific project, deliverable, situation or plan of action.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Meeting";
        final String                     end1AttributeName            = "meetings";
        final String                     end1AttributeDescription     = "Related meetings.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end2AttributeName            = "meetingOwner";
        final String                     end2AttributeDescription     = "Person, project, community or team that called the meeting.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0135 Meetings defines how to record meetings and todos.
     */
    private void add0137Actions()
    {
        this.archiveBuilder.addEnumDef(getToDoStatusEnum());

        this.archiveBuilder.addEntityDef(getToDoEntity());

        this.archiveBuilder.addRelationshipDef(getToDoSourceRelationship());
        this.archiveBuilder.addRelationshipDef(getActionSponsorRelationship());
        this.archiveBuilder.addRelationshipDef(getActionAssignment());
    }


    private EnumDef getToDoStatusEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(ToDoStatus.getOpenTypeGUID(),
                                                        ToDoStatus.getOpenTypeName(),
                                                        ToDoStatus.getOpenTypeDescription(),
                                                        ToDoStatus.getOpenTypeDescriptionGUID(),
                                                        ToDoStatus.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (ToDoStatus toDoStatus : ToDoStatus.values())
        {
            elementDef = archiveHelper.getEnumElementDef(toDoStatus.getOrdinal(),
                                                         toDoStatus.getName(),
                                                         toDoStatus.getDescription(),
                                                         toDoStatus.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (toDoStatus.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getToDoEntity()
    {
        EntityDef entityDef = archiveHelper.getDefaultEntityDef(OpenMetadataType.TO_DO.typeGUID,
                                                                OpenMetadataType.TO_DO.typeName,
                                                                this.archiveBuilder.getEntityDef(OpenMetadataType.REFERENCEABLE.typeName),
                                                                OpenMetadataType.TO_DO.description,
                                                                OpenMetadataType.TO_DO.descriptionGUID,
                                                                OpenMetadataType.TO_DO.wikiURL);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NAME.name;
        final String attribute1Description     = OpenMetadataProperty.NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "creationTime";
        final String attribute3Description     = "When the requested action was identified.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.PRIORITY.name;
        final String attribute4Description     = OpenMetadataProperty.PRIORITY.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.PRIORITY.descriptionGUID;
        final String attribute5Name            = "dueTime";
        final String attribute5Description     = "When the requested action needs to be completed.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = OpenMetadataProperty.TO_DO_STATUS.name;
        final String attribute6Description     = OpenMetadataProperty.TO_DO_STATUS.description;
        final String attribute6DescriptionGUID = OpenMetadataProperty.TO_DO_STATUS.descriptionGUID;
        final String attribute7Name            = "completionTime";
        final String attribute7Description     = "When the requested action was completed.";
        final String attribute7DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute4Name,
                                                        attribute4Description,
                                                        attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute5Name,
                                                         attribute5Description,
                                                         attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(ToDoStatus.getOpenTypeName(),
                                                         attribute6Name,
                                                         attribute6Description,
                                                         attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute7Name,
                                                         attribute7Description,
                                                         attribute7DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getToDoSourceRelationship()
    {
        final String guid            = "a0b7ba50-4c97-4b76-9a7d-c6a00e1be646";
        final String name            = "ToDoSource";
        final String description     = "The source of the to do, such as a person, meeting or a governance action.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "actionSource";
        final String                     end1AttributeDescription     = "Source of the to do request.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ToDo";
        final String                     end2AttributeName            = "actions";
        final String                     end2AttributeDescription     = "Requests to perform actions related to this element.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getActionSponsorRelationship()
    {
        final String guid            = "aca1277b-bf1c-42f5-9b3b-fbc2c9047325";
        final String name            = "ActionSponsor";
        final String description     = "Identifies the sponsor that requires the action (ToDo) to be completed.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "toDoSponsor";
        final String                     end1AttributeDescription     = "Element such as person, team, rule, incident, project, that is driving the need for the action.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ToDo";
        final String                     end2AttributeName            = "trackedActions";
        final String                     end2AttributeDescription     = "Actions that need to be completed.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }

    private RelationshipDef getActionAssignment()
    {
        final String guid            = "af2b5fab-8f83-4a2b-b749-1e6219f61f79";
        final String name            = "ActionAssignment";
        final String description     = "A person who has been assigned to complete the to do (action).";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "PersonRole";
        final String                     end1AttributeName            = "assignedResources";
        final String                     end1AttributeDescription     = "One or more people assigned to complete the action (to do).";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ToDo";
        final String                     end2AttributeName            = "toDoList";
        final String                     end2AttributeDescription     = "List of toDos assigned to this person's role.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0140 Communities describe communities of people who have similar interests.
     */
    private void add0140Communities()
    {
        this.archiveBuilder.addEnumDef(getCommunityMembershipTypeEnum());

        this.archiveBuilder.addEntityDef(getCommunityEntity());
        this.archiveBuilder.addEntityDef(getCommunityMemberEntity());

        this.archiveBuilder.addRelationshipDef(getCommunityMembershipRelationship());
    }


    private EnumDef getCommunityMembershipTypeEnum()
    {
        final String guid            = "b0ef45bf-d12b-4b6f-add6-59c14648d750";
        final String name            = "CommunityMembershipType";
        final String description     = "Type of membership to a community.";
        final String descriptionGUID = null;

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID);

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        final int    element1Ordinal         = 0;
        final String element1Value           = "Contributor";
        final String element1Description     = "Participant in the community.";
        final String element1DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element1Ordinal,
                                                     element1Value,
                                                     element1Description,
                                                     element1DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element2Ordinal         = 1;
        final String element2Value           = "Administrator";
        final String element2Description     = "Administrator of the community.";
        final String element2DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element2Ordinal,
                                                     element2Value,
                                                     element2Description,
                                                     element2DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element3Ordinal         = 2;
        final String element3Value           = "Leader";
        final String element3Description     = "Leader of the community.";
        final String element3DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element3Ordinal,
                                                     element3Value,
                                                     element3Description,
                                                     element3DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element4Ordinal         = 3;
        final String element4Value           = "Observer";
        final String element4Description     = "Observer of the community.";
        final String element4DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element4Ordinal,
                                                     element4Value,
                                                     element4Description,
                                                     element4DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element99Ordinal         = 99;
        final String element99Value           = "Other";
        final String element99Description     = "Another role in the community.";
        final String element99DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element99Ordinal,
                                                     element99Value,
                                                     element99Description,
                                                     element99DescriptionGUID);
        elementDefs.add(elementDef);

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getCommunityEntity()
    {
        final String guid            = "fbd42379-f6c3-4f08-b6f7-378565cda993";
        final String name            = "Community";
        final String description     = "A group of people with a common interest or skill.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NAME.name;
        final String attribute1Description     = OpenMetadataProperty.NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "mission";
        final String attribute3Description     = "Purpose of the community.";
        final String attribute3DescriptionGUID = null;


        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getCommunityMemberEntity()
    {
        final String guid            = "fbd42379-f6c3-4f09-b6f7-378565cda993";
        final String name            = "CommunityMember";
        final String description     = "A person who has joined a community.";
        final String descriptionGUID = null;

        final String superTypeName = "PersonRole";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getCommunityMembershipRelationship()
    {
        final String guid            = "7c7da1a3-01b3-473e-972e-606eff0cb112";
        final String name            = "CommunityMembership";
        final String description     = "Associates an actor profile with a community.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Community";
        final String                     end1AttributeName            = "memberOfCommunity";
        final String                     end1AttributeDescription     = "Communities that the person is a member of.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "CommunityMember";
        final String                     end2AttributeName            = "communityMembers";
        final String                     end2AttributeDescription     = "Members of the community.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "membershipType";
        final String attribute1Description     = "Type of membership to the community.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getEnumTypeDefAttribute("CommunityMembershipType",
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0150 Feedback provides all of the collaborative feedback and attachments that can be made by the user
     * community of the Assets.
     */
    private void add0150Feedback()
    {
        this.archiveBuilder.addEnumDef(getStarRatingEnum());
        this.archiveBuilder.addEnumDef(getCommentTypeEnum());

        this.archiveBuilder.addEntityDef(getRatingEntity());
        this.archiveBuilder.addEntityDef(getCommentEntity());
        this.archiveBuilder.addEntityDef(getLikeEntity());
        this.archiveBuilder.addEntityDef(getInformalTagEntity());

        this.archiveBuilder.addRelationshipDef(getAttachedRatingRelationship());
        this.archiveBuilder.addRelationshipDef(getAttachedCommentRelationship());
        this.archiveBuilder.addRelationshipDef(getAttachedLikeRelationship());
        this.archiveBuilder.addRelationshipDef(getAcceptedAnswerRelationship());
        this.archiveBuilder.addRelationshipDef(getAttachedTagRelationship());
    }


    private EnumDef getStarRatingEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(StarRating.getOpenTypeGUID(),
                                                        StarRating.getOpenTypeName(),
                                                        StarRating.getOpenTypeDescription(),
                                                        StarRating.getOpenTypeDescriptionGUID(),
                                                        StarRating.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (StarRating enumValue : StarRating.values())
        {
            elementDef = archiveHelper.getEnumElementDef(enumValue.getOrdinal(),
                                                         enumValue.getName(),
                                                         enumValue.getDescription(),
                                                         enumValue.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (enumValue.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EnumDef getCommentTypeEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(CommentType.getOpenTypeGUID(),
                                                        CommentType.getOpenTypeName(),
                                                        CommentType.getOpenTypeDescription(),
                                                        CommentType.getOpenTypeDescriptionGUID(),
                                                        CommentType.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (CommentType commentType : CommentType.values())
        {
            elementDef = archiveHelper.getEnumElementDef(commentType.getOrdinal(),
                                                         commentType.getName(),
                                                         commentType.getDescription(),
                                                         commentType.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (commentType.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getRatingEntity()
    {
        final String guid            = "7299d721-d17f-4562-8286-bcd451814478";
        final String name            = "Rating";
        final String description     = "Quantitative feedback related to an item.";
        final String descriptionGUID = null;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                null,
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "anchorGUID";
        final String attribute1Description     = "Optional identifier of the referenceable that this rating is connected to (either directly or indirectly).";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "stars";
        final String attribute2Description     = "Rating level provided.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "review";
        final String attribute3Description     = "Additional comments associated with the rating.";
        final String attribute3DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("StarRating",
                                                         attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getCommentEntity()
    {
         EntityDef entityDef = archiveHelper.getDefaultEntityDef(OpenMetadataType.COMMENT.typeGUID,
                                                                OpenMetadataType.COMMENT.typeName,
                                                                this.archiveBuilder.getEntityDef(OpenMetadataType.REFERENCEABLE.typeName),
                                                                OpenMetadataType.COMMENT.description,
                                                                OpenMetadataType.COMMENT.descriptionGUID,
                                                                OpenMetadataType.COMMENT.wikiURL);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "anchorGUID";
        final String attribute1Description     = "Optional identifier of the referenceable that this comment is connected to (either directly or indirectly).";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "text";
        final String attribute2Description     = "Feedback comments or additional information.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "type";
        final String attribute3Description     = "Type of comment.";
        final String attribute3DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("CommentType",
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getLikeEntity()
    {
        final String guid            = "deaa5ca0-47a0-483d-b943-d91c76744e01";
        final String name            = "Like";
        final String description     = "Boolean type of rating expressing a favorable impression.";
        final String descriptionGUID = null;

        EntityDef entityDef =  archiveHelper.getDefaultEntityDef(guid,
                                                                 name,
                                                                 null,
                                                                 description,
                                                                 descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "anchorGUID";
        final String attribute1Description     = "Optional identifier of the referenceable that this comment is connected to (either directly or indirectly).";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getInformalTagEntity()
    {
        final String guid            = "ba846a7b-2955-40bf-952b-2793ceca090a";
        final String name            = "InformalTag";
        final String description     = "An descriptive tag for an item.";
        final String descriptionGUID = null;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                null,
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.IS_PUBLIC.name;
        final String attribute1Description     = "Is the tag visible to more than the originator?";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "tagName";
        final String attribute2Description     = "Descriptive name of the tag.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "tagDescription";
        final String attribute3Description     = "More detail on the meaning of the tag.";
        final String attribute3DescriptionGUID = null;

        property = archiveHelper.getBooleanTypeDefAttribute(attribute1Name,
                                                            attribute1Description,
                                                            attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getAttachedRatingRelationship()
    {
        final String guid            = "0aaad9e9-9cc5-4ad8-bc2e-c1099bab6344";
        final String name            = "AttachedRating";
        final String description     = "Links a rating to an item.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "ratingAnchor";
        final String                     end1AttributeDescription     = "Element that is rated.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Rating";
        final String                     end2AttributeName            = "starRatings";
        final String                     end2AttributeDescription     = "Accumulated ratings.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.IS_PUBLIC.name;
        final String attribute1Description     = "Is the attached rating visible to more than the originator?";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getBooleanTypeDefAttribute(attribute1Name,
                                                            attribute1Description,
                                                            attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getAttachedCommentRelationship()
    {
        final String guid            = "0d90501b-bf29-4621-a207-0c8c953bdac9";
        final String name            = "AttachedComment";
        final String description     = "Links a comment to an item, or another comment.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "commentAnchor";
        final String                     end1AttributeDescription     = "Element that this comment relates.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Comment";
        final String                     end2AttributeName            = "comments";
        final String                     end2AttributeDescription     = "Accumulated comments.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.IS_PUBLIC.name;
        final String attribute1Description     = "Is the attached comment visible to more than the originator?";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getBooleanTypeDefAttribute(attribute1Name,
                                                            attribute1Description,
                                                            attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getAttachedLikeRelationship()
    {
        final String guid            = "e2509715-a606-415d-a995-61d00503dad4";
        final String name            = "AttachedLike";
        final String description     = "Links a like to an item.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "likeAnchor";
        final String                     end1AttributeDescription     = "Element that is liked.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Like";
        final String                     end2AttributeName            = "likes";
        final String                     end2AttributeDescription     = "Accumulated likes.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.IS_PUBLIC.name;
        final String attribute1Description     = "Is the attached like visible to more than the originator?";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getBooleanTypeDefAttribute(attribute1Name,
                                                            attribute1Description,
                                                            attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getAcceptedAnswerRelationship()
    {
        final String guid            = "ecf1a3ca-adc5-4747-82cf-10ec590c5c69";
        final String name            = "AcceptedAnswer";
        final String description     = "Identifies a comment as answering a question asked in another comment.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Comment";
        final String                     end1AttributeName            = "answeredQuestions";
        final String                     end1AttributeDescription     = "Questions that now has an accepted answer.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Comment";
        final String                     end2AttributeName            = "acceptedAnswers";
        final String                     end2AttributeDescription     = "Accumulated answers.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.IS_PUBLIC.name;
        final String attribute1Description     = "Is the attached answer visible to more than the originator?";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getBooleanTypeDefAttribute(attribute1Name,
                                                            attribute1Description,
                                                            attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getAttachedTagRelationship()
    {
        final String guid            = "4b1641c4-3d1a-4213-86b2-d6968b6c65ab";
        final String name            = "AttachedTag";
        final String description     = "Links an informal tag to an item.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "taggedElement";
        final String                     end1AttributeDescription     = "Element that is tagged.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "InformalTag";
        final String                     end2AttributeName            = "tags";
        final String                     end2AttributeDescription     = "Accumulated tags.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.IS_PUBLIC.name;
        final String attribute1Description     = "Is the attached tag visible to more than the originator?";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getBooleanTypeDefAttribute(attribute1Name,
                                                            attribute1Description,
                                                            attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0155 Crowd Sourcing describes the people involved in crowd sourcing new metadata content.
     */
    private void add0155CrowdSourcing()
    {
        this.archiveBuilder.addEnumDef(getCrowdSourcingRoleEnum());

        this.archiveBuilder.addEntityDef(getCrowdSourcingContributorEntity());

        this.archiveBuilder.addRelationshipDef(getCrowdSourcingContributionRelationship());
    }


    private EnumDef getCrowdSourcingRoleEnum()
    {
        final String guid            = "0ded50c2-17cc-4ecf-915e-908e66dbb27f";
        final String name            = "CrowdSourcingRole";
        final String description     = "Type of contributor to new information and/or assets.";
        final String descriptionGUID = null;

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID);

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        final int    element1Ordinal         = 0;
        final String element1Value           = "Proposer";
        final String element1Description     = "Actor that creates the initial version.";
        final String element1DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element1Ordinal,
                                                     element1Value,
                                                     element1Description,
                                                     element1DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element2Ordinal         = 1;
        final String element2Value           = "Reviewer";
        final String element2Description     = "Actor that provided feedback.";
        final String element2DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element2Ordinal,
                                                     element2Value,
                                                     element2Description,
                                                     element2DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element3Ordinal         = 2;
        final String element3Value           = "Supporter";
        final String element3Description     = "Actor that agrees with the definition.";
        final String element3DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element3Ordinal,
                                                     element3Value,
                                                     element3Description,
                                                     element3DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element4Ordinal         = 3;
        final String element4Value           = "Approver";
        final String element4Description     = "Actor that declares the definition should be used.";
        final String element4DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element4Ordinal,
                                                     element4Value,
                                                     element4Description,
                                                     element4DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element5Ordinal         = 99;
        final String element5Value           = "Other";
        final String element5Description     = "Another role.";
        final String element5DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element5Ordinal,
                                                     element5Value,
                                                     element5Description,
                                                     element5DescriptionGUID);
        elementDefs.add(elementDef);

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getCrowdSourcingContributorEntity()
    {
        final String guid            = "3a84c94c-ac6f-4be1-a72a-07dcec7b1fe3";
        final String name            = "CrowdSourcingContributor";
        final String description     = "Person contributing new content.";
        final String descriptionGUID = null;

        final String superTypeName = "PersonRole";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }

    private RelationshipDef getCrowdSourcingContributionRelationship()
    {
        final String guid            = "4db83564-b200-4956-94a4-c95a5c30e65a";
        final String name            = "CrowdSourcingContribution";
        final String description     = "Defines one of the actors contributing content to a new description or asset.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "contributions";
        final String                     end1AttributeDescription     = "Items that this person has contributed.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "CrowdSourcingContributor";
        final String                     end2AttributeName            = "contributors";
        final String                     end2AttributeDescription     = "Person contributing content.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "roleType";
        final String attribute1Description     = "Type of contribution.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getEnumTypeDefAttribute("CrowdSourcingRole",
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0160 Notes describes note logs and notes within them.  Notes are kept by the owners/stewards working on the
     * Assets.
     */
    private void add0160Notes()
    {
        this.archiveBuilder.addEntityDef(getNoteEntryEntity());
        this.archiveBuilder.addEntityDef(getNoteLogEntity());
        this.archiveBuilder.addEntityDef(getNoteLogAuthorEntity());

        this.archiveBuilder.addRelationshipDef(getAttachedNoteLogRelationship());
        this.archiveBuilder.addRelationshipDef(getAttachedNoteLogEntryRelationship());
        this.archiveBuilder.addRelationshipDef(getNoteLogAuthorshipRelationship());
    }

    private EntityDef getNoteEntryEntity()
    {
        final String guid            = "2a84d94c-ac6f-4be1-a72a-07dcec7b1fe3";
        final String name            = "NoteEntry";
        final String description     = "An entry in a note log.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "title";
        final String attribute1Description     = "Title of the note entry.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "text";
        final String attribute2Description     = "Text of the note entry.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = OpenMetadataProperty.IS_PUBLIC.name;
        final String attribute3Description     = "Is the note visible to more than the note log authors?";
        final String attribute3DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getBooleanTypeDefAttribute(attribute3Name,
                                                            attribute3Description,
                                                            attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getNoteLogAuthorEntity()
    {
        final String guid            = "3a84d94c-ac6f-4be1-a72a-07dbec7b1fe3";
        final String name            = "NoteLogAuthor";
        final String description     = "A person adding notes to a note log.";
        final String descriptionGUID = null;

        final String superTypeName = "PersonRole";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getNoteLogEntity()
    {
        final String guid            = "646727c7-9ad4-46fa-b660-265489ad96c6";
        final String name            = "NoteLog";
        final String description     = "An ordered list of related notes.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NAME.name;
        final String attribute1Description     = OpenMetadataProperty.NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.IS_PUBLIC.name;
        final String attribute3Description     = "Is the note log visible to more than the note log authors?";
        final String attribute3DescriptionGUID = null;


        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getBooleanTypeDefAttribute(attribute3Name,
                                                            attribute3Description,
                                                            attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getAttachedNoteLogRelationship()
    {
        final String guid            = "4f798c0c-6769-4a2d-b489-d2714d89e0a4";
        final String name            = "AttachedNoteLog";
        final String description     = "Links a note log to an item.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "describes";
        final String                     end1AttributeDescription     = "Subject of the note log.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "NoteLog";
        final String                     end2AttributeName            = "noteLogs";
        final String                     end2AttributeDescription     = "Log of related notes.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.IS_PUBLIC.name;
        final String attribute1Description     = "Is the attached note log visible to more than the originator?";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getBooleanTypeDefAttribute(attribute1Name,
                                                            attribute1Description,
                                                            attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getNoteLogAuthorshipRelationship()
    {
        final String guid            = "8f798c0c-6769-4a2d-b489-12714d89e0a4";
        final String name            = "NoteLogAuthorship";
        final String description     = "Links a note log to an author.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "NoteLogAuthor";
        final String                     end1AttributeName            = "authors";
        final String                     end1AttributeDescription     = "Person contributing to the note log.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "NoteLog";
        final String                     end2AttributeName            = "authoredNoteLogs";
        final String                     end2AttributeDescription     = "Note log containing contributions.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getAttachedNoteLogEntryRelationship()
    {
        final String guid            = "38edecc6-f385-4574-8144-524a44e3e712";
        final String name            = "AttachedNoteLogEntry";
        final String description     = "Link between a note log and one of its note log entries.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "NoteLog";
        final String                     end1AttributeName            = "logs";
        final String                     end1AttributeDescription     = "Logs that this entry relates.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "NoteEntry";
        final String                     end2AttributeName            = "entries";
        final String                     end2AttributeDescription     = "Accumulated notes.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * ========================================
     * AREA 2: connectors and assets
     */

    /**
     * Area 2 covers the different types of Assets and the Connection information used by the Open Connector Framework
     * (OCF).
     */
    private void addArea2Types()
    {
        this.add0201ConnectorsAndConnections();
        this.add0205ConnectionLinkage();
        this.add0210DataStores();
        this.add0212DeployedAPIs();
        this.add0215SoftwareComponents();
        this.add0220FilesAndFolders();
        this.add0221DocumentStores();
        this.add0222GraphStores();
        this.add0223EventsAndLogs();
        this.add0224Databases();
        this.add0225MetadataRepositories();
        this.add0227Keystores();
        this.add0230CodeTables();
        this.add0235InformationView();
        this.add0239Reports();
        this.add0290Ports();
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0201 Connectors and Connections defines the details of the Connection that describes the connector type
     * and endpoint for a specific connector instance.
     */
    private void add0201ConnectorsAndConnections()
    {
        this.archiveBuilder.addEntityDef(getConnectionEntity());
        this.archiveBuilder.addEntityDef(getConnectorTypeEntity());

        this.archiveBuilder.addRelationshipDef(getConnectionEndpointRelationship());
        this.archiveBuilder.addRelationshipDef(getConnectionConnectorTypeRelationship());
    }


    private EntityDef getConnectionEntity()
    {
        final String guid            = "114e9f8f-5ff3-4c32-bd37-a7eb42712253";
        final String name            = "Connection";
        final String description     = "A set of properties to identify and configure a connector instance.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "securedProperties";
        final String attribute3Description     = "Private properties accessible only to the connector.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "configurationProperties";
        final String attribute4Description     = "Specific configuration properties used to configure the behaviour of the connector.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "userId";
        final String attribute5Description     = "User identity that the connector should use.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "clearPassword";
        final String attribute6Description     = "Password for the userId in clear text.";
        final String attribute6DescriptionGUID = null;
        final String attribute7Name            = "encryptedPassword";
        final String attribute7Description     = "Encrypted password that the connector needs to decrypt before use.";
        final String attribute7DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringObjectTypeDefAttribute(attribute3Name,
                                                                    attribute3Description,
                                                                    attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringObjectTypeDefAttribute(attribute4Name,
                                                                    attribute4Description,
                                                                    attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute7Name,
                                                           attribute7Description,
                                                           attribute7DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getConnectorTypeEntity()
    {
        final String guid            = "954421eb-33a6-462d-a8ca-b5709a1bd0d4";
        final String name            = "ConnectorType";
        final String description     = "A set of properties describing a type of connector.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "connectorProviderClassName";
        final String attribute3Description     = "Name of the Java class that implements this connector type's open connector framework (OCF) connector provider.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "recognizedAdditionalProperties";
        final String attribute4Description     = "List of additional connection property names supported by the connector implementation.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "recognizedSecuredProperties";
        final String attribute5Description     = "List of secured connection property names supported by the connector implementation.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "recognizedConfigurationProperties";
        final String attribute6Description     = "List of secured connection property names supported by the connector implementation.";
        final String attribute6DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute4Name,
                                                                attribute4Description,
                                                                attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute5Name,
                                                                attribute5Description,
                                                                attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute6Name,
                                                                attribute6Description,
                                                                attribute6DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getConnectionEndpointRelationship()
    {
        final String guid            = "887a7132-d6bc-4b92-a483-e80b60c86fb2";
        final String name            = "ConnectionEndpoint";
        final String description     = "A link between a connection and the endpoint that the connector should use.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Endpoint";
        final String                     end1AttributeName            = "connectionEndpoint";
        final String                     end1AttributeDescription     = "Server endpoint that provides access to the asset.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Connection";
        final String                     end2AttributeName            = "connections";
        final String                     end2AttributeDescription     = "Connections to this endpoint.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getConnectionConnectorTypeRelationship()
    {
        final String guid            = "e542cfc1-0b4b-42b9-9921-f0a5a88aaf96";
        final String name            = "ConnectionConnectorType";
        final String description     = "A link between a connection and the connector type that should be used.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Connection";
        final String                     end1AttributeName            = "connections";
        final String                     end1AttributeDescription     = "Connections using this connector type.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ConnectorType";
        final String                     end2AttributeName            = "connectorType";
        final String                     end2AttributeDescription     = "Type of connector to use for the asset.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0205 Connection Links defines the relationship between the connection and an Asset, plus the nesting
     * of connections for information virtualization support.
     */
    private void add0205ConnectionLinkage()
    {
        this.archiveBuilder.addEntityDef(getVirtualConnectionEntity());

        this.archiveBuilder.addRelationshipDef(getEmbeddedConnectionRelationship());
        this.archiveBuilder.addRelationshipDef(getConnectionToAssetRelationship());
    }


    private EntityDef getVirtualConnectionEntity()
    {
        final String guid            = "82f9c664-e59d-484c-a8f3-17088c23a2f3";
        final String name            = "VirtualConnection";
        final String description     = "A connector for a virtual resource that needs to retrieve data from multiple places.";
        final String descriptionGUID = null;

        final String superTypeName = "Connection";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getEmbeddedConnectionRelationship()
    {
        final String guid            = "eb6dfdd2-8c6f-4f0d-a17d-f6ce4799f64f";
        final String name            = "EmbeddedConnection";
        final String description     = "A link between a virtual connection and one of the connections it depends on.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "VirtualConnection";
        final String                     end1AttributeName            = "supportingVirtualConnections";
        final String                     end1AttributeDescription     = "Virtual connections using this connection.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Connection";
        final String                     end2AttributeName            = "embeddedConnections";
        final String                     end2AttributeDescription     = "Connections embedded in this virtual connection.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute2Name            = "arguments";
        final String attribute2Description     = "Additional arguments needed by the virtual connector when using each connection.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringObjectTypeDefAttribute(attribute2Name,
                                                                    attribute2Description,
                                                                    attribute2DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getConnectionToAssetRelationship()
    {
        final String guid            = "e777d660-8dbe-453e-8b83-903771f054c0";
        final String name            = "ConnectionToAsset";
        final String description     = "Link between a connection and the description of the asset it can be used to access.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Connection";
        final String                     end1AttributeName            = "connections";
        final String                     end1AttributeDescription     = "Connections to this asset.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.ASSET.typeName;
        final String                     end2AttributeName            = "asset";
        final String                     end2AttributeDescription     = "Asset that can be accessed with this connection.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "assetSummary";
        final String attribute1Description     = "Description of the asset that is retrieved through this connection.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0210 DataStores describe physical data store assets.
     */
    private void add0210DataStores()
    {
        this.archiveBuilder.addEntityDef(getDataStoreEntity());

        this.archiveBuilder.addRelationshipDef(getDataContentForDataSetRelationship());

        this.archiveBuilder.addClassificationDef(getDataStoreEncodingClassification());
    }


    private EntityDef getDataStoreEntity()
    {
        final String guid            = OpenMetadataType.DATA_STORE.typeGUID;
        final String name            = OpenMetadataType.DATA_STORE.typeName;
        final String description     = OpenMetadataType.DATA_STORE.description;
        final String descriptionGUID = OpenMetadataType.DATA_STORE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.DATA_STORE.wikiURL;

        final String superTypeName = OpenMetadataType.ASSET.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.STORE_CREATE_TIME.name;
        final String attribute1Description     = OpenMetadataProperty.STORE_CREATE_TIME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.STORE_CREATE_TIME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.STORE_UPDATE_TIME.name;
        final String attribute2Description     = OpenMetadataProperty.STORE_UPDATE_TIME.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.STORE_UPDATE_TIME.descriptionGUID;

        property = archiveHelper.getDateTypeDefAttribute(attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getDataContentForDataSetRelationship()
    {
        final String guid            = OpenMetadataType.DATA_CONTENT_FOR_DATA_SET_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.DATA_CONTENT_FOR_DATA_SET_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.DATA_CONTENT_FOR_DATA_SET_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.DATA_CONTENT_FOR_DATA_SET_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.DATA_CONTENT_FOR_DATA_SET_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.BOTH;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.ASSET.typeName;
        final String                     end1AttributeName            = "dataContent";
        final String                     end1AttributeDescription     = "Assets supporting a data set.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.DATA_SET.typeName;
        final String                     end2AttributeName            = "supportedDataSets";
        final String                     end2AttributeDescription     = "Data sets that use this asset.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private ClassificationDef getDataStoreEncodingClassification()
    {
        final String guid            = OpenMetadataType.DATA_STORE_ENCODING_CLASSIFICATION.typeGUID;
        final String name            = OpenMetadataType.DATA_STORE_ENCODING_CLASSIFICATION.typeName;
        final String description     = OpenMetadataType.DATA_STORE_ENCODING_CLASSIFICATION.description;
        final String descriptionGUID = OpenMetadataType.DATA_STORE_ENCODING_CLASSIFICATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.DATA_STORE_ENCODING_CLASSIFICATION.wikiURL;

        final String linkedToEntity = OpenMetadataType.DATA_STORE.typeName;

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 descriptionWiki,
                                                                                 this.archiveBuilder.getEntityDef(linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.ENCODING.name;
        final String attribute1Description     = OpenMetadataProperty.ENCODING.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.ENCODING.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.ENCODING_LANGUAGE.name;
        final String attribute2Description     = OpenMetadataProperty.ENCODING_LANGUAGE.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.ENCODING_LANGUAGE.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.ENCODING_DESCRIPTION.name;
        final String attribute3Description     = OpenMetadataProperty.ENCODING_DESCRIPTION.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.ENCODING_DESCRIPTION.descriptionGUID;
        final String attribute4Name            = OpenMetadataProperty.ENCODING_PROPERTIES.name;
        final String attribute4Description     = OpenMetadataProperty.ENCODING_PROPERTIES.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.ENCODING_PROPERTIES.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute4Name,
                                                                    attribute4Description,
                                                                    attribute4DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0212 Deployed APIs defines an API that has been deployed to IT Infrastructure
     */
    private void add0212DeployedAPIs()
    {
        this.archiveBuilder.addEntityDef(getDeployedAPIEntity());

        this.archiveBuilder.addRelationshipDef(getAPIEndpointRelationship());

        this.archiveBuilder.addClassificationDef(getRequestResponseInterfaceClassification());
        this.archiveBuilder.addClassificationDef(getListenerInterfaceClassification());
        this.archiveBuilder.addClassificationDef(getPublisherInterfaceClassification());
    }


    private EntityDef getDeployedAPIEntity()
    {
        final String guid            = "7dbb3e63-138f-49f1-97b4-66313871fc14";
        final String name            = "DeployedAPI";
        final String description     = "A callable interface running at an endpoint.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.ASSET.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getAPIEndpointRelationship()
    {
        final String guid            = "de5b9501-3ad4-4803-a8b2-e311c72a4336";
        final String name            = "APIEndpoint";
        final String description     = "The endpoint for a deployed API.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "DeployedAPI";
        final String                     end1AttributeName            = "supportedAPIs";
        final String                     end1AttributeDescription     = "APIs that can be called from this endpoint.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Endpoint";
        final String                     end2AttributeName            = "accessEndpoints";
        final String                     end2AttributeDescription     = "Endpoints used to call this API.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private ClassificationDef getRequestResponseInterfaceClassification()
    {
        final String guid            = "14a29330-e830-4343-a41e-d57e2cec82f8";
        final String name            = "RequestResponseInterface";
        final String description     = "Identifies an API that supports a request response interaction style.";
        final String descriptionGUID = null;

        final String linkedToEntity = "DeployedAPI";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getListenerInterfaceClassification()
    {
        final String guid            = "4099d2ed-2a5e-4c44-8443-9de4e378a4ba";
        final String name            = "ListenerInterface";
        final String description     = "Identifies an API that listens for incoming events and processes them.";
        final String descriptionGUID = null;

        final String linkedToEntity = "DeployedAPI";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getPublisherInterfaceClassification()
    {
        final String guid            = "4fdedcd5-b186-4bee-887a-02fa29a10750";
        final String name            = "PublisherInterface";
        final String description     = "Identifies an API that sends out events to other listening components.";
        final String descriptionGUID = null;

        final String linkedToEntity = "DeployedAPI";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0215 Software Components defines a generic Asset for a software component.
     */
    private void add0215SoftwareComponents()
    {
        this.archiveBuilder.addEntityDef(getDeployedSoftwareComponentEntity());
    }


    private EntityDef getDeployedSoftwareComponentEntity()
    {
        final String guid            = "486af62c-dcfd-4859-ab24-eab2e380ecfd";
        final String name            = "DeployedSoftwareComponent";
        final String description     = "A packaged and deployed software component supporting a well-defined function.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.PROCESS.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "implementationLanguage";
        final String attribute1Description     = "Name of the language used to implement this component.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0220 Files and Folders provides the definitions for describing filesystems and their content
     */
    private void add0220FilesAndFolders()
    {
        this.archiveBuilder.addEntityDef(getFileFolderEntity());
        this.archiveBuilder.addEntityDef(getDataFolderEntity());
        this.archiveBuilder.addEntityDef(getDataFileEntity());
        this.archiveBuilder.addEntityDef(getCSVFileEntity());
        this.archiveBuilder.addEntityDef(getAvroFileEntity());
        this.archiveBuilder.addEntityDef(getJSONFileEntity());

        this.archiveBuilder.addRelationshipDef(getFolderHierarchyRelationship());
        this.archiveBuilder.addRelationshipDef(getNestedFileRelationship());
        this.archiveBuilder.addRelationshipDef(getLinkedFileRelationship());

        this.archiveBuilder.addClassificationDef(getFileSystemClassification());
    }


    private EntityDef getFileFolderEntity()
    {
        final String guid            = OpenMetadataType.FILE_FOLDER.typeGUID;
        final String name            = OpenMetadataType.FILE_FOLDER.typeName;
        final String description     = OpenMetadataType.FILE_FOLDER.description;
        final String descriptionGUID = OpenMetadataType.FILE_FOLDER.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.FILE_FOLDER.wikiURL;

        final String superTypeName = OpenMetadataType.DATA_STORE.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    private EntityDef getDataFolderEntity()
    {
        final String guid            = OpenMetadataType.DATA_FOLDER.typeGUID;
        final String name            = OpenMetadataType.DATA_FOLDER.typeName;
        final String description     = OpenMetadataType.DATA_FOLDER.description;
        final String descriptionGUID = OpenMetadataType.DATA_FOLDER.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.DATA_FOLDER.wikiURL;

        final String superTypeName = OpenMetadataType.FILE_FOLDER.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    private EntityDef getDataFileEntity()
    {
        final String guid            = OpenMetadataType.DATA_FILE.typeGUID;
        final String name            = OpenMetadataType.DATA_FILE.typeName;
        final String description     = OpenMetadataType.DATA_FILE.description;
        final String descriptionGUID = OpenMetadataType.DATA_FILE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.DATA_FILE.wikiURL;

        final String superTypeName = OpenMetadataType.DATA_STORE.typeName;

        EntityDef entityDef =  archiveHelper.getDefaultEntityDef(guid,
                                                                 name,
                                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                                 description,
                                                                 descriptionGUID,
                                                                 descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.FILE_TYPE.name;
        final String attribute1Description     = OpenMetadataProperty.FILE_TYPE.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.FILE_TYPE.descriptionGUID;


        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getCSVFileEntity()
    {
        final String guid            = "2ccb2117-9cee-47ca-8150-9b3a543adcec";
        final String name            = "CSVFile";
        final String description     = "A description of a comma separated value (CSV) file";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_FILE.typeName;

        EntityDef entityDef =  archiveHelper.getDefaultEntityDef(guid,
                                                                 name,
                                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                                 description,
                                                                 descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "delimiterCharacter";
        final String attribute1Description     = "Character used between each column.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "quoteCharacter";
        final String attribute2Description     = "The character used to group the content of the column that contains one or more delimiter characters.";
        final String attribute2DescriptionGUID = null;


        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                            attribute1Description,
                                                            attribute1DescriptionGUID);
        properties.add(property);

        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getAvroFileEntity()
    {
        final String guid            = OpenMetadataType.AVRO_FILE.typeGUID;
        final String name            = OpenMetadataType.AVRO_FILE.typeName;
        final String description     = OpenMetadataType.AVRO_FILE.description;
        final String descriptionGUID = OpenMetadataType.AVRO_FILE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.AVRO_FILE.wikiURL;

        final String superTypeName   = OpenMetadataType.DATA_FILE.typeName;

        return  archiveHelper.getDefaultEntityDef(guid,
                                                  name,
                                                  this.archiveBuilder.getEntityDef(superTypeName),
                                                  description,
                                                  descriptionGUID,
                                                  descriptionWiki);
    }


    private EntityDef getJSONFileEntity()
    {
        final String guid            = OpenMetadataType.JSON_FILE.typeGUID;
        final String name            = OpenMetadataType.JSON_FILE.typeName;
        final String description     = OpenMetadataType.JSON_FILE.description;
        final String descriptionGUID = OpenMetadataType.JSON_FILE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.JSON_FILE.wikiURL;

        final String superTypeName   = OpenMetadataType.DATA_FILE.typeName;

        return  archiveHelper.getDefaultEntityDef(guid,
                                                  name,
                                                  this.archiveBuilder.getEntityDef(superTypeName),
                                                  description,
                                                  descriptionGUID,
                                                  descriptionWiki);
    }


    private RelationshipDef getFolderHierarchyRelationship()
    {
        final String guid            = "48ac9028-45dd-495d-b3e1-622685b54a01";
        final String name            = "FolderHierarchy";
        final String description     = "A nested relationship between two file folders.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "FileFolder";
        final String                     end1AttributeName            = "parentFolder";
        final String                     end1AttributeDescription     = "Parent folder.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "FileFolder";
        final String                     end2AttributeName            = "nestedFolder";
        final String                     end2AttributeDescription     = "Folders embedded in this folder.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getNestedFileRelationship()
    {
        final String guid            = "4cb88900-1446-4eb6-acea-29cd9da45e63";
        final String name            = "NestedFile";
        final String description     = "The link between a data file and its containing folder.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "FileFolder";
        final String                     end1AttributeName            = "homeFolder";
        final String                     end1AttributeDescription     = "Identifies the containing folder of this datafile.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.DATA_FILE.typeName;
        final String                     end2AttributeName            = "nestedFiles";
        final String                     end2AttributeDescription     = "Files stored in this folder.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getLinkedFileRelationship()
    {
        final String guid            = "970a3405-fde1-4039-8249-9aa5f56d5151";
        final String name            = "LinkedFile";
        final String description     = "A data file that is linked to a file folder (rather than stored in it).";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "FileFolder";
        final String                     end1AttributeName            = "linkedFolders";
        final String                     end1AttributeDescription     = "Folders that this file is linked to.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.DATA_FILE.typeName;
        final String                     end2AttributeName            = "linkedFiles";
        final String                     end2AttributeDescription     = "Files linked to the folder.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private ClassificationDef getFileSystemClassification()
    {
        final String guid            = "cab5ba1d-cfd3-4fca-857d-c07711fc4157";
        final String name            = "FileSystem";
        final String description     = "A capability that supports a store of files organized into a hierarchy of file folders.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.typeName;

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "format";
        final String attribute1Description     = "Format of the file system.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "encryption";
        final String attribute2Description     = "Level of encryption used on the filesystem (if any).";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0221 Document Stores define both simple document stores and content management systems
     */
    private void add0221DocumentStores()
    {
        this.archiveBuilder.addEntityDef(getMediaFileEntity());
        this.archiveBuilder.addEntityDef(getMediaCollectionEntity());
        this.archiveBuilder.addEntityDef(getDocumentEntity());
        this.archiveBuilder.addEntityDef(getDocumentStoreEntity());

        this.archiveBuilder.addRelationshipDef(getGroupedMediaRelationship());
        this.archiveBuilder.addRelationshipDef(getLinkedMediaRelationship());

        this.archiveBuilder.addClassificationDef(getContentManagerClassification());
    }


    private EntityDef getMediaFileEntity()
    {
        EntityDef entityDef = archiveHelper.getDefaultEntityDef(OpenMetadataType.MEDIA_FILE.typeGUID,
                                                                OpenMetadataType.MEDIA_FILE.typeName,
                                                                this.archiveBuilder.getEntityDef(OpenMetadataType.DATA_FILE.typeName),
                                                                OpenMetadataType.MEDIA_FILE.description,
                                                                OpenMetadataType.MEDIA_FILE.descriptionGUID,
                                                                OpenMetadataType.MEDIA_FILE.wikiURL);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.EMBEDDED_METADATA.name;
        final String attribute1Description     = OpenMetadataProperty.EMBEDDED_METADATA.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.EMBEDDED_METADATA.descriptionGUID;


        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute1Name,
                                                                    attribute1Description,
                                                                    attribute1DescriptionGUID);
        properties.add(property);


        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getMediaCollectionEntity()
    {
        final String guid            = "0075d603-1627-41c5-8cae-f5458d1247fe";
        final String name            = "MediaCollection";
        final String description     = "A group of related media files.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_SET.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getDocumentEntity()
    {
        return archiveHelper.getDefaultEntityDef(OpenMetadataType.DOCUMENT.typeGUID,
                                                 OpenMetadataType.DOCUMENT.typeName,
                                                 this.archiveBuilder.getEntityDef(OpenMetadataType.MEDIA_FILE.typeName),
                                                 OpenMetadataType.DOCUMENT.description,
                                                 OpenMetadataType.DOCUMENT.descriptionGUID,
                                                 OpenMetadataType.DOCUMENT.wikiURL);
    }


    private EntityDef getDocumentStoreEntity()
    {
        final String guid            = "37156790-feac-4e1a-a42e-88858ae6f8e1";
        final String name            = "DocumentStore";
        final String description     = "Identifies a data store as one that contains documents.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_STORE.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getGroupedMediaRelationship()
    {
        final String guid            = "7d881574-461d-475c-ab44-077451528cb8";
        final String name            = "GroupedMedia";
        final String description     = "Links a media file into a data set.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "MediaCollection";
        final String                     end1AttributeName            = "dataSetMembership";
        final String                     end1AttributeDescription     = "Identifies the data sets this media file belongs to.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.MEDIA_FILE.typeName;
        final String                     end2AttributeName            = "dataSetMembers";
        final String                     end2AttributeDescription     = "Media files that make up this media collection.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getLinkedMediaRelationship()
    {
        final String guid            = "cee3a190-fc8d-4e53-908a-f1b9689581e0";
        final String name            = "LinkedMedia";
        final String description     = "Links a media file to another media file and describes relationship.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.MEDIA_FILE.typeName;
        final String                     end1AttributeName            = "linkedMediaFiles";
        final String                     end1AttributeDescription     = "Link to related media files.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.MEDIA_FILE.typeName;
        final String                     end2AttributeName            = "linkedMediaFiles";
        final String                     end2AttributeDescription     = "Link to related media files.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private ClassificationDef getContentManagerClassification()
    {
        final String guid            = "fa4df7b5-cb6d-475c-889e-8f3b7ca564d3";
        final String name            = "ContentManager";
        final String description     = "Identifies a server as a manager of controlled documents and related media.";
        final String descriptionGUID = null;

        final String linkedToEntity = "SoftwareServer";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0222GraphStores()
    {
        this.archiveBuilder.addEntityDef(getGraphStoreEntity());
    }


    private EntityDef getGraphStoreEntity()
    {
        final String guid            = "86de3633-eec8-4bf9-aad1-e92df1ca2024";
        final String name            = "GraphStore";
        final String description     = "Identifies a data store as one that contains one or more graphs.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_STORE.typeName;

        EntityDef entityDef =  archiveHelper.getDefaultEntityDef(guid,
                                                                 name,
                                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                                 description,
                                                                 descriptionGUID);
        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "type";
        final String attribute1Description     = "Type of graph store.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0223 Events and Logs describes events, log files and event management.
     */
    private void add0223EventsAndLogs()
    {
        this.archiveBuilder.addEntityDef(getSubscriberListEntity());
        this.archiveBuilder.addEntityDef(getTopicEntity());
        this.archiveBuilder.addEntityDef(getLogFileEntity());


        this.archiveBuilder.addRelationshipDef(getTopicSubscribersRelationship());

        this.archiveBuilder.addClassificationDef(getNotificationManagerClassification());
    }


    private EntityDef getSubscriberListEntity()
    {
        final String guid            = "69751093-35f9-42b1-944b-ba6251ff513d";
        final String name            = "SubscriberList";
        final String description     = "A data set containing a list of endpoints registered to receive events from a topic.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_SET.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getTopicEntity()
    {
        final String guid            = OpenMetadataType.TOPIC.typeGUID;
        final String name            = OpenMetadataType.TOPIC.typeName;
        final String description     = OpenMetadataType.TOPIC.description;
        final String descriptionGUID = OpenMetadataType.TOPIC.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.TOPIC.wikiURL;

        final String superTypeName = OpenMetadataType.DATA_SET.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.TOPIC_TYPE.name;
        final String attribute1Description     = OpenMetadataProperty.TOPIC_TYPE.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.TOPIC_TYPE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getLogFileEntity()
    {
        final String guid            = OpenMetadataType.LOG_FILE.typeGUID;
        final String name            = OpenMetadataType.LOG_FILE.typeName;
        final String description     = OpenMetadataType.LOG_FILE.description;
        final String descriptionGUID = OpenMetadataType.LOG_FILE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.LOG_FILE.wikiURL;

        final String superTypeName = OpenMetadataType.DATA_FILE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "type";
        final String attribute1Description     = "Type of log file.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getTopicSubscribersRelationship()
    {
        final String guid            = "bc91a28c-afb9-41a7-8eb2-fc8b5271fe9e";
        final String name            = "TopicSubscribers";
        final String description     = "Links the list of subscribers to a topic.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "SubscriberList";
        final String                     end1AttributeName            = "subscribers";
        final String                     end1AttributeDescription     = "The endpoints subscribed to this topic.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Topic";
        final String                     end2AttributeName            = "topics";
        final String                     end2AttributeDescription     = "The topics used by this subscriber list.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private ClassificationDef getNotificationManagerClassification()
    {
        final String guid            = "3e7502a7-396a-4737-a106-378c9c94c105";
        final String name            = "NotificationManager";
        final String description     = "Identifies a server capability that is distributing events from a topic to its subscriber list.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.typeName;

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0224 Databases describe database servers
     */
    private void add0224Databases()
    {
        this.archiveBuilder.addEntityDef(getDeployedDatabaseSchemaEntity());
        this.archiveBuilder.addEntityDef(getDatabaseEntity());

        this.archiveBuilder.addClassificationDef(getDatabaseServerClassification());
    }


    private EntityDef getDeployedDatabaseSchemaEntity()
    {
        final String guid            = "eab811ec-556a-45f1-9091-bc7ac8face0f";
        final String name            = "DeployedDatabaseSchema";
        final String description     = "A collection of database tables and views running in a database server.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_SET.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getDatabaseEntity()
    {
        final String guid            = "0921c83f-b2db-4086-a52c-0d10e52ca078";
        final String name            = "Database";
        final String description     = "A data store containing relational data.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_STORE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "type";
        final String attribute1Description     = "Type of database.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "version";
        final String attribute2Description     = "Version of the database.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "instance";
        final String attribute3Description     = "Name of the database instance.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "importedFrom";
        final String attribute4Description     = "Name of the connector where database is imported from.";
        final String attribute4DescriptionGUID = null;        

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);        

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private ClassificationDef getDatabaseServerClassification()
    {
        final String guid            = "6bb58cc9-ed9e-4f75-b2f2-6d308554eb52";
        final String name            = "DatabaseServer";
        final String description     = "Identifies a server as one that manages one or more databases.";
        final String descriptionGUID = null;

        final String linkedToEntity = "SoftwareServer";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "type";
        final String attribute1Description     = "Type of database server.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "version";
        final String attribute2Description     = "Version of the database server software.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute3Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0225MetadataRepositories()
    {
        this.archiveBuilder.addEntityDef(getEnterpriseAccessLayerEntity());
        this.archiveBuilder.addEntityDef(getCohortMemberEntity());
        this.archiveBuilder.addEntityDef(getMetadataRepositoryCohortEntity());
        this.archiveBuilder.addEntityDef(getMetadataCollectionEntity());
        this.archiveBuilder.addEntityDef(getMetadataRepositoryEntity());
        this.archiveBuilder.addEntityDef(getCohortRegistryStoreEntity());

        this.archiveBuilder.addRelationshipDef(getMetadataCohortPeerRelationship());
        this.archiveBuilder.addRelationshipDef(getCohortMemberMetadataCollectionRelationship());

        this.archiveBuilder.addClassificationDef(getMetadataServerClassification());
        this.archiveBuilder.addClassificationDef(getRepositoryProxyClassification());
    }


    private EntityDef getEnterpriseAccessLayerEntity()
    {
        final String guid            = "39444bf9-638e-4124-a5f9-1b8f3e1b008b";
        final String name            = "EnterpriseAccessLayer";
        final String description     = "Repository services for the Open Metadata Access Services (OMAS) supporting federated queries and aggregated events from the connected cohorts.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "topicRoot";
        final String attribute1Description     = "Root of topic names used by the Open Metadata access Services (OMASs).";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "metadataCollectionId";
        final String attribute2Description     = "Unique identifier for the metadata collection accessed through this enterprise access layer.";
        final String attribute2DescriptionGUID = null;


        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getCohortMemberEntity()
    {
        final String guid            = "42063797-a78a-4720-9353-52026c75f667";
        final String name            = "CohortMember";
        final String description     = "A capability enabling a server to access an open metadata repository cohort.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getMetadataRepositoryCohortEntity()
    {
        final String guid            = "43e7dca2-c7b4-4cdf-a1ea-c9d4f7093893";
        final String name            = "MetadataRepositoryCohort";
        final String description     = "A group of collaborating open metadata repositories.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = "topic";
        final String attribute2Description     = "Name of the topic used to exchange registration, type definitions and metadata instances between the members of the open metadata repository cohort.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getMetadataCollectionEntity()
    {
        final String guid            = "ea3b15af-ed0e-44f7-91e4-bdb299dd4976";
        final String name            = "MetadataCollection";
        final String description     = "A data set containing metadata.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_SET.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "metadataCollectionId";
        final String attribute1Description     = "Unique identifier for the metadata collection managed in the local repository.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }




    private EntityDef getMetadataRepositoryEntity()
    {
        final String guid            = "c40397bd-eab0-4b2e-bffb-e7fa0f93a5a9";
        final String name            = "MetadataRepository";
        final String description     = "A data store containing metadata.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_STORE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "type";
        final String attribute1Description     = "Type of metadata repository.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getCohortRegistryStoreEntity()
    {
        final String guid            = "2bfdcd0d-68bb-42c3-ae75-e9fb6c3dff70";
        final String name            = "CohortRegistryStore";
        final String description     = "A data store containing cohort membership registration details.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_STORE.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getMetadataCohortPeerRelationship()
    {
        final String guid            = "954cdba1-3d69-4db1-bf0e-d59fd2c25a27";
        final String name            = "MetadataCohortPeer";
        final String description     = "A metadata repository's registration with an open metadata cohort.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "MetadataRepositoryCohort";
        final String                     end1AttributeName            = "registeredWithCohorts";
        final String                     end1AttributeDescription     = "Identifies which cohorts this cohort member is registered with.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "CohortMember";
        final String                     end2AttributeName            = "cohortMembership";
        final String                     end2AttributeDescription     = "Members of this cohort.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "registrationDate";
        final String attribute1Description     = "Date first registered with the cohort.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getDateTypeDefAttribute(attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getCohortMemberMetadataCollectionRelationship()
    {
        final String guid            = "8b9dd3ea-057b-4709-9b42-f16098523907";
        final String name            = "CohortMemberMetadataCollection";
        final String description     = "The local metadata collection associated with a cohort peer.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "CohortMember";
        final String                     end1AttributeName            = "cohortMember";
        final String                     end1AttributeDescription     = "Cohort registry representing this metadata collection on the metadata highway.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "MetadataCollection";
        final String                     end2AttributeName            = "localMetadataCollection";
        final String                     end2AttributeDescription     = "Metadata to exchange with the cohorts.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private ClassificationDef getMetadataServerClassification()
    {
        final String guid            = "74a256ad-4022-4518-a446-c65fe082d4d3";
        final String name            = "MetadataServer";
        final String description     = "A server hosting a metadata collection.";
        final String descriptionGUID = null;

        final String linkedToEntity = "SoftwareServer";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "format";
        final String attribute1Description     = "format of supported metadata.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "type";
        final String attribute2Description     = "Type of metadata server.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private ClassificationDef getRepositoryProxyClassification()
    {
        final String guid            = "ae81c35e-7078-46f0-9b2c-afc99accf3ec";
        final String name            = "RepositoryProxy";
        final String description     = "A server acting as an open metadata adapter for a metadata repository.";
        final String descriptionGUID = null;

        final String linkedToEntity = "SoftwareServer";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "type";
        final String attribute1Description     = "Type of repository proxy.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0227Keystores()
    {
        this.archiveBuilder.addEntityDef(getKeystoreFileEntity());
        this.archiveBuilder.addEntityDef(geSecretsCollectionEntity());
    }


    private EntityDef getKeystoreFileEntity()
    {
        return archiveHelper.getDefaultEntityDef(OpenMetadataType.KEYSTORE_FILE.typeGUID,
                                                 OpenMetadataType.KEYSTORE_FILE.typeName,
                                                 this.archiveBuilder.getEntityDef(OpenMetadataType.DATA_FILE.typeName),
                                                 OpenMetadataType.KEYSTORE_FILE.description,
                                                 OpenMetadataType.KEYSTORE_FILE.descriptionGUID,
                                                 OpenMetadataType.KEYSTORE_FILE.wikiURL);
    }


    private EntityDef geSecretsCollectionEntity()
    {
        return archiveHelper.getDefaultEntityDef(OpenMetadataType.SECRETS_COLLECTION.typeGUID,
                                                 OpenMetadataType.SECRETS_COLLECTION.typeName,
                                                 this.archiveBuilder.getEntityDef(OpenMetadataType.DATA_SET.typeName),
                                                 OpenMetadataType.SECRETS_COLLECTION.description,
                                                 OpenMetadataType.SECRETS_COLLECTION.descriptionGUID,
                                                 OpenMetadataType.SECRETS_COLLECTION.wikiURL);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0230CodeTables()
    {
        this.archiveBuilder.addEntityDef(getReferenceCodeTableEntity());
        this.archiveBuilder.addEntityDef(getReferenceCodeMappingTableEntity());

    }


    private EntityDef getReferenceCodeTableEntity()
    {
        final String guid            = "201f48c5-4e4b-41dc-9c5f-0bc9742190cf";
        final String name            = "ReferenceCodeTable";
        final String description     = "A data set containing code values and their translations.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_SET.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getReferenceCodeMappingTableEntity()
    {
        final String guid            = "9c6ec0c6-0b26-4414-bffe-089144323213";
        final String name            = "ReferenceCodeMappingTable";
        final String description     = "A data set containing mappings between code values from different data sets.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_SET.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0235InformationView()
    {
        this.archiveBuilder.addEntityDef(getInformationViewEntity());

    }


    private EntityDef getInformationViewEntity()
    {
        final String guid            = "68d7b905-6438-43be-88cf-5de027b4aaaf";
        final String name            = "InformationView";
        final String description     = "A data set containing selected data items from one or more data stores or data sets.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_SET.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "id";
        final String attribute1Description     = "Id of view.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "comment";
        final String attribute2Description     = "Comment";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "nativeClass";
        final String attribute3Description     = "Native class used by the client to represent this entity.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "createdTime";
        final String attribute4Description     = "Information View create time.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "lastModifiedTime";
        final String attribute5Description     = "Information View last modified time.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "lastModifier";
        final String attribute6Description     = "Information View last modifier.";
        final String attribute6DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                attribute1Description,
                attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                attribute2Description,
                attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                attribute3Description,
                attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute4Name,
                attribute4Description,
                attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute5Name,
                attribute5Description,
                attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                attribute6Description,
                attribute6DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0239Reports()
    {
        this.archiveBuilder.addEntityDef(getFormEntity());
        this.archiveBuilder.addEntityDef(getDeployedReportEntity());
    }


    private EntityDef getFormEntity()
    {
        final String guid            = "8078e3d1-0c63-4ace-aafa-68498b39ccd6";
        final String name            = "Form";
        final String description     = "A collection of data items used to request activity.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_SET.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getDeployedReportEntity()
    {
        final String guid            = "e9077f4f-955b-4d7b-b1f7-12ee769ff0c3";
        final String name            = "DeployedReport";
        final String description     = "A collection if data items that describe a situation.  This is an instance of a report.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.DATA_SET.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "id";
        final String attribute1Description     = "Id of report.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "author";
        final String attribute2Description     = "Author of the report.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "url";
        final String attribute3Description     = "url of the report.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "createdTime";
        final String attribute4Description     = "Report create time.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "lastModifiedTime";
        final String attribute5Description     = "Report last modified time.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "lastModifier";
        final String attribute6Description     = "Report last modifier.";
        final String attribute6DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute4Name,
                                                         attribute4Description,
                                                         attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute5Name,
                                                         attribute5Description,
                                                         attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private void add0290Ports()
    {
        this.archiveBuilder.addEnumDef(getPortTypeEnum());

        this.archiveBuilder.addEntityDef(getPortEntity());
        this.archiveBuilder.addEntityDef(getPortAliasEntity());
        this.archiveBuilder.addEntityDef(getPortImplementationEntity());


        this.archiveBuilder.addRelationshipDef(getProcessPortRelationship());
        this.archiveBuilder.addRelationshipDef(getPortDelegationRelationship());
    }

    /**
     * The PortType enum describes the type of a port
     *
     * @return PortType EnumDef
     */
    private EnumDef getPortTypeEnum()
    {
        final String guid            = "b57Fbce7-42ac-71D1-D6a6-9f62Cb7C6dc3";
        final String name            = "PortType";
        final String description     = "Descriptor for a port that indicates its type.";
        final String descriptionGUID = null;

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID);

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        final int    element1Ordinal         = 0;
        final String element1Value           = "INPUT_PORT";
        final String element1Description     = "Data is passed into the process.";
        final String element1DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element1Ordinal,
                                                    element1Value,
                                                    element1Description,
                                                    element1DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element2Ordinal         = 1;
        final String element2Value           = "OUTPUT_PORT";
        final String element2Description     = "Data is produced by the process.";
        final String element2DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element2Ordinal,
                                                    element2Value,
                                                    element2Description,
                                                    element2DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element3Ordinal         = 2;
        final String element3Value           = "INOUT_PORT";
        final String element3Description     = "A request-response interface is provided by the process.";
        final String element3DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element3Ordinal,
                                                    element3Value,
                                                    element3Description,
                                                    element3DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element4Ordinal         = 3;
        final String element4Value           = "OUTIN_PORT";
        final String element4Description     = "A request-response call is made by the process.";
        final String element4DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element4Ordinal,
                                                    element4Value,
                                                    element4Description,
                                                    element4DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element5Ordinal         = 99;
        final String element5Value           = "OTHER";
        final String element5Description     = "None of the above.";
        final String element5DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element5Ordinal,
                                                    element5Value,
                                                    element5Description,
                                                    element5DescriptionGUID);
        elementDefs.add(elementDef);

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }

    /**
     * The Port entity describes the input or output of a process
     *
     * @return Port EntityDef
     */
    private EntityDef getPortEntity()
    {
        /*
         * Build the Entity
         */
        final String guid            = "e3d9FD9F-d5eD-2aed-CC98-0bc21aB6f71C";
        final String name            = "Port";
        final String description     = "An interface where data flows in and/or out of the process.";
        final String descriptionGUID = null;
        final String superTypeName   = OpenMetadataType.REFERENCEABLE.typeName;


        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);


        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;

        final String attribute2Name            = "portType";
        final String attribute2Description     = "Type of port";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        property = archiveHelper.getEnumTypeDefAttribute("PortType",
                                                         attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }

    /**
     * The PortAlias describes the input/output of a higher-level process
     *
     * @return PortAlias EntityDef
     */
    private EntityDef getPortAliasEntity()
    {
        /*
         * Build the Entity
         */
        final String guid            = "DFa5aEb1-bAb4-c25B-bDBD-B95Ce6fAB7F5";
        final String name            = "PortAlias";
        final String description     = "Entity that describes the port for a composition process.";
        final String descriptionGUID = null;
        final String superTypeName   = "Port";


        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);

    }

    /**
     * The PortImplementation describes the input/output of a lowest-level process
     *
     * @return PortImplementation EntityDef
     */
    private EntityDef getPortImplementationEntity()
    {
        /*
         * Build the Entity
         */
        final String guid            = "ADbbdF06-a6A3-4D5F-7fA3-DB4Cb0eDeC0E";
        final String name            = "PortImplementation";
        final String description     = "Entity that describes a port with a concrete implementation.";
        final String descriptionGUID = null;
        final String superTypeName   = "Port";


        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }

    /**
     * The PortDelegation relationship describes the relationship between a more granular and a more abstract port
     * @return PortDelegation RelationshipDef
     */
    private RelationshipDef getPortDelegationRelationship()
    {
        /*
         * Build the relationship
         */
        final String guid            = "98bB8BA1-dc6A-eb9D-32Cf-F837bEbCbb8E";
        final String name            = "PortDelegation";
        final String description     = "A relationship between a more granular and a more abstract port";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Port";
        final String                     end1AttributeName            = "delegatingFrom";
        final String                     end1AttributeDescription     = "Higher level Port";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);

        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Port";
        final String                     end2AttributeName            = "delegatingTo";
        final String                     end2AttributeDescription     = "Lower level port";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);


        return relationshipDef;
    }

    /**
     * The ProcessPort relationship describes the link between a port and the process used by the port.
     * @return ProcessPort RelationshipDef
     */
    private RelationshipDef getProcessPortRelationship()
    {
        /*
         * Build the relationship
         */
        final String guid            = "fB4E00CF-37e4-88CE-4a94-233BAdB84DA2";
        final String name            = "ProcessPort";
        final String description     = "A link between a process and one of its ports.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.PROCESS.typeName;
        final String                     end1AttributeName            = "owningProcess";
        final String                     end1AttributeDescription     = "Process linked to the port";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Port";
        final String                     end2AttributeName            = "ports";
        final String                     end2AttributeDescription     = "Port to the process";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);


        return relationshipDef;
    }

    /*
     * ========================================
     * AREA 3: glossary
     */

    /**
     * Area 3 covers semantic definitions.
     */
    private void addArea3Types()
    {
        this.add0310Glossary();
        this.add0320CategoryHierarchy();
        this.add0330Terms();
        this.add0340Dictionary();
        this.add0350RelatedTerms();
        this.add0360Contexts();
        this.add0370SemanticAssignment();
        this.add0380SpineObjects();
        this.add0385ControlledGlossaryDevelopment();
        this.add0390GlossaryProject();
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0310 Glossary covers the top-level glossary object that organizes glossary terms.
     */
    private void add0310Glossary()
    {
        this.archiveBuilder.addEntityDef(getGlossaryEntity());
        this.archiveBuilder.addEntityDef(getExternalGlossaryLinkEntity());

        this.archiveBuilder.addRelationshipDef(getExternalSourcedGlossaryRelationship());

        this.archiveBuilder.addClassificationDef(getTaxonomyClassification());
        this.archiveBuilder.addClassificationDef(getCanonicalVocabularyClassification());
    }


    private EntityDef getGlossaryEntity()
    {
        final String guid            = "36f66863-9726-4b41-97ee-714fd0dc6fe4";
        final String name            = "Glossary";
        final String description     = "A collection of related glossary terms.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "language";
        final String attribute3Description     = "Natural language used in the glossary.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.USAGE.name;
        final String attribute4Description     = OpenMetadataProperty.USAGE.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.USAGE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getExternalGlossaryLinkEntity()
    {
        final String guid            = "183d2935-a950-4d74-b246-eac3664b5a9d";
        final String name            = "ExternalGlossaryLink";
        final String description     = "The location of a glossary stored outside of the open metadata ecosystem.";
        final String descriptionGUID = null;

        final String superTypeName = "ExternalReference";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getExternalSourcedGlossaryRelationship()
    {
        final String guid            = "7786a39c-436b-4538-acc7-d595b5856add";
        final String name            = "ExternallySourcedGlossary";
        final String description     = "Link between an open metadata glossary and a related glossary stored outside of the open metadata ecosystem.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Glossary";
        final String                     end1AttributeName            = "localGlossary";
        final String                     end1AttributeDescription     = "Local glossary that relates to this external glossary.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ExternalGlossaryLink";
        final String                     end2AttributeName            = "externalGlossaryLink";
        final String                     end2AttributeDescription     = "Link to a related external glossary.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private ClassificationDef getTaxonomyClassification()
    {
        final String guid            = "37116c51-e6c9-4c37-942e-35d48c8c69a0";
        final String name            = "Taxonomy";
        final String description     = "Identifies a glossary that includes a taxonomy.";
        final String descriptionGUID = null;

        final String linkedToEntity = "Glossary";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "organizingPrinciple";
        final String attribute1Description     = "Characteristics that influence the organization of the taxonomy.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private ClassificationDef getCanonicalVocabularyClassification()
    {
        final String guid            = "33ad3da2-0910-47be-83f1-daee018a4c05";
        final String name            = "CanonicalVocabulary";
        final String description     = "Identifies a glossary that contains unique terms.";
        final String descriptionGUID = null;

        final String linkedToEntity = "Glossary";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "scope";
        final String attribute1Description     = "Scope of influence for this canonical glossary.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0320 Category Hierarchy adds categories to the glossary that allow terms to be defined in taxonomies
     * or hierarchies of folders.
     */
    private void add0320CategoryHierarchy()
    {
        this.archiveBuilder.addEntityDef(getGlossaryCategoryEntity());

        this.archiveBuilder.addRelationshipDef(getCategoryAnchorRelationship());
        this.archiveBuilder.addRelationshipDef(getCategoryHierarchyLinkRelationship());
        this.archiveBuilder.addRelationshipDef(getLibraryCategoryReferenceRelationship());

        this.archiveBuilder.addClassificationDef(getSubjectAreaClassification());
    }


    private EntityDef getGlossaryCategoryEntity()
    {
        final String guid            = "e507485b-9b5a-44c9-8a28-6967f7ff3672";
        final String name            = "GlossaryCategory";
        final String description     = "A collection of related glossary terms.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getCategoryAnchorRelationship()
    {
        final String guid            = "c628938e-815e-47db-8d1c-59bb2e84e028";
        final String name            = "CategoryAnchor";
        final String description     = "Connects a glossary category with its owning glossary.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Glossary";
        final String                     end1AttributeName            = "anchor";
        final String                     end1AttributeDescription     = "Owning glossary for this category.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GlossaryCategory";
        final String                     end2AttributeName            = "categories";
        final String                     end2AttributeDescription     = "Categories owned by this glossary.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getCategoryHierarchyLinkRelationship()
    {
        final String guid            = "71e4b6fb-3412-4193-aff3-a16eccd87e8e";
        final String name            = "CategoryHierarchyLink";
        final String description     = "Relationship between two glossary categories used to create nested categories.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GlossaryCategory";
        final String                     end1AttributeName            = "superCategory";
        final String                     end1AttributeDescription     = "Identifies the parent category.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GlossaryCategory";
        final String                     end2AttributeName            = "subcategories";
        final String                     end2AttributeDescription     = "Glossary categories nested inside this category.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getLibraryCategoryReferenceRelationship()
    {
        final String guid            = "3da21cc9-3cdc-4d87-89b5-c501740f00b2";
        final String name            = "LibraryCategoryReference";
        final String description     = "Links a glossary category to a corresponding category in an external glossary.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GlossaryCategory";
        final String                     end1AttributeName            = "localCategories";
        final String                     end1AttributeDescription     = "Related local glossary categories.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ExternalGlossaryLink";
        final String                     end2AttributeName            = "externalGlossaryCategories";
        final String                     end2AttributeDescription     = "Links to related external glossaries.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "identifier";
        final String attribute1Description     = "Identifier of the corresponding element from the external glossary.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute3Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute4Name            = "lastVerified";
        final String attribute4Description     = "Date when this reference was last checked.";
        final String attribute4DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute4Name,
                                                         attribute4Description,
                                                         attribute4DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private ClassificationDef getSubjectAreaClassification()
    {
        final String guid            = "480e6993-35c5-433a-b50b-0f5c4063fb5d";
        final String name            = "SubjectArea";
        final String description     = "Identifies an element as part of a subject area definition.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.REFERENCEABLE.typeName;

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NAME.name;
        final String attribute1Description     = OpenMetadataProperty.NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;


        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0330 Terms brings in the glossary term that captures a single semantic meaning.
     */
    private void add0330Terms()
    {
        this.archiveBuilder.addEnumDef(getTermRelationshipStatusEnum());

        this.archiveBuilder.addEntityDef(getGlossaryTermEntity());

        this.archiveBuilder.addRelationshipDef(getTermAnchorRelationship());
        this.archiveBuilder.addRelationshipDef(getTermCategorizationRelationship());
        this.archiveBuilder.addRelationshipDef(getLibraryTermReferenceRelationship());
    }

    private EnumDef getTermRelationshipStatusEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(GlossaryTermRelationshipStatus.getOpenTypeGUID(),
                                                        GlossaryTermRelationshipStatus.getOpenTypeName(),
                                                        GlossaryTermRelationshipStatus.getOpenTypeDescription(),
                                                        GlossaryTermRelationshipStatus.getOpenTypeDescriptionGUID(),
                                                        GlossaryTermRelationshipStatus.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (GlossaryTermRelationshipStatus enumValues : GlossaryTermRelationshipStatus.values())
        {
            elementDef = archiveHelper.getEnumElementDef(enumValues.getOrdinal(),
                                                         enumValues.getName(),
                                                         enumValues.getDescription(),
                                                         enumValues.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (enumValues.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getGlossaryTermEntity()
    {
        EntityDef entityDef = archiveHelper.getDefaultEntityDef(OpenMetadataType.GLOSSARY_TERM.typeGUID,
                                                                OpenMetadataType.GLOSSARY_TERM.typeName,
                                                                this.archiveBuilder.getEntityDef(OpenMetadataType.REFERENCEABLE.typeName),
                                                                OpenMetadataType.GLOSSARY_TERM.description,
                                                                OpenMetadataType.GLOSSARY_TERM.descriptionGUID,
                                                                OpenMetadataType.GLOSSARY_TERM.wikiURL);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.SUMMARY.name;
        final String attribute2Description     = OpenMetadataProperty.SUMMARY.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.SUMMARY.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute3Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute4Name            = OpenMetadataProperty.EXAMPLES.name;
        final String attribute4Description     = OpenMetadataProperty.EXAMPLES.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.EXAMPLES.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.ABBREVIATION.name;
        final String attribute5Description     = OpenMetadataProperty.ABBREVIATION.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.ABBREVIATION.descriptionGUID;
        final String attribute6Name            = OpenMetadataProperty.USAGE.name;
        final String attribute6Description     = OpenMetadataProperty.USAGE.description;
        final String attribute6DescriptionGUID = OpenMetadataProperty.USAGE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getTermAnchorRelationship()
    {
        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(OpenMetadataType.TERM_ANCHOR.typeGUID,
                                                                                OpenMetadataType.TERM_ANCHOR.typeName,
                                                                                null,
                                                                                OpenMetadataType.TERM_ANCHOR.description,
                                                                                OpenMetadataType.TERM_ANCHOR.descriptionGUID,
                                                                                OpenMetadataType.TERM_ANCHOR.wikiURL,
                                                                                ClassificationPropagationRule.NONE);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.GLOSSARY_TYPE_NAME;
        final String                     end1AttributeName            = "anchor";
        final String                     end1AttributeDescription     = "Owning glossary.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.GLOSSARY_TERM.typeName;
        final String                     end2AttributeName            = "terms";
        final String                     end2AttributeDescription     = "Terms owned by this glossary.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getTermCategorizationRelationship()
    {
        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(OpenMetadataType.TERM_CATEGORIZATION.typeGUID,
                                                                                OpenMetadataType.TERM_CATEGORIZATION.typeName,
                                                                                null,
                                                                                OpenMetadataType.TERM_CATEGORIZATION.description,
                                                                                OpenMetadataType.TERM_CATEGORIZATION.descriptionGUID,
                                                                                OpenMetadataType.TERM_CATEGORIZATION.wikiURL,
                                                                                ClassificationPropagationRule.NONE);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.GLOSSARY_CATEGORY_TYPE_NAME;
        final String                     end1AttributeName            = "categories";
        final String                     end1AttributeDescription     = "Glossary categories that this term is linked to.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.GLOSSARY_TERM.typeName;
        final String                     end2AttributeName            = "terms";
        final String                     end2AttributeDescription     = "Glossary terms linked to this category.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = "status";
        final String attribute2Description     = "Status of the relationship.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermRelationshipStatus.getOpenTypeName(),
                                                         attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getLibraryTermReferenceRelationship()
    {
        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(OpenMetadataType.LIBRARY_TERM_REFERENCE.typeGUID,
                                                                                OpenMetadataType.LIBRARY_TERM_REFERENCE.typeName,
                                                                                null,
                                                                                OpenMetadataType.LIBRARY_TERM_REFERENCE.description,
                                                                                OpenMetadataType.LIBRARY_TERM_REFERENCE.descriptionGUID,
                                                                                OpenMetadataType.LIBRARY_TERM_REFERENCE.wikiURL,
                                                                                ClassificationPropagationRule.NONE);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.GLOSSARY_TERM.typeName;
        final String                     end1AttributeName            = "localTerms";
        final String                     end1AttributeDescription     = "Related local glossary terms.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ExternalGlossaryLink";
        final String                     end2AttributeName            = "externalGlossaryTerms";
        final String                     end2AttributeDescription     = "Links to related external glossaries.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "identifier";
        final String attribute1Description     = "Identifier of the corresponding element from the external glossary.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute3Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute4Name            = "lastVerified";
        final String attribute4Description     = "Date when this reference was last checked.";
        final String attribute4DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute4Name,
                                                         attribute4Description,
                                                         attribute4DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0340 Dictionary provides classifications for a term that define what type of term it is and
     * how it intended to be used.
     */
    private void add0340Dictionary()
    {
        this.archiveBuilder.addEnumDef(getActivityTypeEnum());

        this.archiveBuilder.addClassificationDef(getActivityDescriptionClassification());
        this.archiveBuilder.addClassificationDef(getAbstractConceptClassification());
        this.archiveBuilder.addClassificationDef(getDataValueClassification());
    }


    private EnumDef getActivityTypeEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(GlossaryTermActivityType.getOpenTypeGUID(),
                                                        GlossaryTermActivityType.getOpenTypeName(),
                                                        GlossaryTermActivityType.getOpenTypeDescription(),
                                                        GlossaryTermActivityType.getOpenTypeDescriptionGUID(),
                                                        GlossaryTermActivityType.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (GlossaryTermActivityType enumValues : GlossaryTermActivityType.values())
        {
            elementDef = archiveHelper.getEnumElementDef(enumValues.getOrdinal(),
                                                         enumValues.getName(),
                                                         enumValues.getDescription(),
                                                         enumValues.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (enumValues.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private ClassificationDef getActivityDescriptionClassification()
    {
        ClassificationDef classificationDef = archiveHelper.getClassificationDef(OpenMetadataType.ACTIVITY_DESCRIPTION_CLASSIFICATION.typeGUID,
                                                                                 OpenMetadataType.ACTIVITY_DESCRIPTION_CLASSIFICATION.typeName,
                                                                                 null,
                                                                                 OpenMetadataType.ACTIVITY_DESCRIPTION_CLASSIFICATION.description,
                                                                                 OpenMetadataType.ACTIVITY_DESCRIPTION_CLASSIFICATION.descriptionGUID,
                                                                                 OpenMetadataType.ACTIVITY_DESCRIPTION_CLASSIFICATION.wikiURL,
                                                                                 this.archiveBuilder.getEntityDef(OpenMetadataType.GLOSSARY_TERM_TYPE_NAME),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.ACTIVITY_TYPE.name;
        final String attribute1Description     = "Classification of the activity.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermActivityType.getOpenTypeName(),
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private ClassificationDef getAbstractConceptClassification()
    {
        final String guid            = "9d725a07-4abf-4939-a268-419d200b69c2";
        final String name            = "AbstractConcept";
        final String description     = "Identifies that this glossary term describes an abstract concept.";
        final String descriptionGUID = null;

        final String linkedToEntity = "GlossaryTerm";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getDataValueClassification()
    {
        final String guid            = "ab253e31-3d8a-45a7-8592-24329a189b9e";
        final String name            = "DataValue";
        final String description     = "Identifies that this glossary term describes a data value.";
        final String descriptionGUID = null;

        final String linkedToEntity = "GlossaryTerm";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0350 Related Terms provides a selection of semantic relationships
     */
    private void add0350RelatedTerms()
    {
        this.archiveBuilder.addRelationshipDef(getRelatedTermRelationship());
        this.archiveBuilder.addRelationshipDef(getSynonymRelationship());
        this.archiveBuilder.addRelationshipDef(getAntonymRelationship());
        this.archiveBuilder.addRelationshipDef(getPreferredTermRelationship());
        this.archiveBuilder.addRelationshipDef(getReplacementTermRelationship());
        this.archiveBuilder.addRelationshipDef(getTranslationRelationship());
        this.archiveBuilder.addRelationshipDef(getISARelationshipRelationship());
        this.archiveBuilder.addRelationshipDef(getValidValueRelationship());
    }


    private RelationshipDef getRelatedTermRelationship()
    {
        final String guid            = "b1161696-e563-4cf9-9fd9-c0c76e47d063";
        final String name            = "RelatedTerm";
        final String description     = "Link between similar glossary terms.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GlossaryTerm";
        final String                     end1AttributeName            = "seeAlso";
        final String                     end1AttributeDescription     = "Related glossary terms.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GlossaryTerm";
        final String                     end2AttributeName            = "seeAlso";
        final String                     end2AttributeDescription     = "Related glossary terms.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = "expression";
        final String attribute2Description     = "An expression that explains the relationship.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "status";
        final String attribute3Description     = "The status of or confidence in the relationship.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute4Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute5Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermRelationshipStatus.getOpenTypeName(),
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getSynonymRelationship()
    {
        final String guid            = "74f4094d-dba2-4ad9-874e-d422b69947e2";
        final String name            = "Synonym";
        final String description     = "Link between glossary terms that have the same meaning.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GlossaryTerm";
        final String                     end1AttributeName            = "synonyms";
        final String                     end1AttributeDescription     = "Glossary terms with the same meaning.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GlossaryTerm";
        final String                     end2AttributeName            = "synonyms";
        final String                     end2AttributeDescription     = "Glossary terms with the same meaning.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = "expression";
        final String attribute2Description     = "An expression that explains the relationship.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "status";
        final String attribute3Description     = "The status of or confidence in the relationship.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute4Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute5Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermRelationshipStatus.getOpenTypeName(),
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getAntonymRelationship()
    {
        final String guid            = "ea5e126a-a8fa-4a43-bcfa-309a98aa0185";
        final String name            = "Antonym";
        final String description     = "Link between glossary terms that have the opposite meaning.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GlossaryTerm";
        final String                     end1AttributeName            = "antonyms";
        final String                     end1AttributeDescription     = "Glossary terms with the opposite meaning.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GlossaryTerm";
        final String                     end2AttributeName            = "antonyms";
        final String                     end2AttributeDescription     = "Glossary terms with the opposite meaning.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = "expression";
        final String attribute2Description     = "An expression that explains the relationship.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "status";
        final String attribute3Description     = "The status of or confidence in the relationship.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute4Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute5Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermRelationshipStatus.getOpenTypeName(),
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getPreferredTermRelationship()
    {
        final String guid            = "8ac8f9de-9cdd-4103-8a33-4cb204b78c2a";
        final String name            = "PreferredTerm";
        final String description     = "Link to an alternative term that the organization prefer is used.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GlossaryTerm";
        final String                     end1AttributeName            = "alternateTerms";
        final String                     end1AttributeDescription     = "Alternative glossary terms.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GlossaryTerm";
        final String                     end2AttributeName            = "preferredTerms";
        final String                     end2AttributeDescription     = "Related glossary terms.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = "expression";
        final String attribute2Description     = "An expression that explains the relationship.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "status";
        final String attribute3Description     = "The status of or confidence in the relationship.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute4Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute5Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermRelationshipStatus.getOpenTypeName(),
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getReplacementTermRelationship()
    {
        final String guid            = "3bac5f35-328b-4bbd-bfc9-3b3c9ba5e0ed";
        final String name            = "ReplacementTerm";
        final String description     = "Link to a glossary term that is replacing an obsolete glossary term.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GlossaryTerm";
        final String                     end1AttributeName            = "replacedTerms";
        final String                     end1AttributeDescription     = "Replaced glossary terms.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GlossaryTerm";
        final String                     end2AttributeName            = "replacementTerms";
        final String                     end2AttributeDescription     = "Replacement glossary terms.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = "expression";
        final String attribute2Description     = "An expression that explains the relationship.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "status";
        final String attribute3Description     = "The status of or confidence in the relationship.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute4Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute5Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermRelationshipStatus.getOpenTypeName(),
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getTranslationRelationship()
    {
        final String guid            = "6ae42e95-efc5-4256-bfa8-801140a29d2a";
        final String name            = "Translation";
        final String description     = "Link between glossary terms that provide different natural language translation of the same concept.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GlossaryTerm";
        final String                     end1AttributeName            = "translations";
        final String                     end1AttributeDescription     = "Translations of glossary term.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GlossaryTerm";
        final String                     end2AttributeName            = "translations";
        final String                     end2AttributeDescription     = "Translations of glossary term.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = "expression";
        final String attribute2Description     = "An expression that explains the relationship.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "status";
        final String attribute3Description     = "The status of or confidence in the relationship.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute4Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute5Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermRelationshipStatus.getOpenTypeName(),
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getISARelationshipRelationship()
    {
        final String guid            = "50fab7c7-68bc-452f-b8eb-ec76829cac85";
        final String name            = "ISARelationship";
        final String description     = "Link between a more general glossary term and a more specific definition.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GlossaryTerm";
        final String                     end1AttributeName            = "classifies";
        final String                     end1AttributeDescription     = "More specific glossary terms.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GlossaryTerm";
        final String                     end2AttributeName            = "isA";
        final String                     end2AttributeDescription     = "More general glossary terms.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = "expression";
        final String attribute2Description     = "An expression that explains the relationship.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "status";
        final String attribute3Description     = "The status of or confidence in the relationship.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute4Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute5Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermRelationshipStatus.getOpenTypeName(),
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getValidValueRelationship()
    {
        final String guid            = "707a156b-e579-4482-89a5-de5889da1971";
        final String name            = "ValidValue";
        final String description     = "Link between glossary terms where one defines one of the data values for the another.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GlossaryTerm";
        final String                     end1AttributeName            = "validValueFor";
        final String                     end1AttributeDescription     = "Glossary terms for data items that can be set to this value.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GlossaryTerm";
        final String                     end2AttributeName            = "validValues";
        final String                     end2AttributeDescription     = "Glossary terms for data values that can be used with data items represented by this glossary term.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = "expression";
        final String attribute2Description     = "An expression that explains the relationship.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "status";
        final String attribute3Description     = "The status of or confidence in the relationship.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute4Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute5Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermRelationshipStatus.getOpenTypeName(),
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0360 Contexts allows Glossary Terms to linked to specific contexts (also defined with Glossary Terms).
     */
    private void add0360Contexts()
    {
        this.archiveBuilder.addRelationshipDef(getUsedInContextRelationship());

        this.archiveBuilder.addClassificationDef(getContextDefinitionClassification());
    }

    private RelationshipDef getUsedInContextRelationship()
    {
        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(OpenMetadataType.USED_IN_CONTEXT.typeGUID,
                                                                                OpenMetadataType.USED_IN_CONTEXT.typeName,
                                                                                null,
                                                                                OpenMetadataType.USED_IN_CONTEXT.description,
                                                                                OpenMetadataType.USED_IN_CONTEXT.descriptionGUID,
                                                                                OpenMetadataType.USED_IN_CONTEXT.wikiURL,
                                                                                ClassificationPropagationRule.NONE);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GlossaryTerm";
        final String                     end1AttributeName            = "contextRelevantTerms";
        final String                     end1AttributeDescription     = "Glossary terms used in this specific context.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GlossaryTerm";
        final String                     end2AttributeName            = "usedInContexts";
        final String                     end2AttributeDescription     = "Glossary terms describing the contexts where this term is used.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = "expression";
        final String attribute2Description     = "An expression that explains the relationship.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "status";
        final String attribute3Description     = "The status of or confidence in the relationship.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute4Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute5Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermRelationshipStatus.getOpenTypeName(),
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private ClassificationDef getContextDefinitionClassification()
    {
        final String guid            = "54f9f41a-3871-4650-825d-59a41de01330";
        final String name            = "ContextDefinition";
        final String description     = "Identifies a glossary term that describes a context where processing or decisions occur.";
        final String descriptionGUID = null;

        final String linkedToEntity = "GlossaryTerm";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = "scope";
        final String attribute2Description     = "Scope of influence of the context.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0370 Semantic Assignment defines a relationship between a Glossary Term and an Asset
     */
    private void add0370SemanticAssignment()
    {
        this.archiveBuilder.addEnumDef(getTermAssignmentStatusEnum());

        this.archiveBuilder.addRelationshipDef(getSemanticAssignmentRelationship());
    }


    private EnumDef getTermAssignmentStatusEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(GlossaryTermAssignmentStatus.getOpenTypeGUID(),
                                                        GlossaryTermAssignmentStatus.getOpenTypeName(),
                                                        GlossaryTermAssignmentStatus.getOpenTypeDescription(),
                                                        GlossaryTermAssignmentStatus.getOpenTypeDescriptionGUID(),
                                                        GlossaryTermAssignmentStatus.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (GlossaryTermAssignmentStatus enumValues : GlossaryTermAssignmentStatus.values())
        {
            elementDef = archiveHelper.getEnumElementDef(enumValues.getOrdinal(),
                                                         enumValues.getName(),
                                                         enumValues.getDescription(),
                                                         enumValues.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (enumValues.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private RelationshipDef getSemanticAssignmentRelationship()
    {
        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(OpenMetadataType.SEMANTIC_ASSIGNMENT.typeGUID,
                                                                                OpenMetadataType.SEMANTIC_ASSIGNMENT.typeName,
                                                                                null,
                                                                                OpenMetadataType.SEMANTIC_ASSIGNMENT.description,
                                                                                OpenMetadataType.SEMANTIC_ASSIGNMENT.descriptionGUID,
                                                                                OpenMetadataType.SEMANTIC_ASSIGNMENT.wikiURL,
                                                                                ClassificationPropagationRule.NONE);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "assignedElements";
        final String                     end1AttributeDescription     = "Elements identified as managing data that has the same meaning as this glossary term.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.GLOSSARY_TYPE_NAME;
        final String                     end2AttributeName            = "meaning";
        final String                     end2AttributeDescription     = "Semantic definition for this element.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = "expression";
        final String attribute2Description     = "Expression describing the relationship.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = OpenMetadataProperty.TERM_ASSIGNMENT_STATUS.name;
        final String attribute3Description     = OpenMetadataProperty.TERM_ASSIGNMENT_STATUS.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.TERM_ASSIGNMENT_STATUS.descriptionGUID;
        final String attribute4Name            = OpenMetadataProperty.CONFIDENCE.name;
        final String attribute4Description     = OpenMetadataProperty.CONFIDENCE.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.CONFIDENCE.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute5Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute6Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute6Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute6DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermAssignmentStatus.getOpenTypeName(),
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute4Name,
                                                        attribute4Description,
                                                        attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0380 Spine Objects enables relationships to be established between objects and their attributes.
     */
    private void add0380SpineObjects()
    {
        this.archiveBuilder.addRelationshipDef(getTermHASARelationshipRelationship());
        this.archiveBuilder.addRelationshipDef(getTermISATYPEOFRelationshipRelationship());
        this.archiveBuilder.addRelationshipDef(getTermTYPEDBYRelationshipRelationship());

        this.archiveBuilder.addClassificationDef(getSpineObjectClassification());
        this.archiveBuilder.addClassificationDef(getSpineAttributeClassification());
        this.archiveBuilder.addClassificationDef(getObjectIdentifierClassification());
    }


    private RelationshipDef getTermHASARelationshipRelationship()
    {
        final String guid            = "d67f16d1-5348-419e-ba38-b0bb6fe4ad6c";
        final String name            = "TermHASARelationship";
        final String description     = "Defines the relationship between a spine object and a spine attribute.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GlossaryTerm";
        final String                     end1AttributeName            = "objects";
        final String                     end1AttributeDescription     = "Objects where this attribute may occur.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GlossaryTerm";
        final String                     end2AttributeName            = "attributes";
        final String                     end2AttributeDescription     = "Typical attributes for this object.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "status";
        final String attribute3Description     = "The status of or confidence in the relationship.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute4Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute5Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermRelationshipStatus.getOpenTypeName(),
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getTermISATYPEOFRelationshipRelationship()
    {
        final String guid            = "d5d588c3-46c9-420c-adff-6031802a7e51";
        final String name            = "TermISATypeOFRelationship";
        final String description     = "Defines an inheritance relationship between two spine objects.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GlossaryTerm";
        final String                     end1AttributeName            = "supertypes";
        final String                     end1AttributeDescription     = "Supertypes for this object.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GlossaryTerm";
        final String                     end2AttributeName            = "subtypes";
        final String                     end2AttributeDescription     = "Subtypes for this object.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "status";
        final String attribute3Description     = "The status of or confidence in the relationship.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute4Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute5Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermRelationshipStatus.getOpenTypeName(),
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getTermTYPEDBYRelationshipRelationship()
    {
        final String guid            = "669e8aa4-c671-4ee7-8d03-f37d09b9d006";
        final String name            = "TermTYPEDBYRelationship";
        final String description     = "Defines the relationship between a spine attribute and its type.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GlossaryTerm";
        final String                     end1AttributeName            = "attributesTypedBy";
        final String                     end1AttributeDescription     = "Attributes of this type.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GlossaryTerm";
        final String                     end2AttributeName            = "types";
        final String                     end2AttributeDescription     = "Types for this attribute.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "status";
        final String attribute3Description     = "The status of or confidence in the relationship.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute4Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute5Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute(GlossaryTermRelationshipStatus.getOpenTypeName(),
                                                         attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private ClassificationDef getSpineObjectClassification()
    {
        final String guid            = "a41ee152-de1e-4533-8535-2f8b37897cac";
        final String name            = "SpineObject";
        final String description     = "Identifies a glossary term that describes a type of spine object.";
        final String descriptionGUID = null;

        final String linkedToEntity = "GlossaryTerm";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getSpineAttributeClassification()
    {
        final String guid            = "ccb749ba-34ec-4f71-8755-4d8b383c34c3";
        final String name            = "SpineAttribute";
        final String description     = "Identifies a glossary term that describes an attribute of a spine object.";
        final String descriptionGUID = null;

        final String linkedToEntity = "GlossaryTerm";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getObjectIdentifierClassification()
    {
        final String guid            = "3d1e4389-27de-44fa-8df4-d57bfaf809ea";
        final String name            = "ObjectIdentifier";
        final String description     = "Identifies a glossary term that describes an attribute that can be used to identify an instance.";
        final String descriptionGUID = null;

        final String linkedToEntity = "GlossaryTerm";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0385 ControlledGlossaryDevelopment covers glossary term objects that are created and
     * maintained through a workflow process.
     */
    private void add0385ControlledGlossaryDevelopment()
    {
        this.archiveBuilder.addEntityDef(getControlledGlossaryTermEntity());
    }


    private EntityDef getControlledGlossaryTermEntity()
    {
        final String guid            = "c04e29b2-2d66-48fc-a20d-e59895de6040";
        final String name            = "ControlledGlossaryTerm";
        final String description     = "Defines a glossary term that is developed through a controlled workflow.";
        final String descriptionGUID = null;

        final String superTypeName = "GlossaryTerm";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        ArrayList<InstanceStatus> validInstanceStatusList = new ArrayList<>();
        validInstanceStatusList.add(InstanceStatus.DRAFT);
        validInstanceStatusList.add(InstanceStatus.PROPOSED);
        validInstanceStatusList.add(InstanceStatus.APPROVED);
        validInstanceStatusList.add(InstanceStatus.ACTIVE);
        validInstanceStatusList.add(InstanceStatus.DELETED);
        entityDef.setValidInstanceStatusList(validInstanceStatusList);
        entityDef.setInitialStatus(InstanceStatus.DRAFT);

        return entityDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0390 Glossary Project provides a classification for a project to say it is updating glossary terms.
     */
    private void add0390GlossaryProject()
    {
        this.archiveBuilder.addClassificationDef(getGlossaryProjectClassification());
    }


    private ClassificationDef getGlossaryProjectClassification()
    {
        final String guid            = "43be51a9-2d19-4044-b399-3ba36af10929";
        final String name            = "GlossaryProject";
        final String description     = "Identifies a project that is defining new glossary terms and categories or maintaining an existing glossary.";
        final String descriptionGUID = null;

        final String linkedToEntity = "Project";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    /*
     * ========================================
     * AREA 4: governance
     */

    /**
     * Area 4 models cover the governance entities and relationships.
     */
    private void addArea4Types()
    {
        this.add0401GovernanceDefinitions();
        this.add0405GovernanceDrivers();
        this.add0415GovernanceResponses();
        this.add0417GovernanceProject();
        this.add0420GovernanceControls();
        this.add0421GovernanceConfidentialityLevels();
        this.add0422GovernanceActionClassifications();
        this.add0423SecurityTags();
        this.add0424GovernanceZones();
        this.add0430TechnicalControls();
        this.add0438NamingStandards();
        this.add0440OrganizationalControls();
        this.add0442ProjectCharter();
        this.add0445GovernanceRoles();
        this.add0450GovernanceRollout();
        this.add0455ExceptionManagement();
        this.add0460GovernanceControls();
        this.add0481Licenses();
        this.add0482Certifications();
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0401 The definitions that control the governance of data assets are authored in the metadata repository.
     * They are referenceable and they make use of external references for more information.
     */
    private void add0401GovernanceDefinitions()
    {
        this.archiveBuilder.addEnumDef(getGovernanceDomainEnum());

        this.archiveBuilder.addEntityDef(getGovernanceDefinitionEntity());
        this.archiveBuilder.addEntityDef(getGovernanceOfficerEntity());
    }


    private EnumDef getGovernanceDomainEnum()
    {
        final String guid            = "baa31998-f3cb-47b0-9123-674a701e87bc";
        final String name            = "GovernanceDomain";
        final String description     = "Defines the governance domains that open metadata seeks to unite.";
        final String descriptionGUID = null;

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID);

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        final int    element1Ordinal         = 0;
        final String element1Value           = "All";
        final String element1Description     = "Relevant to all governance domains.";
        final String element1DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element1Ordinal,
                                                     element1Value,
                                                     element1Description,
                                                     element1DescriptionGUID);

        elementDefs.add(elementDef);

        final int    element2Ordinal         = 1;
        final String element2Value           = "Data";
        final String element2Description     = "The data (information) governance domain.";
        final String element2DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element2Ordinal,
                                                     element2Value,
                                                     element2Description,
                                                     element2DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element3Ordinal         = 2;
        final String element3Value           = "Privacy";
        final String element3Description     = "The data privacy domain.";
        final String element3DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element3Ordinal,
                                                     element3Value,
                                                     element3Description,
                                                     element3DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element4Ordinal         = 3;
        final String element4Value           = "Security";
        final String element4Description     = "The security governance domain.";
        final String element4DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element4Ordinal,
                                                     element4Value,
                                                     element4Description,
                                                     element4DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element5Ordinal         = 4;
        final String element5Value           = OpenMetadataType.IT_INFRASTRUCTURE.typeName;
        final String element5Description     = "The IT infrastructure governance domain.";
        final String element5DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element5Ordinal,
                                                     element5Value,
                                                     element5Description,
                                                     element5DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element6Ordinal         = 5;
        final String element6Value           = "SoftwareDevelopment";
        final String element6Description     = "The software development lifecycle governance domain.";
        final String element6DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element6Ordinal,
                                                     element6Value,
                                                     element6Description,
                                                     element6DescriptionGUID);

        elementDefs.add(elementDef);

        final int    element7Ordinal         = 6;
        final String element7Value           = "Corporate";
        final String element7Description     = "The corporate governance domain.";
        final String element7DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element7Ordinal,
                                                     element7Value,
                                                     element7Description,
                                                     element7DescriptionGUID);

        elementDefs.add(elementDef);

        final int    element8Ordinal         = 7;
        final String element8Value           = "AssetManagement";
        final String element8Description     = "The physical asset management governance domain.";
        final String element8DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element8Ordinal,
                                                     element8Value,
                                                     element8Description,
                                                     element8DescriptionGUID);

        elementDefs.add(elementDef);

        final int    element99Ordinal         = 99;
        final String element99Value           = "Other";
        final String element99Description     = "Another governance domain.";
        final String element99DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element99Ordinal,
                                                     element99Value,
                                                     element99Description,
                                                     element99DescriptionGUID);
        elementDefs.add(elementDef);

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getGovernanceDefinitionEntity()
    {
        final String guid            = "578a3500-9ad3-45fe-8ada-e4e9572c37c8";
        final String name            = "GovernanceDefinition";
        final String description     = "Defines an aspect of the governance program.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "title";
        final String attribute1Description     = "Title describing the governance definition.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "summary";
        final String attribute2Description     = "Short summary of the governance definition.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute3Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute4Name            = "scope";
        final String attribute4Description     = "Scope of impact for this governance definition.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "domain";
        final String attribute5Description     = "Governance domain for this governance definition.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "importance";
        final String attribute6Description     = "Relative importance of this governance definition compared to its peers.";
        final String attribute6DescriptionGUID = null;
        final String attribute7Name            = "implications";
        final String attribute7Description     = "Impact on the organization, people and services when adopting the recommendation in this governance definition.";
        final String attribute7DescriptionGUID = null;
        final String attribute8Name            = "outcomes";
        final String attribute8Description     = "Expected outcomes.";
        final String attribute8DescriptionGUID = null;
        final String attribute9Name            = "results";
        final String attribute9Description     = "Actual results.";
        final String attribute9DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("GovernanceDomain",
                                                         attribute5Name,
                                                         attribute5Description,
                                                         attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute7Name,
                                                                attribute7Description,
                                                                attribute7DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute8Name,
                                                                attribute8Description,
                                                                attribute8DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute9Name,
                                                                attribute9Description,
                                                                attribute9DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        ArrayList<InstanceStatus> validInstanceStatusList = new ArrayList<>();
        validInstanceStatusList.add(InstanceStatus.DRAFT);
        validInstanceStatusList.add(InstanceStatus.PROPOSED);
        validInstanceStatusList.add(InstanceStatus.APPROVED);
        validInstanceStatusList.add(InstanceStatus.ACTIVE);
        validInstanceStatusList.add(InstanceStatus.DEPRECATED);
        validInstanceStatusList.add(InstanceStatus.OTHER);
        validInstanceStatusList.add(InstanceStatus.DELETED);
        entityDef.setValidInstanceStatusList(validInstanceStatusList);
        entityDef.setInitialStatus(InstanceStatus.DRAFT);

        return entityDef;
    }


    private EntityDef getGovernanceOfficerEntity()
    {
        final String guid            = "578a3510-9ad3-45fe-8ada-e4e9572c37c8";
        final String name            = "GovernanceOfficer";
        final String description     = "Person responsible for a governance domain.";
        final String descriptionGUID = null;

        final String superTypeName = "PersonRole";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "domain";
        final String attribute1Description     = "Governance domain for this governance officer.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getEnumTypeDefAttribute("GovernanceDomain",
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0405 Governance Drivers defines the factors that drive the governance program.
     */
    private void add0405GovernanceDrivers()
    {
        this.archiveBuilder.addEntityDef(getGovernanceDriverEntity());
        this.archiveBuilder.addEntityDef(getGovernanceStrategyEntity());
        this.archiveBuilder.addEntityDef(getRegulationEntity());
    }


    private EntityDef getGovernanceDriverEntity()
    {
        final String guid            = "c403c109-7b6b-48cd-8eee-df445b258b33";
        final String name            = "GovernanceDriver";
        final String description     = "Defines a reason for having the governance program.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernanceDefinition";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getGovernanceStrategyEntity()
    {
        final String guid            = "3c34f121-07a6-4e95-a07d-9b0ef17b7bbf";
        final String name            = "GovernanceStrategy";
        final String description     = "Defines how the governance program and the supporting capabilities are supporting the business strategy.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernanceDriver";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "businessImperatives";
        final String attribute1Description     = "Goals or required outcomes from the business strategy that is supported by the data strategy.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getArrayStringTypeDefAttribute(attribute1Name,
                                                                attribute1Description,
                                                                attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getRegulationEntity()
    {
        final String guid            = "e3c4293d-8846-4500-b0c0-197d73aba8b0";
        final String name            = "Regulation";
        final String description     = "Identifies a regulation related to data that must be supported.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernanceDriver";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "jurisdiction";
        final String attribute1Description     = "Issuing authority for the regulation.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0415 Governance Responses lay out the approaches, principles and obligations that follow from the
     * governance drivers.
     */
    private void add0415GovernanceResponses()
    {
        this.archiveBuilder.addEntityDef(getGovernancePolicyEntity());
        this.archiveBuilder.addEntityDef(getGovernancePrincipleEntity());
        this.archiveBuilder.addEntityDef(getGovernanceObligationEntity());
        this.archiveBuilder.addEntityDef(getGovernanceApproachEntity());

        this.archiveBuilder.addRelationshipDef(getGovernancePolicyLinkRelationship());
        this.archiveBuilder.addRelationshipDef(getGovernanceResponseRelationship());
    }


    private EntityDef getGovernancePolicyEntity()
    {
        final String guid            = "a7defa41-9cfa-4be5-9059-359022bb016d";
        final String name            = "GovernancePolicy";
        final String description     = "Defines a goal or outcome expected from the organization.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernanceDefinition";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getGovernancePrincipleEntity()
    {
        final String guid            = "3b7d1325-ec2c-44cb-8db0-ce207beb78cf";
        final String name            = "GovernancePrinciple";
        final String description     = "Defines a principle related to how data is managed or used that the organization should ensure remains true.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernancePolicy";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getGovernanceObligationEntity()
    {
        final String guid            = "0cec20d3-aa29-41b7-96ea-1c544ed32537";
        final String name            = "GovernanceObligation";
        final String description     = "Defines a capability, rule or action that is required by a regulation or external party.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernancePolicy";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getGovernanceApproachEntity()
    {
        final String guid            = "2d03ec9d-bd6b-4be9-8e17-95a7ecdbaa67";
        final String name            = "GovernanceApproach";
        final String description     = "Defines a preferred approach to managing or using data.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernancePolicy";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getGovernancePolicyLinkRelationship()
    {
        final String guid            = "0c42c999-4cac-4da4-afab-0e381f3a818e";
        final String name            = "GovernancePolicyLink";
        final String description     = "Links related governance policies together.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GovernancePolicy";
        final String                     end1AttributeName            = "linkingPolicies";
        final String                     end1AttributeDescription     = "Policies that are dependent on this policy.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GovernancePolicy";
        final String                     end2AttributeName            = "linkedPolicies";
        final String                     end2AttributeDescription     = "Policies that further define aspects of this policy.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getGovernanceResponseRelationship()
    {
        final String guid            = "8845990e-7fd9-4b79-a19d-6c4730dadd6b";
        final String name            = "GovernanceResponse";
        final String description     = "Links a governance policy to a governance driver that it is supporting.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GovernanceDriver";
        final String                     end1AttributeName            = "drivers";
        final String                     end1AttributeDescription     = "Drivers that justify this policy.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GovernancePolicy";
        final String                     end2AttributeName            = "policies";
        final String                     end2AttributeDescription     = "Governance policies that support this governance driver.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "rationale";
        final String attribute1Description     = "Describes the reasoning for defining the policy in support of the driver.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);


        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0417 Governance Project provides a classification for a project
     */
    private void add0417GovernanceProject()
    {
        this.archiveBuilder.addClassificationDef(getGovernanceProjectClassification());
    }


    private ClassificationDef getGovernanceProjectClassification()
    {
        final String guid            = "37142317-4125-4046-9514-71dc5031563f";
        final String name            = "GovernanceProject";
        final String description     = "Identifies that a project is rolling out capability to support the governance program.";
        final String descriptionGUID = null;

        final String linkedToEntity = "Project";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0420 Governance Controls define the implementation of the governance policies
     */
    private void add0420GovernanceControls()
    {
        this.archiveBuilder.addEntityDef(getGovernanceControlEntity());
        this.archiveBuilder.addEntityDef(getTechnicalControlEntity());
        this.archiveBuilder.addEntityDef(getOrganizationalControlEntity());

        this.archiveBuilder.addRelationshipDef(getGovernanceImplementationRelationship());
        this.archiveBuilder.addRelationshipDef(getGovernanceControlLinkRelationship());
    }


    private EntityDef getGovernanceControlEntity()
    {
        final String guid            = "c794985e-a10b-4b6c-9dc2-6b2e0a2901d3";
        final String name            = "GovernanceControl";
        final String description     = "An implementation of a governance capability.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernanceDefinition";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "implementationDescription";
        final String attribute1Description     = "Description of how this governance control should be implemented.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getTechnicalControlEntity()
    {
        final String guid            = "d8f6eb5b-36f0-49bd-9b25-bf16f370d1ec";
        final String name            = "TechnicalControl";
        final String description     = "A governance control that is implemented using technology.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernanceControl";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getOrganizationalControlEntity()
    {
        final String guid            = "befa1458-79b8-446a-b813-536700e60fa8";
        final String name            = "OrganizationalControl";
        final String description     = "A governance control that is implemented using organization structure, training, roles manual procedures and reviews.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernanceControl";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getGovernanceImplementationRelationship()
    {
        final String guid            = "787eaf46-7cf2-4096-8d6e-671a0819d57e";
        final String name            = "GovernanceImplementation";
        final String description     = "A link between a governance control and the governance driver it is implementing.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GovernancePolicy";
        final String                     end1AttributeName            = "policies";
        final String                     end1AttributeDescription     = "The policies that are supported by this control.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GovernanceControl";
        final String                     end2AttributeName            = "implementations";
        final String                     end2AttributeDescription     = "The governance controls that implement this policy.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "rationale";
        final String attribute1Description     = "The reasons for implementing the policy using this control.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getGovernanceControlLinkRelationship()
    {
        final String guid            = "806933fb-7925-439b-9876-922a960d2ba1";
        final String name            = "GovernanceControlLink";
        final String description     = "A link between two related governance controls.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GovernanceControl";
        final String                     end1AttributeName            = "linkingControls";
        final String                     end1AttributeDescription     = "Governance controls that ate dependent on this control.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GovernanceControl";
        final String                     end2AttributeName            = "linkedControls";
        final String                     end2AttributeDescription     = "Governance controls that support the implementation of this control.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0421 Governance Confidentiality Level defines the organization's levels of confidentiality.
     */
    private void add0421GovernanceConfidentialityLevels()
    {
        this.archiveBuilder.addEntityDef(getGovernanceConfidentialityLevelEntity());
    }


    private EntityDef getGovernanceConfidentialityLevelEntity()
    {
        final String guid            = "49dd320b-4850-4838-9b78-f1285f0e6d2f";
        final String name            = "GovernanceConfidentialityLevel";
        final String description     = "A definition of a confidentiality level.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "level";
        final String attribute1Description     = "Numeric value for the confidentiality - the higher the number, the more confidential";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute2Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute3Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getIntTypeDefAttribute(attribute1Name,
                                                        attribute1Description,
                                                        attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0422 Governance Action Classifications provide the key classification that drive information governance.
     */
    private void add0422GovernanceActionClassifications()
    {
        this.archiveBuilder.addEnumDef(getGovernanceClassificationStatusEnum());
        this.archiveBuilder.addEnumDef(getConfidentialityLevelEnum());
        this.archiveBuilder.addEnumDef(getConfidenceLevelEnum());
        this.archiveBuilder.addEnumDef(getRetentionBasisEnum());
        this.archiveBuilder.addEnumDef(getCriticalityLevelEnum());

        this.archiveBuilder.addClassificationDef(getConfidentialityClassification());
        this.archiveBuilder.addClassificationDef(getConfidenceClassification());
        this.archiveBuilder.addClassificationDef(getRetentionClassification());
        this.archiveBuilder.addClassificationDef(getCriticalityClassification());
    }


    private EnumDef getGovernanceClassificationStatusEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(GovernanceClassificationStatus.getOpenTypeGUID(),
                                                        GovernanceClassificationStatus.getOpenTypeName(),
                                                        GovernanceClassificationStatus.getOpenTypeDescription(),
                                                        GovernanceClassificationStatus.getOpenTypeDescriptionGUID(),
                                                        GovernanceClassificationStatus.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (CriticalityLevel enumValues : CriticalityLevel.values())
        {
            elementDef = archiveHelper.getEnumElementDef(enumValues.getOrdinal(),
                                                         enumValues.getName(),
                                                         enumValues.getDescription(),
                                                         enumValues.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (enumValues.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }

    private EnumDef getConfidentialityLevelEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(ConfidentialityLevel.getOpenTypeGUID(),
                                                        ConfidentialityLevel.getOpenTypeName(),
                                                        ConfidentialityLevel.getOpenTypeDescription(),
                                                        ConfidentialityLevel.getOpenTypeDescriptionGUID(),
                                                        ConfidentialityLevel.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (ConfidentialityLevel enumValues : ConfidentialityLevel.values())
        {
            elementDef = archiveHelper.getEnumElementDef(enumValues.getOrdinal(),
                                                         enumValues.getName(),
                                                         enumValues.getDescription(),
                                                         enumValues.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (enumValues.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }

    private EnumDef getConfidenceLevelEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(ConfidenceLevel.getOpenTypeGUID(),
                                                        ConfidenceLevel.getOpenTypeName(),
                                                        ConfidenceLevel.getOpenTypeDescription(),
                                                        ConfidenceLevel.getOpenTypeDescriptionGUID(),
                                                        ConfidenceLevel.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (ConfidenceLevel enumValues : ConfidenceLevel.values())
        {
            elementDef = archiveHelper.getEnumElementDef(enumValues.getOrdinal(),
                                                         enumValues.getName(),
                                                         enumValues.getDescription(),
                                                         enumValues.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (enumValues.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }

    private EnumDef getRetentionBasisEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(RetentionBasis.getOpenTypeGUID(),
                                                        RetentionBasis.getOpenTypeName(),
                                                        RetentionBasis.getOpenTypeDescription(),
                                                        RetentionBasis.getOpenTypeDescriptionGUID(),
                                                        RetentionBasis.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (RetentionBasis enumValues : RetentionBasis.values())
        {
            elementDef = archiveHelper.getEnumElementDef(enumValues.getOrdinal(),
                                                         enumValues.getName(),
                                                         enumValues.getDescription(),
                                                         enumValues.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (enumValues.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }

    private EnumDef getCriticalityLevelEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(CriticalityLevel.getOpenTypeGUID(),
                                                        CriticalityLevel.getOpenTypeName(),
                                                        CriticalityLevel.getOpenTypeDescription(),
                                                        CriticalityLevel.getOpenTypeDescriptionGUID(),
                                                        CriticalityLevel.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (CriticalityLevel enumValues : CriticalityLevel.values())
        {
            elementDef = archiveHelper.getEnumElementDef(enumValues.getOrdinal(),
                                                         enumValues.getName(),
                                                         enumValues.getDescription(),
                                                         enumValues.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (enumValues.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }

    private ClassificationDef getConfidentialityClassification()
    {
        ClassificationDef classificationDef = archiveHelper.getClassificationDef(OpenMetadataType.CONFIDENTIALITY_CLASSIFICATION.typeGUID,
                                                                                 OpenMetadataType.CONFIDENTIALITY_CLASSIFICATION.typeName,
                                                                                 null,
                                                                                 OpenMetadataType.CONFIDENTIALITY_CLASSIFICATION.description,
                                                                                 OpenMetadataType.CONFIDENTIALITY_CLASSIFICATION.descriptionGUID,
                                                                                 OpenMetadataType.CONFIDENTIALITY_CLASSIFICATION.wikiURL,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         OpenMetadataType.REFERENCEABLE.typeName),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "status";
        final String attribute1Description     = "Status of this classification.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = OpenMetadataProperty.CONFIDENCE.name;
        final String attribute2Description     = OpenMetadataProperty.CONFIDENCE.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.CONFIDENCE.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute3Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute4Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute4Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.NOTES.name;
        final String attribute5Description     = OpenMetadataProperty.NOTES.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.NOTES.descriptionGUID;
        final String attribute6Name            = "level";
        final String attribute6Description     = "Level of confidentiality.";
        final String attribute6DescriptionGUID = null;

        property = archiveHelper.getEnumTypeDefAttribute("GovernanceClassificationStatus",
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute2Name,
                                                        attribute2Description,
                                                        attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute6Name,
                                                        attribute6Description,
                                                        attribute6DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }

    private ClassificationDef getConfidenceClassification()
    {
        final String guid            = "25d8f8d5-2998-4983-b9ef-265f58732965";
        final String name            = "Confidence";
        final String description     = "Defines the level of confidence that should be placed in the accuracy of related data items.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.REFERENCEABLE.typeName;

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "status";
        final String attribute1Description     = "Status of this classification.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = OpenMetadataProperty.CONFIDENCE.name;
        final String attribute2Description     = OpenMetadataProperty.CONFIDENCE.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.CONFIDENCE.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute3Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute4Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute4Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.NOTES.name;
        final String attribute5Description     = OpenMetadataProperty.NOTES.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.NOTES.descriptionGUID;
        final String attribute6Name            = "level";
        final String attribute6Description     = "Level of confidence in the quality of this data.";
        final String attribute6DescriptionGUID = null;


        property = archiveHelper.getEnumTypeDefAttribute("GovernanceClassificationStatus",
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute2Name,
                                                        attribute2Description,
                                                        attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("ConfidenceLevel",
                                                         attribute6Name,
                                                         attribute6Description,
                                                         attribute6DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }

    private ClassificationDef getRetentionClassification()
    {
        ClassificationDef classificationDef = archiveHelper.getClassificationDef(OpenMetadataType.RETENTION_CLASSIFICATION.typeGUID,
                                                                                 OpenMetadataType.RETENTION_CLASSIFICATION.typeName,
                                                                                 null,
                                                                                 OpenMetadataType.RETENTION_CLASSIFICATION.description,
                                                                                 OpenMetadataType.RETENTION_CLASSIFICATION.descriptionGUID,
                                                                                 OpenMetadataType.RETENTION_CLASSIFICATION.wikiURL,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         OpenMetadataType.REFERENCEABLE.typeName),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "status";
        final String attribute1Description     = "Status of this classification.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = OpenMetadataProperty.CONFIDENCE.name;
        final String attribute2Description     = OpenMetadataProperty.CONFIDENCE.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.CONFIDENCE.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute3Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute4Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute4Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.NOTES.name;
        final String attribute5Description     = OpenMetadataProperty.NOTES.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.NOTES.descriptionGUID;
        final String attribute6Name            = "basis";
        final String attribute6Description     = "Basis on which the retention period is defined.";
        final String attribute6DescriptionGUID = null;
        final String attribute7Name            = "associatedGUID";
        final String attribute7Description     = "Related entity used to determine the retention period.";
        final String attribute7DescriptionGUID = null;
        final String attribute8Name            = "archiveAfter";
        final String attribute8Description     = "Date when archiving can take place.";
        final String attribute8DescriptionGUID = null;
        final String attribute9Name            = "deleteAfter";
        final String attribute9Description     = "Date when delete can take place.";
        final String attribute9DescriptionGUID = null;

        property = archiveHelper.getEnumTypeDefAttribute("GovernanceClassificationStatus",
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute2Name,
                                                        attribute2Description,
                                                        attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("RetentionBasis",
                                                         attribute6Name,
                                                         attribute6Description,
                                                         attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute7Name,
                                                           attribute7Description,
                                                           attribute7DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute8Name,
                                                         attribute8Description,
                                                         attribute8DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute9Name,
                                                         attribute9Description,
                                                         attribute9DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }

    private ClassificationDef getCriticalityClassification()
    {
        ClassificationDef classificationDef = archiveHelper.getClassificationDef(OpenMetadataType.CRITICALITY_CLASSIFICATION.typeGUID,
                                                                                 OpenMetadataType.CRITICALITY_CLASSIFICATION.typeName,
                                                                                 null,
                                                                                 OpenMetadataType.CRITICALITY_CLASSIFICATION.description,
                                                                                 OpenMetadataType.CRITICALITY_CLASSIFICATION.descriptionGUID,
                                                                                 OpenMetadataType.CRITICALITY_CLASSIFICATION.wikiURL,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         OpenMetadataType.REFERENCEABLE.typeName),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "status";
        final String attribute1Description     = "Status of this classification.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = OpenMetadataProperty.CONFIDENCE.name;
        final String attribute2Description     = OpenMetadataProperty.CONFIDENCE.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.CONFIDENCE.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute3Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute4Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute4Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.NOTES.name;
        final String attribute5Description     = OpenMetadataProperty.NOTES.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.NOTES.descriptionGUID;
        final String attribute6Name            = "level";
        final String attribute6Description     = "How critical is this data to the organization.";
        final String attribute6DescriptionGUID = null;


        property = archiveHelper.getEnumTypeDefAttribute("GovernanceClassificationStatus",
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute2Name,
                                                        attribute2Description,
                                                        attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("CriticalityLevel",
                                                         attribute6Name,
                                                         attribute6Description,
                                                         attribute6DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0423 Security Tags define the labels and properties used by rules in a security engine.
     */
    private void add0423SecurityTags()
    {
        this.archiveBuilder.addClassificationDef(getSecurityTagsClassification());
    }

    private ClassificationDef getSecurityTagsClassification()
    {
        final String guid            = "a0b07a86-9fd3-40ca-bb9b-fe83c6981deb";
        final String name            = "SecurityTags";
        final String description     = "Defines labels and properties used by a security engine.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.REFERENCEABLE.typeName;

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(OpenMetadataType.REFERENCEABLE.typeName),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "securityLabels";
        final String attribute1Description     = "Labels that apply to the referenceable.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "securityProperties";
        final String attribute2Description     = "Properties that apply to the referenceable.";
        final String attribute2DescriptionGUID = null;



        property = archiveHelper.getArrayStringTypeDefAttribute(attribute1Name,
                                                                attribute1Description,
                                                                attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringObjectTypeDefAttribute(attribute2Name,
                                                                    attribute2Description,
                                                                    attribute2DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0424 Governance Zones define the zones used to group assets according to their use.
     */
    private void add0424GovernanceZones()
    {
        this.archiveBuilder.addClassificationDef(getAssetZoneMembershipClassification());

        this.archiveBuilder.addEntityDef(getGovernanceZoneEntity());

        this.archiveBuilder.addRelationshipDef(getZoneGovernanceRelationship());
        this.archiveBuilder.addRelationshipDef(getZoneHierarchyRelationship());
    }


    private ClassificationDef getAssetZoneMembershipClassification()
    {
        final String guid            = "a1c17a86-9fd3-40ca-bb9b-fe83c6981deb";
        final String name            = "AssetZoneMembership";
        final String description     = "Defines the asset's membership of the governance zones.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.ASSET.typeName;

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "zoneMembership";
        final String attribute1Description     = "List of governance zones for the asset.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getArrayStringTypeDefAttribute(attribute1Name,
                                                                attribute1Description,
                                                                attribute1DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private EntityDef getGovernanceZoneEntity()
    {
        final String guid            = "290a192b-42a7-449a-935a-269ca62cfdac";
        final String name            = "GovernanceZone";
        final String description     = "Defines a collection of assets that are suitable for a particular usage or are governed by a particular process.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "criteria";
        final String attribute3Description     = "Definition of the types of assets that belong in this zone.";
        final String attribute3DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getZoneGovernanceRelationship()
    {
        final String guid            = "4c4d1d9c-a9fc-4305-8b71-4e891c0f9ae0";
        final String name            = "ZoneGovernance";
        final String description     = "Links a governance zone to a governance definition that applies to all of the members of the zone.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GovernanceZone";
        final String                     end1AttributeName            = "governedZones";
        final String                     end1AttributeDescription     = "The collections of assets governed by this definition.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GovernanceDefinition";
        final String                     end2AttributeName            = "governedBy";
        final String                     end2AttributeDescription     = "Governance definitions for this zone.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);


        return relationshipDef;
    }


    private RelationshipDef getZoneHierarchyRelationship()
    {
        final String guid            = "ee6cf469-cb4d-4c3b-a4c7-e2da1236d139";
        final String name            = "ZoneHierarchy";
        final String description     = "Creates a controlling hierarchy for governance zones.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GovernanceZone";
        final String                     end1AttributeName            = "inheritsFromZone";
        final String                     end1AttributeDescription     = "The zone that provides additional governance requirements.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GovernanceZone";
        final String                     end2AttributeName            = "controlsZone";
        final String                     end2AttributeDescription     = "The zones that are also governed in the same way.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);


        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0430 Technical Controls describe specific types of technical controls
     */
    private void add0430TechnicalControls()
    {
        this.archiveBuilder.addEntityDef(getGovernanceRuleEntity());
        this.archiveBuilder.addEntityDef(getGovernanceProcessEntity());

        this.archiveBuilder.addRelationshipDef(getGovernanceRuleImplementationRelationship());
        this.archiveBuilder.addRelationshipDef(getGovernanceProcessImplementationRelationship());
    }


    private EntityDef getGovernanceRuleEntity()
    {
        final String guid            = "8f954380-12ce-4a2d-97c6-9ebe250fecf8";
        final String name            = "GovernanceRule";
        final String description     = "Technical control expressed as a logic expression.";
        final String descriptionGUID = null;

        final String superTypeName = "TechnicalControl";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getGovernanceProcessEntity()
    {
        final String guid            = "b68b5d9d-6b79-4f3a-887f-ec0f81c54aea";
        final String name            = "GovernanceProcess";
        final String description     = "Technical control expressed as a sequence of tasks.";
        final String descriptionGUID = null;

        final String superTypeName = "TechnicalControl";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getGovernanceRuleImplementationRelationship()
    {
        final String guid            = "e701a5c8-c1ba-4b75-8257-e0a6569eda48";
        final String name            = "GovernanceRuleImplementation";
        final String description     = "Identifies the implementation of a governance rule.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GovernanceRule";
        final String                     end1AttributeName            = "implementsGovernanceRules";
        final String                     end1AttributeDescription     = "The rules that are implemented by this component.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "DeployedSoftwareComponent";
        final String                     end2AttributeName            = "implementations";
        final String                     end2AttributeDescription     = "The software components that implement this governance rule.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NOTES.name;
        final String attribute1Description     = OpenMetadataProperty.NOTES.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NOTES.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getGovernanceProcessImplementationRelationship()
    {
        final String guid            = "a5a7b08a-73fd-4026-a9dd-d0fe55bea8a4";
        final String name            = "GovernanceProcessImplementation";
        final String description     = "Identifies the implementation of a governance process.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GovernanceProcess";
        final String                     end1AttributeName            = "implementsGovernanceProcesses";
        final String                     end1AttributeDescription     = "The processes that are implemented by this component.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.PROCESS.typeName;
        final String                     end2AttributeName            = "implementations";
        final String                     end2AttributeDescription     = "The processes that implement this governance process.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NOTES.name;
        final String attribute1Description     = OpenMetadataProperty.NOTES.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NOTES.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0438 Naming Standards provides definitions for laying our naming standards for schemas and assets.
     */
    private void add0438NamingStandards()
    {
        this.archiveBuilder.addEntityDef(getNamingStandardRuleEntity());
        this.archiveBuilder.addEntityDef(getNamingStandardRuleSetEntity());

        this.archiveBuilder.addClassificationDef(getPrimeWordClassification());
        this.archiveBuilder.addClassificationDef(getClassWordClassification());
        this.archiveBuilder.addClassificationDef(getModifierClassification());
    }


    private EntityDef getNamingStandardRuleEntity()
    {
        final String guid            = "52505b06-98a5-481f-8a32-db9b02afabfc";
        final String name            = "NamingStandardRule";
        final String description     = "Describes a parsing rule used to create compliant names.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernanceRule";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "namePattern";
        final String attribute1Description     = "Format of the naming standard rule.";
        final String attribute1DescriptionGUID = null;


        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);


        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getNamingStandardRuleSetEntity()
    {
        final String guid            = "ba70f506-1f81-4890-bb4f-1cb1d99c939e";
        final String name            = "NamingStandardRuleSet";
        final String description     = "Describes a collection of related naming standard rules.";
        final String descriptionGUID = null;

        final String superTypeName = "Collection";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private ClassificationDef getPrimeWordClassification()
    {
        final String guid            = "3ea1ea66-8923-4662-8628-0bacef3e9c5f";
        final String name            = "PrimeWord";
        final String description     = "Describes a primary noun, used in naming standards.";
        final String descriptionGUID = null;

        final String linkedToEntity = "GlossaryTerm";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getClassWordClassification()
    {
        final String guid            = "feac4bd9-37d9-4437-82f6-618ce3e2793e";
        final String name            = "ClassWord";
        final String description     = "Describes classifying or grouping noun, using in naming standards.";
        final String descriptionGUID = null;

        final String linkedToEntity = "GlossaryTerm";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getModifierClassification()
    {
        final String guid            = "dfc70bed-7e8b-4060-910c-59c7473f23a3";
        final String name            = "NamingConventionRule";
        final String description     = "Describes modifying noun or adverb, used in naming standards.";
        final String descriptionGUID = null;

        final String linkedToEntity = "GlossaryTerm";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0440 Organizational Controls describe organizational roles and responsibilities.
     */
    private void add0440OrganizationalControls()
    {
        this.archiveBuilder.addEnumDef(getBusinessCapabilityTypeEnum());

        this.archiveBuilder.addEntityDef(getOrganizationEntity());
        this.archiveBuilder.addEntityDef(getBusinessCapabilityEntity());
        this.archiveBuilder.addEntityDef(getGovernanceResponsibilityEntity());
        this.archiveBuilder.addEntityDef(getGovernanceProcedureEntity());

        this.archiveBuilder.addClassificationDef(getAssetOriginClassification());

        this.archiveBuilder.addRelationshipDef(getOrganizationCapabilityRelationship());
        this.archiveBuilder.addRelationshipDef(getResponsibilityStaffContactRelationship());
        this.archiveBuilder.addRelationshipDef(getBusinessCapabilityControlsRelationship());
    }


    private EnumDef getBusinessCapabilityTypeEnum()
    {
        final String guid            = "fb7c40cf-8d95-48ff-ba8b-e22bff6f5a91";
        final String name            = "BusinessCapabilityType";
        final String description     = "Defines the type or category of business capability.";
        final String descriptionGUID = null;

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID);

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        final int    element1Ordinal         = 0;
        final String element1Value           = "Unclassified";
        final String element1Description     = "The business capability has not been classified.";
        final String element1DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element1Ordinal,
                                                     element1Value,
                                                     element1Description,
                                                     element1DescriptionGUID);
        elementDefs.add(elementDef);
        enumDef.setDefaultValue(elementDef);

        final int    element2Ordinal         = 1;
        final String element2Value           = "BusinessService";
        final String element2Description     = "A functional business capability.";
        final String element2DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element2Ordinal,
                                                     element2Value,
                                                     element2Description,
                                                     element2DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element3Ordinal         = 2;
        final String element3Value           = "BusinessArea";
        final String element3Description     = "A collection of related business services.";
        final String element3DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element3Ordinal,
                                                     element3Value,
                                                     element3Description,
                                                     element3DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element99Ordinal         = 99;
        final String element99Value           = "Other";
        final String element99Description     = "Another governance definition status.";
        final String element99DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element99Ordinal,
                                                     element99Value,
                                                     element99Description,
                                                     element99DescriptionGUID);
        elementDefs.add(elementDef);

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getOrganizationEntity()
    {
        final String guid            = "50a61105-35be-4ee3-8b99-bdd958ed0685";
        final String name            = "Organization";
        final String description     = "Describes a specific organization.";
        final String descriptionGUID = null;

        final String superTypeName = "Team";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getBusinessCapabilityEntity()
    {
        final String guid            = "7cc6bcb2-b573-4719-9412-cf6c3f4bbb15";
        final String name            = "BusinessCapability";
        final String description     = "Describes a function, capability or skill set.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "type";
        final String attribute1Description     = "Type of business capability.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getEnumTypeDefAttribute("BusinessCapabilityType",
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getGovernanceResponsibilityEntity()
    {
        final String guid            = "89a76b24-deb8-45bf-9304-a578a610326f";
        final String name            = "GovernanceResponsibility";
        final String description     = "Describes a responsibility of a person, team or organization that supports the implementation of a governance driver.";
        final String descriptionGUID = null;

        final String superTypeName = "OrganizationalControl";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getGovernanceProcedureEntity()
    {
        final String guid            = "69055d10-51dc-4c2b-b21f-d76fad3f8ef3";
        final String name            = "GovernanceProcedure";
        final String description     = "Describes set of tasks that a person, team or organization performs to support the implementation of a governance driver.";
        final String descriptionGUID = null;

        final String superTypeName = "OrganizationalControl";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private ClassificationDef getAssetOriginClassification()
    {
        final String guid = "e530c566-03d2-470a-be69-6f52bfbd5fb7";
        final String name            = "AssetOrigin";
        final String description     = "Describes the origin of an asset.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.ASSET.typeName;

        ClassificationDef classificationDef =  archiveHelper.getClassificationDef(guid,
                                                                                  name,
                                                                                  null,
                                                                                  description,
                                                                                  descriptionGUID,
                                                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                                                  false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "organization";
        final String attribute1Description     = "Unique identifier (GUID) of the organization where this asset originated from.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "businessCapability";
        final String attribute2Description     = "Unique identifier (GUID) of the business capability where this asset originated from.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "otherOriginValues";
        final String attribute3Description     = "Descriptive labels describing origin of the asset.";
        final String attribute3DescriptionGUID = null;


        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute3Name,
                                                                    attribute3Description,
                                                                    attribute3DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private RelationshipDef getOrganizationCapabilityRelationship()
    {
        final String guid            = "47f0ad39-db77-41b0-b406-36b1598e0ba7";
        final String name            = "OrganizationalCapability";
        final String description     = "Describes the relationship between a team and the business capabilities it supports.";
        final String descriptionGUID = null;


        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "BusinessCapability";
        final String                     end1AttributeName            = "supportsBusinessCapabilities";
        final String                     end1AttributeDescription     = "The business capabilities that this team supports.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "Team";
        final String                     end2AttributeName            = "supportingTeams";
        final String                     end2AttributeDescription     = "The teams that support this business capability.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getResponsibilityStaffContactRelationship()
    {
        final String guid            = "49f2ecb5-6bf7-4324-9824-ac98d595c404";
        final String name            = "ResponsibilityStaffContact";
        final String description     = "Identifies a person, team or engine assigned to a governance responsibility.";
        final String descriptionGUID = null;


        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GovernanceResponsibility";
        final String                     end1AttributeName            = "contactFor";
        final String                     end1AttributeDescription     = "The governance responsibilities that this team or person is assigned to.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ActorProfile";
        final String                     end2AttributeName            = "assignedStaff";
        final String                     end2AttributeDescription     = "The people, teams and/or engines that are supporting this governance responsibility.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "context";
        final String attribute1Description     = "The context in which this person, team or engine is to be contacted.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getBusinessCapabilityControlsRelationship()
    {
        final String guid            = "b5de932a-738c-4c69-b852-09fec2b9c678";
        final String name            = "BusinessCapabilityControls";
        final String description     = "Identifies a business capability that supports a governance control.";
        final String descriptionGUID = null;


        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GovernanceControl";
        final String                     end1AttributeName            = "implementsControls";
        final String                     end1AttributeDescription     = "The governance controls that this business capability supports.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "BusinessCapability";
        final String                     end2AttributeName            = "affectedBusinessCapabilities";
        final String                     end2AttributeDescription     = "The business capabilities that implement or support this governance responsibility.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "rationale";
        final String attribute1Description     = "Documents reasons for assigning the control to this business capability.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0442ProjectCharter()
    {
        this.archiveBuilder.addEntityDef(getProjectCharterEntity());

        this.archiveBuilder.addRelationshipDef(getProjectCharterLinkRelationship());
    }


    private EntityDef getProjectCharterEntity()
    {
        final String guid            = "f96b5a32-42c1-4a74-8f77-70a81cec783d";
        final String name            = "ProjectCharter";
        final String description     = "Describes the goals, scope and authority of a project.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "mission";
        final String attribute1Description     = "The high-level goal of the project.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "projectType";
        final String attribute2Description     = "Short description of type of the project.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "purposes";
        final String attribute3Description     = "List of purposes for having the project.";
        final String attribute3DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute3Name,
                                                                attribute3Description,
                                                                attribute3DescriptionGUID);
        properties.add(property);


        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getProjectCharterLinkRelationship()
    {
        final String guid            = "f081808d-545a-41cb-a9aa-c4f074a16c78";
        final String name            = "ProjectCharterLink";
        final String description     = "Links a Project with its Charter.";
        final String descriptionGUID = null;


        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Project";
        final String                     end1AttributeName            = "projects";
        final String                     end1AttributeDescription     = "The projects guided by this charter.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ProjectCharter";
        final String                     end2AttributeName            = "charter";
        final String                     end2AttributeDescription     = "The charter guiding this project.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0445GovernanceRoles()
    {
        // todo call is currently being made from 0010 base model
        /* this.archiveBuilder.addEnumDef(getAssetOwnerTypeEnum()); */

        this.archiveBuilder.addClassificationDef(getAssetOwnershipClassification());

        this.archiveBuilder.addEntityDef(getGovernanceRoleEntity());
        this.archiveBuilder.addEntityDef(getAssetOwnerEntity());
        this.archiveBuilder.addEntityDef(getSubjectAreaOwnerEntity());

        this.archiveBuilder.addRelationshipDef(getGovernanceRoleAssignmentRelationship());
        this.archiveBuilder.addRelationshipDef(getGovernanceResponsibilityAssignmentRelationship());
    }


    private EnumDef getAssetOwnerTypeEnum()
    {
        final String guid            = "9548390c-69f5-4dc6-950d-6feeee257b56";
        final String name            = "AssetOwnerType";
        final String description     = "Defines the type of identifier for an asset's owner.";
        final String descriptionGUID = null;

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID);

        List<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef       elementDef;

        final int    element1Ordinal         = 0;
        final String element1Value           = "UserId";
        final String element1Description     = "The owner's userId is specified (default).";
        final String element1DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element1Ordinal,
                                                     element1Value,
                                                     element1Description,
                                                     element1DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element2Ordinal         = 1;
        final String element2Value           = "ProfileId";
        final String element2Description     = "The unique identifier (guid) of the profile of the owner.";
        final String element2DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element2Ordinal,
                                                     element2Value,
                                                     element2Description,
                                                     element2DescriptionGUID);
        elementDefs.add(elementDef);


        final int    element99Ordinal         = 99;
        final String element99Value           = "Other";
        final String element99Description     = "Another type of owner identifier, probably not supported by open metadata.";
        final String element99DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element99Ordinal,
                                                     element99Value,
                                                     element99Description,
                                                     element99DescriptionGUID);
        elementDefs.add(elementDef);

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private ClassificationDef getAssetOwnershipClassification()
    {
        final String guid = "d531c566-03d2-470a-be69-6f52cabd5fb9";
        final String name            = "AssetOwnership";
        final String description     = "Describes the ownership of an asset.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.ASSET.typeName;

        ClassificationDef classificationDef =  archiveHelper.getClassificationDef(guid,
                                                                                  name,
                                                                                  null,
                                                                                  description,
                                                                                  descriptionGUID,
                                                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                                                  false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "owner";
        final String attribute1Description     = "Identifier of the person or process that owns the asset.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "ownerType";
        final String attribute2Description     = "Type of identifier used for owner property.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("AssetOwnerType",
                                                         attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }

    private EntityDef getGovernanceRoleEntity()
    {
        final String guid            = "de2d7f2e-1759-44e3-b8a6-8af53e8fb0ee";
        final String name            = "GovernanceRole";
        final String description     = "Describes a set of goals, tasks and skills that can be assigned a person and contribute to the governance of a resource.";
        final String descriptionGUID = null;

        final String superTypeName = "PersonRole";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getAssetOwnerEntity()
    {
        final String guid            = "ac406bf8-e53e-49f1-9088-2af28eeee285";
        final String name            = "AssetOwner";
        final String description     = "A role defining a responsibility to manage an asset.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernanceRole";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getSubjectAreaOwnerEntity()
    {
        final String guid            = "c6fe40af-cdd6-4ca7-98c4-353d2612921f";
        final String name            = "SubjectAreaOwner";
        final String description     = "A role defining a responsibility to manage the development and maintenance of a subject area.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernanceRole";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getGovernanceRoleAssignmentRelationship()
    {
        final String guid            = "cb10c107-b7af-475d-aab0-d78b8297b982";
        final String name            = "GovernanceRoleAssignment";
        final String description     = "Identifies a person assigned to perform a specific responsibility for a specific resource.";
        final String descriptionGUID = null;


        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "governanceAssignments";
        final String                     end1AttributeDescription     = "The resources assigned to this person.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "PersonRole";
        final String                     end2AttributeName            = "governedByRoles";
        final String                     end2AttributeDescription     = "The roles assigned to this element.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getGovernanceResponsibilityAssignmentRelationship()
    {
        final String guid            = "cb15c107-b7af-475d-aab0-d78b8297b982";
        final String name            = "GovernanceResponsibilityAssignment";
        final String description     = "Identifies a role that will perform a governance responsibility.";
        final String descriptionGUID = null;


        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GovernanceRole";
        final String                     end1AttributeName            = "performedByRoles";
        final String                     end1AttributeDescription     = "The roles assigned to this responsibility.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GovernanceResponsibility";
        final String                     end2AttributeName            = "governanceResponsibilities";
        final String                     end2AttributeDescription     = "The responsibilities performed by this role.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }



    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0450GovernanceRollout()
    {
        this.archiveBuilder.addEntityDef(getGovernanceMetricEntity());

        this.archiveBuilder.addRelationshipDef(getGovernanceDefinitionMetricRelationship());
        this.archiveBuilder.addRelationshipDef(getGovernanceResultsRelationship());

        this.archiveBuilder.addClassificationDef(getGovernanceMeasurementsDataSetClassification());
    }


    private EntityDef getGovernanceMetricEntity()
    {
        final String guid            = "9ada8e7b-823c-40f7-adf8-f164aabda77e";
        final String name            = "GovernanceMetric";
        final String description     = "A definition for how the effectiveness of the governance program is measured.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "measurement";
        final String attribute3Description     = "Format or description of the measurements captured for this metric.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "target";
        final String attribute4Description     = "Definition of the measurement values that the governance definitions are trying to achieve.";
        final String attribute4DescriptionGUID = null;


        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getGovernanceDefinitionMetricRelationship()
    {
        final String guid            = "e076fbb3-54f5-46b8-8f1e-a7cb7e792673";
        final String name            = "GovernanceDefinitionMetric";
        final String description     = "Link between a governance definition and a governance metric used to measure this definition.";
        final String descriptionGUID = null;


        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GovernanceMetric";
        final String                     end1AttributeName            = "metrics";
        final String                     end1AttributeDescription     = "The metrics that measure the landscape against this governance definition.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "GovernanceDefinition";
        final String                     end2AttributeName            = "measuredDefinitions";
        final String                     end2AttributeDescription     = "The governance definitions that are measured by this metric.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "rationale";
        final String attribute1Description     = "Documents reasons for using the metric to measure the governance definition.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getGovernanceResultsRelationship()
    {
        final String guid            = "89c3c695-9e8d-4660-9f44-ed971fd55f88";
        final String name            = "GovernanceResults";
        final String description     = "Link between a governance metric and a data set used to gather measurements from the landscape.";
        final String descriptionGUID = null;


        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "GovernanceMetric";
        final String                     end1AttributeName            = "metrics";
        final String                     end1AttributeDescription     = "The governance metrics that are captured in this data set.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.DATA_SET.typeName;
        final String                     end2AttributeName            = "measurements";
        final String                     end2AttributeDescription     = "The data set that captures the measurements for this governance metric.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "query";
        final String attribute1Description     = "Defines how the data items from the data set are converted in measurements for the metric.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private ClassificationDef getGovernanceMeasurementsDataSetClassification()
    {
        final String guid            = "789f2e89-accd-4489-8eca-dc43b432c022";
        final String name            = "GovernanceMeasurementsResultsDataSet";
        final String description     = "A data file containing measurements for a governance metric.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.DATA_SET.typeName;

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0455ExceptionManagement()
    {
        this.archiveBuilder.addClassificationDef(getExceptionBacklogClassification());
        this.archiveBuilder.addClassificationDef(getAuditLogClassification());
        this.archiveBuilder.addClassificationDef(getMeteringLogClassification());
        this.archiveBuilder.addClassificationDef(getStewardshipServerClassification());
        this.archiveBuilder.addClassificationDef(getGovernanceDaemonClassification());
    }


    private ClassificationDef getExceptionBacklogClassification()
    {
        final String guid            = "b3eceea3-aa02-4d84-8f11-da4953e64b5f";
        final String name            = "ExceptionBacklog";
        final String description     = "A data set containing exceptions that need to be resolved";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.DATA_SET.typeName;

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NOTES.name;
        final String attribute1Description     = OpenMetadataProperty.NOTES.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NOTES.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute2Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute3Name            = "process";
        final String attribute3Description     = "Unique identifier of the automated process that processes this exception backlog.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute4Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private ClassificationDef getAuditLogClassification()
    {
        final String guid            = "449be034-6cc8-4f1b-859f-a8b9ff8ee7a1";
        final String name            = "AuditLog";
        final String description     = "A data set of related audit log records.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.DATA_SET.typeName;

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NOTES.name;
        final String attribute1Description     = OpenMetadataProperty.NOTES.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NOTES.descriptionGUID;
        final String attribute2Name            = "process";
        final String attribute2Description     = "Unique identifier of the automated process that processes this exception backlog.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute3Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private ClassificationDef getMeteringLogClassification()
    {
        final String guid            = "161b37c9-1d51-433b-94ce-5a760a198236";
        final String name            = "MeteringLog";
        final String description     = "A data set containing records that can be used to identify usage of resources.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.DATA_SET.typeName;

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NOTES.name;
        final String attribute1Description     = OpenMetadataProperty.NOTES.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NOTES.descriptionGUID;
        final String attribute2Name            = "process";
        final String attribute2Description     = "Unique identifier of the automated process that processes this exception backlog.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute3Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private ClassificationDef getStewardshipServerClassification()
    {
        final String guid            = "eaaeaa31-6f8b-4ed5-88fe-422ed3733158";
        final String name            = "StewardshipServer";
        final String description     = "A server dedicated to managing stewardship activity relating to governance of data.";
        final String descriptionGUID = null;

        final String linkedToEntity = "SoftwareServer";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getGovernanceDaemonClassification()
    {
        final String guid            = "7815f222-529d-4902-8f0b-e37cbc779885";
        final String name            = "GovernanceDaemon";
        final String description     = "A server dedicated to managing activity relating to governance of data.";
        final String descriptionGUID = null;

        final String linkedToEntity = "SoftwareServer";

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0460GovernanceControls()
    {
        this.archiveBuilder.addClassificationDef(getControlPointClassification());
        this.archiveBuilder.addClassificationDef(getVerificationPointClassification());
        this.archiveBuilder.addClassificationDef(getEnforcementPointClassification());
    }


    private ClassificationDef getControlPointClassification()
    {
        final String guid            = "acf8b73e-3545-435d-ba16-fbfae060dd28";
        final String name            = "ControlPoint";
        final String description     = "A task in a process where a person must make a decision on the right action to take.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.REFERENCEABLE.typeName;

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getVerificationPointClassification()
    {
        final String guid            = "12d78c95-3879-466d-883f-b71f6477a741";
        final String name            = "VerificationPoint";
        final String description     = "A governance rule that tests if a required condition is true or raises an exception if not.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.REFERENCEABLE.typeName;

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private ClassificationDef getEnforcementPointClassification()
    {
        final String guid            = "f4ce104e-7430-4c30-863d-60f6af6394d9";
        final String name            = "EnforcementPoint";
        final String description     = "A governance rule that ensures a required condition is true.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.REFERENCEABLE.typeName;

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0481Licenses()
    {
        this.archiveBuilder.addEntityDef(getLicenseTypeEntity());

        this.archiveBuilder.addRelationshipDef(getLicenseRelationship());
    }


    private EntityDef getLicenseTypeEntity()
    {
        final String guid            = "046a049d-5f80-4e5b-b0ae-f3cf6009b513";
        final String name            = "LicenseType";
        final String description     = "A type of license that sets out specific terms and conditions for the use of an asset.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernanceDefinition";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "details";
        final String attribute1Description     = "Description of the rights, terms and conditions associated with the licence.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getLicenseRelationship()
    {
        final String guid            = "35e53b7f-2312-4d66-ae90-2d4cb47901ee";
        final String name            = "License";
        final String description     = "Link between an asset and its license.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "licensed";
        final String                     end1AttributeDescription     = "Items licensed by this type of license.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "LicenseType";
        final String                     end2AttributeName            = "licenses";
        final String                     end2AttributeDescription     = "The types of licenses that apply.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "licenseGUID";
        final String attribute1Description     = "Unique identifier of the actual license.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "start";
        final String attribute2Description     = "Start date for the license.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "end";
        final String attribute3Description     = "End date for the license.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "conditions";
        final String attribute4Description     = "Any special conditions or endorsements over the basic license type.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "licensedBy";
        final String attribute5Description     = "Person or organization that owns the intellectual property.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "custodian";
        final String attribute6Description     = "The person, engine or organization tht will ensure the license is honored.";
        final String attribute6DescriptionGUID = null;
        final String attribute7Name            = "licensee";
        final String attribute7Description     = "The person or organization that holds the license.";
        final String attribute7DescriptionGUID = null;
        final String attribute8Name            = OpenMetadataProperty.NOTES.name;
        final String attribute8Description     = OpenMetadataProperty.NOTES.description;
        final String attribute8DescriptionGUID = OpenMetadataProperty.NOTES.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute7Name,
                                                           attribute7Description,
                                                           attribute7DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute8Name,
                                                           attribute8Description,
                                                           attribute8DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0482Certifications()
    {
        this.archiveBuilder.addEntityDef(getCertificationTypeEntity());

        this.archiveBuilder.addRelationshipDef(getCertificationRelationship());
        this.archiveBuilder.addRelationshipDef(getRegulationCertificationTypeRelationship());
    }


    private EntityDef getCertificationTypeEntity()
    {
        final String guid            = "97f9ffc9-e2f7-4557-ac12-925257345eea";
        final String name            = "CertificationType";
        final String description     = "A specific type of certification required by a regulation.";
        final String descriptionGUID = null;

        final String superTypeName = "GovernanceDefinition";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "details";
        final String attribute1Description     = "Description of the requirements associated with the certification.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getCertificationRelationship()
    {
        final String guid            = "390559eb-6a0c-4dd7-bc95-b9074caffa7f";
        final String name            = "Certification";
        final String description     = "An awarded certification of a specific type.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "certifies";
        final String                     end1AttributeDescription     = "Items certified by this type of certification.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "CertificationType";
        final String                     end2AttributeName            = "certifications";
        final String                     end2AttributeDescription     = "The types of certifications that apply.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "certificateGUID";
        final String attribute1Description     = "Unique identifier of the actual certificate.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "start";
        final String attribute2Description     = "Start date for the certification.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "end";
        final String attribute3Description     = "End date for the certification.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "conditions";
        final String attribute4Description     = "Any special conditions or endorsements over the basic certification type.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "certifiedBy";
        final String attribute5Description     = "Person or organization awarded the certification.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "custodian";
        final String attribute6Description     = "The person, engine or organization that will ensure the certification is honored.";
        final String attribute6DescriptionGUID = null;
        final String attribute7Name            = "recipient";
        final String attribute7Description     = "The person or organization that received the certification.";
        final String attribute7DescriptionGUID = null;
        final String attribute8Name            = OpenMetadataProperty.NOTES.name;
        final String attribute8Description     = OpenMetadataProperty.NOTES.description;
        final String attribute8DescriptionGUID = OpenMetadataProperty.NOTES.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute7Name,
                                                           attribute7Description,
                                                           attribute7DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute8Name,
                                                           attribute8Description,
                                                           attribute8DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getRegulationCertificationTypeRelationship()
    {
        final String guid            = "be12ff15-0721-4a7e-8c98-334eaa884bdf";
        final String name            = "RegulationCertificationType";
        final String description     = "Identifies a certification required by a regulation.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Regulation";
        final String                     end1AttributeName            = "relatedRegulations";
        final String                     end1AttributeDescription     = "Regulations that require this type of certification.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "CertificationType";
        final String                     end2AttributeName            = "requiredCertifications";
        final String                     end2AttributeDescription     = "The certifications required by this regulation.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * ========================================
     * AREA 5: standards
     */


    /**
     * Area 5 covers information architecture models and schemas.
     */
    private void addArea5Types()
    {
        this.add0501SchemaElements();
        this.add0503AssetSchemas();
        this.add0504ImplementationSnippets();
        this.add0505SchemaAttributes();
        this.add0507BoundedSchemaTypes();
        this.add0510SchemaLinkElements();
        this.add0511SchemaMapElements();
        this.add0512DerivedSchemaAttributes();
        this.add0530TabularSchemas();
        this.add0531DocumentSchemas();
        this.add0532ObjectSchemas();
        this.add0533GraphSchemas();
        this.add0534RelationalSchemas();
        this.add0535EventSchemas();
        this.add0536APISchemas();
        this.add0540DataClasses();
        this.add0545ReferenceData();
        this.add0565DesignModelElements();
        this.add0566DesignModelOrganization();
        this.add0568DesignModelScoping();
        this.add0569DesignModelImplementation();
        this.add0570MetaModel();
        this.add0571ConceptModels();
        this.add0595DesignPatterns();
        this.add0598LineageRelationships();
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0501 Schema Elements define the structure of data is described by a schema.  Schemas are nested structures
     * of schema elements.
     */
    private void add0501SchemaElements()
    {
        this.archiveBuilder.addEntityDef(getSchemaElementEntity());
        this.archiveBuilder.addEntityDef(getSchemaTypeEntity());
        this.archiveBuilder.addEntityDef(getSchemaTypeChoiceEntity());
        this.archiveBuilder.addEntityDef(getLiteralSchemaTypeEntity());
        this.archiveBuilder.addEntityDef(getSimpleSchemaTypeEntity());
        this.archiveBuilder.addEntityDef(getPrimitiveSchemaTypeEntity());
        this.archiveBuilder.addEntityDef(getEnumSchemaTypeEntity());
        this.archiveBuilder.addRelationshipDef(getSchemaTypeOptionRelationship());
    }


    private EntityDef getSchemaElementEntity()
    {
        final String guid            = "718d4244-8559-49ed-ad5a-10e5c305a656";
        final String name            = "SchemaElement";
        final String description     = "An element that is part of a schema definition.";
        final String descriptionGUID = null;
        final String superTypeName   = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "anchorGUID";
        final String attribute1Description     = "Optional identification of the referenceable that this schema element is a part of.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute2Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute3Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;


        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getSchemaTypeEntity()
    {
        final String guid            = "5bd4a3e7-d22d-4a3d-a115-066ee8e0754f";
        final String name            = "SchemaType";
        final String description     = "A specific type description.";
        final String descriptionGUID = null;
        final String superTypeName   = "SchemaElement";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "namespace";
        final String attribute1Description     = "Prefix for element names to ensure uniqueness.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "versionNumber";
        final String attribute2Description     = "Version of the schema type.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "author";
        final String attribute3Description     = "User name of the person or process that created the schema type.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.USAGE.name;
        final String attribute4Description     = OpenMetadataProperty.USAGE.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.USAGE.descriptionGUID;
        final String attribute5Name            = "encodingStandard";
        final String attribute5Description     = "Format of the schema.";
        final String attribute5DescriptionGUID = null;



        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getSchemaTypeChoiceEntity()
    {
        final String guid            = "5caf954a-3e33-4cbd-b17d-8b8613bd2db8";
        final String name            = "SchemaTypeChoice";
        final String description     = "A list of alternative schema types for attribute.";
        final String descriptionGUID = null;
        final String superTypeName   = "SchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getLiteralSchemaTypeEntity()
    {
        final String guid            = "520ebb91-c4eb-4d46-a3b1-974875cdcf0d";
        final String name            = "LiteralSchemaType";
        final String description     = "A fixed simple value.";
        final String descriptionGUID = null;
        final String superTypeName   = "SchemaType";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "dataType";
        final String attribute1Description     = "Type name for the data stored in this schema element.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "fixedValue";
        final String attribute2Description     = "Fixed value for data stored in this schema element.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getSimpleSchemaTypeEntity()
    {
        final String guid            = "b5ec6e07-6419-4225-9dc4-fb55aba255c6";
        final String name            = "SimpleSchemaType";
        final String description     = "A single valued type.";
        final String descriptionGUID = null;
        final String superTypeName   = "SchemaType";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "dataType";
        final String attribute1Description     = "Type name for the data stored in this schema element.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "defaultValue";
        final String attribute2Description     = "Initial value for data stored in this schema element.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getPrimitiveSchemaTypeEntity()
    {
        final String guid            = "f0f75fba-9136-4082-8352-0ad74f3c36ed";
        final String name            = "PrimitiveSchemaType";
        final String description     = "A specific primitive type.";
        final String descriptionGUID = null;
        final String superTypeName   = "SimpleSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getEnumSchemaTypeEntity()
    {
        final String guid            = "24b092ac-42e9-43dc-aeca-eb034ce307d9";
        final String name            = "EnumSchemaType";
        final String description     = "A single valued type with fixed list of valid values.";
        final String descriptionGUID = null;
        final String superTypeName   = "SimpleSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getSchemaTypeOptionRelationship()
    {
        final String guid            = "eb4f1f98-c649-4560-8a46-da17c02764a9";
        final String name            = "SchemaTypeOption";
        final String description     = "The list of alternative schema types.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "SchemaTypeChoice";
        final String                     end1AttributeName            = "schemaOptionalUses";
        final String                     end1AttributeDescription     = "Potential place where this schema type is used.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SchemaType";
        final String                     end2AttributeName            = "schemaOptions";
        final String                     end2AttributeDescription     = "Possible structure of the content of this element.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0503 Asset Schemas shows how assets are linked to schemas
     */
    private void add0503AssetSchemas()
    {
        this.archiveBuilder.addRelationshipDef(getAssetSchemaTypeRelationship());
    }


    private RelationshipDef getAssetSchemaTypeRelationship()
    {
        final String guid            = "815b004d-73c6-4728-9dd9-536f4fe803cd";
        final String name            = "AssetSchemaType";
        final String description     = "The structure of an asset.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.ASSET.typeName;
        final String                     end1AttributeName            = "describesAssets";
        final String                     end1AttributeDescription     = "Asset that conforms to the schema type.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SchemaType";
        final String                     end2AttributeName            = "schema";
        final String                     end2AttributeDescription     = "Structure of the content of this asset.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0504 Implementation snippets enable code fragments defining data structures to be linked with the
     * relevant schema to show how the schema should be implemented.
     */
    private void add0504ImplementationSnippets()
    {
        this.archiveBuilder.addEntityDef(getImplementationSnippetEntity());

        this.archiveBuilder.addRelationshipDef(getSchemaTypeSnippetRelationship());
        this.archiveBuilder.addRelationshipDef(getSchemaTypeImplementationRelationship());
    }


    private EntityDef getImplementationSnippetEntity()
    {
        final String guid            = "49990755-2faa-4a62-a1f3-9124b9c73df4";
        final String name            = "ImplementationSnippet";
        final String description     = "A concrete implementation example for a schema element.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "snippet";
        final String attribute1Description     = "Concrete implementation of the schema type.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "implementationLanguage";
        final String attribute2Description     = "Type of implementation.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "version";
        final String attribute3Description     = "Version number of the snippet.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.USAGE.name;
        final String attribute4Description     = OpenMetadataProperty.USAGE.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.USAGE.descriptionGUID;
        final String attribute5Name            = "curator";
        final String attribute5Description     = "User name of the person or process that is maintaining the snippet.";
        final String attribute5DescriptionGUID = null;


        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getSchemaTypeSnippetRelationship()
    {
        final String guid            = "6aab4ec6-f0c6-4c40-9f50-ac02a3483358";
        final String name            = "SchemaTypeSnippet";
        final String description     = "Link between a schema type and an implementation snippet.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "SchemaType";
        final String                     end1AttributeName            = "templateForSchemaTypes";
        final String                     end1AttributeDescription     = "Logical structure for data.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ImplementationSnippet";
        final String                     end2AttributeName            = "implementationSnippets";
        final String                     end2AttributeDescription     = "Template implementation of the schema type.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getSchemaTypeImplementationRelationship()
    {
        final String guid            = "eed5565d-7ac2-46fe-9a26-4722fad8d993";
        final String name            = "SchemaTypeImplementation";
        final String description     = "Link between a schema type and an implementation.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "SchemaType";
        final String                     end1AttributeName            = "implementationSchemaTypes";
        final String                     end1AttributeDescription     = "Logical structure for the data.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.PROCESS.typeName;
        final String                     end2AttributeName            = "implementations";
        final String                     end2AttributeDescription     = "Concrete implementation of the schema type.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0505 Schema Attributes begins to fill out the structure of a schema.
     */
    private void add0505SchemaAttributes()
    {
        this.archiveBuilder.addEnumDef(getDataItemSortOrderEnum());

        this.archiveBuilder.addEntityDef(getSchemaAttributeEntity());
        this.archiveBuilder.addEntityDef(getComplexSchemaTypeEntity());
        this.archiveBuilder.addEntityDef(getStructSchemaTypeEntity());

        this.archiveBuilder.addRelationshipDef(getAttributeForSchemaRelationship());
        this.archiveBuilder.addRelationshipDef(getSchemaAttributeTypeRelationship());
        this.archiveBuilder.addRelationshipDef(getNestedSchemaAttributeRelationship());

        this.archiveBuilder.addClassificationDef(getTypeEmbeddedAttributeClassification());
    }

    private EnumDef getDataItemSortOrderEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(DataItemSortOrder.getOpenTypeGUID(),
                                                        DataItemSortOrder.getOpenTypeName(),
                                                        DataItemSortOrder.getOpenTypeDescription(),
                                                        DataItemSortOrder.getOpenTypeDescriptionGUID(),
                                                        DataItemSortOrder.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (DataItemSortOrder enumValues : DataItemSortOrder.values())
        {
            elementDef = archiveHelper.getEnumElementDef(enumValues.getOrdinal(),
                                                         enumValues.getName(),
                                                         enumValues.getDescription(),
                                                         enumValues.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (enumValues.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getSchemaAttributeEntity()
    {
        EntityDef entityDef = archiveHelper.getDefaultEntityDef(OpenMetadataType.SCHEMA_ATTRIBUTE.typeGUID,
                                                                OpenMetadataType.SCHEMA_ATTRIBUTE.typeName,
                                                                this.archiveBuilder.getEntityDef(OpenMetadataType.SCHEMA_ELEMENT_TYPE_NAME),
                                                                OpenMetadataType.SCHEMA_ATTRIBUTE.description,
                                                                OpenMetadataType.SCHEMA_ATTRIBUTE.descriptionGUID,
                                                                OpenMetadataType.SCHEMA_ATTRIBUTE.wikiURL);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "position";
        final String attribute1Description     = "Location of the attribute in the parent schema's list of attributes, starting at zero.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "minCardinality";
        final String attribute2Description     = "Minimum number of occurrences of this attribute allowed (0 = optional, 1+ = mandatory).";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "maxCardinality";
        final String attribute3Description     = "Maximum number of occurrences of this attribute allowed (-1 = infinite).";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "allowsDuplicateValues";
        final String attribute4Description     = "When multiple occurrences are allowed, indicates whether duplicates of the same value are allowed or not.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "orderedValues";
        final String attribute5Description     = "When multiple occurrences are allowed, indicates whether the values are ordered or not.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "defaultValueOverride";
        final String attribute6Description     = "Initial value for the attribute (overriding the default value of its type.";
        final String attribute6DescriptionGUID = null;
        final String attribute7Name            = "nativeClass";
        final String attribute7Description     = "Native class used by the client to represent this element.";
        final String attribute7DescriptionGUID = null;
        final String attribute8Name            = "name";
        final String attribute8Description     = "Name of schema attribute (deprecated).";
        final String attribute8DescriptionGUID = null;
        final String attribute9Name            = "aliases";
        final String attribute9Description     = "List of aliases for attribute.";
        final String attribute9DescriptionGUID = null;
        final String attribute10Name            = "sortOrder";
        final String attribute10Description     = "Suggested ordering of values in this attribute.";
        final String attribute10DescriptionGUID = null;

        property = archiveHelper.getIntTypeDefAttribute(attribute1Name,
                                                        attribute1Description,
                                                        attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute2Name,
                                                        attribute2Description,
                                                        attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute3Name,
                                                        attribute3Description,
                                                        attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getBooleanTypeDefAttribute(attribute4Name,
                                                            attribute4Description,
                                                            attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getBooleanTypeDefAttribute(attribute5Name,
                                                            attribute5Description,
                                                            attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute7Name,
                                                           attribute7Description,
                                                           attribute7DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute8Name,
                                                           attribute8Description,
                                                           attribute8DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute9Name,
                                                                attribute9Description,
                                                                attribute9DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("DataItemSortOrder",
                                                         attribute10Name,
                                                         attribute10Description,
                                                         attribute10DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getComplexSchemaTypeEntity()
    {
        final String guid            = "786a6199-0ce8-47bf-b006-9ace1c5510e4";
        final String name            = "ComplexSchemaType";
        final String description     = "A schema type that has a complex structure of nested attributes and types.";
        final String descriptionGUID = null;

        final String superTypeName = "SchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getStructSchemaTypeEntity()
    {
        final String guid            = "a13b409f-fd67-4506-8d94-14dfafd250a4";
        final String name            = "StructSchemaType";
        final String description     = "A schema type that has a list of attributes, typically of different types.";
        final String descriptionGUID = null;

        final String superTypeName = "ComplexSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getAttributeForSchemaRelationship()
    {
        final String guid            = "86b176a2-015c-44a6-8106-54d5d69ba661";
        final String name            = "AttributeForSchema";
        final String description     = "Link between a complex schema type and its attributes.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "ComplexSchemaType";
        final String                     end1AttributeName            = "parentSchemas";
        final String                     end1AttributeDescription     = "Schema types using this attribute.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;
        final String                     end2AttributeName            = "attributes";
        final String                     end2AttributeDescription     = "The attributes defining the internal structure of the schema type.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getSchemaAttributeTypeRelationship()
    {
        final String guid            = "2d955049-e59b-45dd-8e62-cde1add59f9e";
        final String name            = "SchemaAttributeType";
        final String description     = "The schema type for an attribute.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;
        final String                     end1AttributeName            = "usedInSchemas";
        final String                     end1AttributeDescription     = "Occurrences of this schema type in other schemas.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SchemaType";
        final String                     end2AttributeName            = "type";
        final String                     end2AttributeDescription     = "The structure of this attribute.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getNestedSchemaAttributeRelationship()
    {
        final String guid            = "0ffb9d87-7074-45da-a9b0-ae0859611133";
        final String name            = "NestedSchemaAttribute";
        final String description     = "The direct parent-child relationship between attributes with an embedded type.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;
        final String                     end1AttributeName            = "parentAttribute";
        final String                     end1AttributeDescription     = "Schema attribute containing this attribute.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                end1AttributeName,
                end1AttributeDescription,
                end1AttributeDescriptionGUID,
                end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;
        final String                     end2AttributeName            = "nestedAttributes";
        final String                     end2AttributeDescription     = "The attributes defining the internal structure of the parent attribute.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                end2AttributeName,
                end2AttributeDescription,
                end2AttributeDescriptionGUID,
                end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private ClassificationDef getTypeEmbeddedAttributeClassification()
    {
        final String guid            = "e2bb76bb-774a-43ff-9045-3a05f663d5d9";
        final String name            = "TypeEmbeddedAttribute";
        final String description     = "Type information embedded within an attribute.";
        final String descriptionGUID = null;

        final String linkedToEntity = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "encodingStandard";
        final String attribute1Description     = "Format of the schema.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "dataType";
        final String attribute2Description     = "Type name for the data stored in this schema element.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0507 Bounded Schema Types adds more types of schema.
     */
    private void add0507BoundedSchemaTypes()
    {
        this.archiveBuilder.addEntityDef(getBoundedSchemaTypeEntity());
        this.archiveBuilder.addEntityDef(getArraySchemaTypeEntity());
        this.archiveBuilder.addEntityDef(getSetSchemaTypeEntity());

        this.archiveBuilder.addRelationshipDef(getBoundedSchemaElementTypeRelationship());
    }


    private EntityDef getBoundedSchemaTypeEntity()
    {
        final String guid            = "77133161-37a9-43f5-aaa3-fd6d7ff92fdb";
        final String name            = "BoundedSchemaType";
        final String description     = "A schema type that limits the number of values that can be stored.";
        final String descriptionGUID = null;

        final String superTypeName = "ComplexSchemaType";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "maximumElements";
        final String attribute1Description     = "Maximum number of values that can be stored - zero for no limit.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getIntTypeDefAttribute(attribute1Name,
                                                        attribute1Description,
                                                        attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getArraySchemaTypeEntity()
    {
        final String guid            = "ba8d29d2-a8a4-41f3-b29f-91ad924dd944";
        final String name            = "ArraySchemaType";
        final String description     = "A schema type that has a list of values of the same type.";
        final String descriptionGUID = null;

        final String superTypeName = "BoundedSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getSetSchemaTypeEntity()
    {
        final String guid            = "b2605d2d-10cd-443c-b3e8-abf15fb051f0";
        final String name            = "SetSchemaType";
        final String description     = "A schema type that is an unordered group of values of the same type.";
        final String descriptionGUID = null;

        final String superTypeName = "BoundedSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getBoundedSchemaElementTypeRelationship()
    {
        final String guid            = "3e844049-e59b-45dd-8e62-cde1add59f9e";
        final String name            = "BoundedSchemaElementType";
        final String description     = "The type of the element within the bounded schema type.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "BoundedSchemaType";
        final String                     end1AttributeName            = "usedInBoundedSchemaType";
        final String                     end1AttributeDescription     = "Use of this element as part of a more complex type.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SchemaType";
        final String                     end2AttributeName            = "boundedSchemaElementType";
        final String                     end2AttributeDescription     = "The structure of the element within this type.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0510 Schema Link Elements defines how a link between two schema elements in different schema element
     * hierarchies are linked.
     */
    private void add0510SchemaLinkElements()
    {
        this.archiveBuilder.addEntityDef(getSchemaLinkElementEntity());

        this.archiveBuilder.addRelationshipDef(getLinkedTypeRelationship());
        this.archiveBuilder.addRelationshipDef(getSchemaLinkToTypeRelationship());
    }


    private EntityDef getSchemaLinkElementEntity()
    {
        final String guid            = "67e08705-2d2a-4df6-9239-1818161a41e0";
        final String name            = "SchemaLinkElement";
        final String description     = "A link to a type in a different schema.";
        final String descriptionGUID = null;
        final String superTypeName   = "SchemaElement";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "linkName";
        final String attribute1Description     = "Name for the element that bridges between two schemas.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "linkProperties";
        final String attribute2Description     = "Any options needed to describe the link.";
        final String attribute2DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute2Name,
                                                                    attribute2Description,
                                                                    attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getLinkedTypeRelationship()
    {
        final String guid            = "292125f7-5660-4533-a48a-478c5611922e";
        final String name            = "LinkedType";
        final String description     = "Link between a link element and its type.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "SchemaLinkElement";
        final String                     end1AttributeName            = "linkedBy";
        final String                     end1AttributeDescription     = "External links to this type.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SchemaType";
        final String                     end2AttributeName            = "linkedType";
        final String                     end2AttributeDescription     = "Types for this attribute.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getSchemaLinkToTypeRelationship()
    {
        final String guid            = "db9583c5-4690-41e5-a580-b4e30a0242d3";
        final String name            = "SchemaLinkToType";
        final String description     = "Link between a schema attribute and a schema link.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;
        final String                     end1AttributeName            = "usedIn";
        final String                     end1AttributeDescription     = "Attributes of this type.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SchemaLinkElement";
        final String                     end2AttributeName            = "externalType";
        final String                     end2AttributeDescription     = "External type for this attribute.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0511 Schema Map Elements defines how a map element is recorded in a schema.
     */
    private void add0511SchemaMapElements()
    {
        this.archiveBuilder.addEntityDef(getMapSchemaTypeEntity());

        this.archiveBuilder.addRelationshipDef(getMapFromElementTypeRelationship());
        this.archiveBuilder.addRelationshipDef(getMapToElementTypeRelationship());
    }


    private EntityDef getMapSchemaTypeEntity()
    {
        final String guid            = "bd4c85d0-d471-4cd2-a193-33b0387a19fd";
        final String name            = "MapSchemaType";
        final String description     = "A schema type for a map between a key and value.";
        final String descriptionGUID = null;

        final String superTypeName = "SchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getMapFromElementTypeRelationship()
    {
        final String guid            = "6189d444-2da4-4cd7-9332-e48a1c340b44";
        final String name            = "MapFromElementType";
        final String description     = "Defines the type of the key for a map schema type.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "MapSchemaType";
        final String                     end1AttributeName            = "parentMapFrom";
        final String                     end1AttributeDescription     = "Used in map.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SchemaType";
        final String                     end2AttributeName            = "mapFromElement";
        final String                     end2AttributeDescription     = "Key for this attribute.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getMapToElementTypeRelationship()
    {
        final String guid            = "8b9856b3-451e-45fc-afc7-fddefd81a73a";
        final String name            = "MapToElementType";
        final String description     = "Defines the type of value for a map schema type.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "MapSchemaType";
        final String                     end1AttributeName            = "parentMapTo";
        final String                     end1AttributeDescription     = "Used in map.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SchemaType";
        final String                     end2AttributeName            = "mapToElement";
        final String                     end2AttributeDescription     = "Value for this map.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }

    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0512 Derived Schema Attributes defines how one attribute can be derived from another.
     */
    private void add0512DerivedSchemaAttributes()
    {
        this.archiveBuilder.addEntityDef(getDerivedSchemaAttributeEntity());

        this.archiveBuilder.addRelationshipDef(getSchemaQueryImplementationRelationship());
    }


    private EntityDef getDerivedSchemaAttributeEntity()
    {
        final String guid            = "cf21abfe-655a-47ba-b9b6-f73394745c80";
        final String name            = "DerivedSchemaAttribute";
        final String description     = "An attribute that is made up of values from another attribute.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "formula";
        final String attribute1Description     = "Transformation used to create the derived data.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "comment";
        final String attribute2Description     = "Comment from source system (deprecated).";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "id";
        final String attribute3Description     = "Id of derived schema attribute (deprecated).";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "aggregatingFunction";
        final String attribute4Description     = "Aggregating function of derived schema attribute (deprecated).";
        final String attribute4DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getSchemaQueryImplementationRelationship()
    {
        final String guid            = "e5d7025d-8b4f-43c7-bcae-1047d650b94a";
        final String name            = "SchemaQueryImplementation";
        final String description     = "Details of how a derived schema attribute is calculated.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "DerivedSchemaAttribute";
        final String                     end1AttributeName            = "usedBy";
        final String                     end1AttributeDescription     = "Use of an attribute to derive another attribute.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;
        final String                     end2AttributeName            = "queryTarget";
        final String                     end2AttributeDescription     = "Used to derive this attribute.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "query";
        final String attribute1Description     = "Details of how the attribute is retrieved.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0530 describes table oriented schemas such as spreadsheets
     */
    private void add0530TabularSchemas()
    {
        this.archiveBuilder.addEntityDef(getTabularSchemaTypeEntity());
        this.archiveBuilder.addEntityDef(getTabularColumnTypeEntity());
        this.archiveBuilder.addEntityDef(getTabularColumnEntity());
    }


    private EntityDef getTabularSchemaTypeEntity()
    {
        final String guid            = "248975ec-8019-4b8a-9caf-084c8b724233";
        final String name            = "TabularSchemaType";
        final String description     = "A schema type for a table oriented data structure.";
        final String descriptionGUID = null;

        final String superTypeName = "ComplexSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getTabularColumnTypeEntity()
    {
        final String guid            = "a7392281-348d-48a4-bad7-f9742d7696fe";
        final String name            = "TabularColumnType";
        final String description     = "A schema type for a column oriented data structure.";
        final String descriptionGUID = null;

        final String superTypeName = "PrimitiveSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getTabularColumnEntity()
    {
        final String guid            = "d81a0425-4e9b-4f31-bc1c-e18c3566da10";
        final String name            = "TabularColumn";
        final String description     = "A column attribute for a table oriented data structure.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0531 Document Schema provide specialized entities for describing document oriented schemas such as JSON
     */
    private void add0531DocumentSchemas()
    {
        this.archiveBuilder.addEntityDef(getDocumentSchemaTypeEntity());
        this.archiveBuilder.addEntityDef(getDocumentSchemaAttributeEntity());
        this.archiveBuilder.addEntityDef(getSimpleDocumentTypeEntity());
        this.archiveBuilder.addEntityDef(getStructDocumentTypeEntity());
        this.archiveBuilder.addEntityDef(getArrayDocumentTypeEntity());
        this.archiveBuilder.addEntityDef(getSetDocumentTypeEntity());
        this.archiveBuilder.addEntityDef(getMapDocumentTypeEntity());
    }


    private EntityDef getDocumentSchemaTypeEntity()
    {
        final String guid            = "33da99cd-8d04-490c-9457-c58908da7794";
        final String name            = "DocumentSchemaType";
        final String description     = "A schema type for a hierarchical data structure.";
        final String descriptionGUID = null;

        final String superTypeName = "ComplexSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getDocumentSchemaAttributeEntity()
    {
        final String guid            = "b5cefb7e-b198-485f-a1d7-8e661012499b";
        final String name            = "DocumentSchemaAttribute";
        final String description     = "A schema attribute for a hierarchical data structure.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getSimpleDocumentTypeEntity()
    {
        final String guid            = "42cfccbf-cc68-4980-8c31-0faf1ee002d3";
        final String name            = "SimpleDocumentType";
        final String description     = "A primitive attribute for a hierarchical data structure.";
        final String descriptionGUID = null;

        final String superTypeName = "PrimitiveSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getStructDocumentTypeEntity()
    {
        final String guid            = "f6245c25-8f73-45eb-8fb5-fa17a5f27649";
        final String name            = "StructDocumentType";
        final String description     = "A structure within a hierarchical data structure.";
        final String descriptionGUID = null;

        final String superTypeName = "StructSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getArrayDocumentTypeEntity()
    {
        final String guid            = "ddd29c67-db9a-45ff-92aa-6d17a12a8ee2";
        final String name            = "ArrayDocumentType";
        final String description     = "An array in a hierarchical data structure.";
        final String descriptionGUID = null;

        final String superTypeName = "ArraySchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getSetDocumentTypeEntity()
    {
        final String guid            = "67228a7a-9d8d-4fa7-b217-17474f1f4ac6";
        final String name            = "SetDocumentType";
        final String description     = "A set in a hierarchical data structure.";
        final String descriptionGUID = null;

        final String superTypeName = "SetSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getMapDocumentTypeEntity()
    {
        final String guid            = "b0f09598-ceb6-415b-befc-563ecadd5727";
        final String name            = "MapDocumentType";
        final String description     = "A map in a hierarchical data structure.";
        final String descriptionGUID = null;

        final String superTypeName = "MapSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0532 Object Schemas provides the specialized entity for an object schema.
     */
    private void add0532ObjectSchemas()
    {
        this.archiveBuilder.addEntityDef(getObjectSchemaTypeEntity());
        this.archiveBuilder.addEntityDef(getObjectAttributeEntity());
    }


    private EntityDef getObjectSchemaTypeEntity()
    {
        final String guid            = "6920fda1-7c07-47c7-84f1-9fb044ae153e";
        final String name            = "ObjectSchemaType";
        final String description     = "A schema attribute for an object.";
        final String descriptionGUID = null;

        final String superTypeName = "ComplexSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getObjectAttributeEntity()
    {
        final String guid            = "ccb408c0-582e-4a3a-a926-7082d53bb669";
        final String name            = "ObjectAttribute";
        final String description     = "An attribute in an object schema type.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0533GraphSchemas()
    {
        this.archiveBuilder.addEntityDef(getGraphSchemaTypeEntity());
        this.archiveBuilder.addEntityDef(getGraphVertexEntity());
        this.archiveBuilder.addEntityDef(getGraphEdgeEntity());

        this.archiveBuilder.addRelationshipDef(getGraphEdgeLinkRelationship());

    }


    private EntityDef getGraphSchemaTypeEntity()
    {
        final String guid            = OpenMetadataType.GRAPH_SCHEMA_TYPE.typeGUID;
        final String name            = OpenMetadataType.GRAPH_SCHEMA_TYPE.typeName;
        final String description     = OpenMetadataType.GRAPH_SCHEMA_TYPE.description;
        final String descriptionGUID = OpenMetadataType.GRAPH_SCHEMA_TYPE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.GRAPH_SCHEMA_TYPE.wikiURL;

        final String superTypeName = OpenMetadataType.SCHEMA_TYPE_TYPE_NAME;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    private EntityDef getGraphVertexEntity()
    {
        final String guid            = OpenMetadataType.GRAPH_VERTEX.typeGUID;
        final String name            = OpenMetadataType.GRAPH_VERTEX.typeName;
        final String description     = OpenMetadataType.GRAPH_VERTEX.description;
        final String descriptionGUID = OpenMetadataType.GRAPH_VERTEX.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.GRAPH_VERTEX.wikiURL;

        final String superTypeName = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    private EntityDef getGraphEdgeEntity()
    {
        final String guid            = OpenMetadataType.GRAPH_EDGE.typeGUID;
        final String name            = OpenMetadataType.GRAPH_EDGE.typeName;
        final String description     = OpenMetadataType.GRAPH_EDGE.description;
        final String descriptionGUID = OpenMetadataType.GRAPH_EDGE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.GRAPH_EDGE.wikiURL;

        final String superTypeName = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    private RelationshipDef getGraphEdgeLinkRelationship()
    {
        final String guid            = OpenMetadataType.GRAPH_EDGE_LINK_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.GRAPH_EDGE_LINK_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.GRAPH_EDGE_LINK_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.GRAPH_EDGE_LINK_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.GRAPH_EDGE_LINK_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.GRAPH_EDGE.typeName;
        final String                     end1AttributeName            = "edges";
        final String                     end1AttributeDescription     = "Edges for this vertex.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.GRAPH_VERTEX.typeName;
        final String                     end2AttributeName            = "vertices";
        final String                     end2AttributeDescription     = "Vertices for this edge.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.LINK_TYPE_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.LINK_TYPE_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.LINK_TYPE_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.RELATIONSHIP_END.name;
        final String attribute2Description     = OpenMetadataProperty.RELATIONSHIP_END.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.RELATIONSHIP_END.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.RELATIONSHIP_END_NAME.name;
        final String attribute3Description     = OpenMetadataProperty.RELATIONSHIP_END_NAME.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.RELATIONSHIP_END_NAME.descriptionGUID;
        final String attribute4Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute4Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.MIN_CARDINALITY.name;
        final String attribute5Description     = OpenMetadataProperty.MIN_CARDINALITY.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.MIN_CARDINALITY.descriptionGUID;
        final String attribute6Name            = OpenMetadataProperty.MAX_CARDINALITY.name;
        final String attribute6Description     = OpenMetadataProperty.MAX_CARDINALITY.description;
        final String attribute6DescriptionGUID = OpenMetadataProperty.MAX_CARDINALITY.descriptionGUID;
        final String attribute7Name            = OpenMetadataProperty.ADDITIONAL_PROPERTIES.name;
        final String attribute7Description     = OpenMetadataProperty.ADDITIONAL_PROPERTIES.description;
        final String attribute7DescriptionGUID = OpenMetadataProperty.ADDITIONAL_PROPERTIES.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute2Name,
                                                        attribute2Description,
                                                        attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute5Name,
                                                        attribute5Description,
                                                        attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute6Name,
                                                        attribute6Description,
                                                        attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute7Name,
                                                                    attribute7Description,
                                                                    attribute7DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0534 Relational Schemas extend the tabular schemas to describe a relational database.
     */
    private void add0534RelationalSchemas()
    {
        this.archiveBuilder.addEntityDef(getRelationalDBSchemaTypeEntity());
        this.archiveBuilder.addEntityDef(getRelationalTableTypeEntity());
        this.archiveBuilder.addEntityDef(getRelationalTableEntity());
        this.archiveBuilder.addEntityDef(getRelationalColumnEntity());
        this.archiveBuilder.addEntityDef(getRelationalColumnTypeEntity());
        this.archiveBuilder.addEntityDef(getDerivedRelationalColumnEntity());

        this.archiveBuilder.addClassificationDef(getPrimaryKeyClassification());
        this.archiveBuilder.addClassificationDef(getRelationalViewClassification());

        this.archiveBuilder.addRelationshipDef(getForeignKeyRelationship());
    }


    private EntityDef getRelationalDBSchemaTypeEntity()
    {
        final String guid            = "f20f5f45-1afb-41c1-9a09-34d8812626a4";
        final String name            = "RelationalDBSchemaType";
        final String description     = "A schema type for a relational database.";
        final String descriptionGUID = null;

        final String superTypeName = "ComplexSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getRelationalTableTypeEntity()
    {
        final String guid            = "1321bcc0-dc6a-48ed-9ca6-0c6f934b0b98";
        final String name            = "RelationalTableType";
        final String description     = "A table type for a relational database.";
        final String descriptionGUID = null;

        final String superTypeName = "TabularSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getRelationalTableEntity()
    {
        final String guid            = "ce7e72b8-396a-4013-8688-f9d973067425";
        final String name            = "RelationalTable";
        final String description     = "A table within a relational database schema type.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getRelationalColumnEntity()
    {
        final String guid            = "aa8d5470-6dbc-4648-9e2f-045e5df9d2f9";
        final String name            = "RelationalColumn";
        final String description     = "A column within a relational table.";
        final String descriptionGUID = null;

        final String superTypeName = "TabularColumn";

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);
        
        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;
        
        final String attribute3Name            = "fraction";
        final String attribute3Description     = "Number of significant digits to the right of decimal point.";
        final String attribute3DescriptionGUID = null;
        final String attribute5Name            = "isUnique";
        final String attribute5Description     = "Data is unique or not.";
        final String attribute5DescriptionGUID = null;
        
        property = archiveHelper.getIntTypeDefAttribute(attribute3Name,
                                                        attribute3Description,
                                                        attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getBooleanTypeDefAttribute(attribute5Name,
                                                            attribute5Description,
                                                            attribute5DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getRelationalColumnTypeEntity()
    {
        final String guid            = "f0438d80-6eb9-4fac-bcc1-5efee5babcfc";
        final String name            = "RelationalColumnType";
        final String description     = "A type for a relational column.";
        final String descriptionGUID = null;

        final String superTypeName = "TabularColumnType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getDerivedRelationalColumnEntity()
    {
        final String guid            = "a9f7d15d-b797-450a-8d56-1ba55490c019";
        final String name            = "DerivedRelationalColumn";
        final String description     = "A relational column that is derived from other columns.";
        final String descriptionGUID = null;

        final String superTypeName = "DerivedSchemaAttribute";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private ClassificationDef getPrimaryKeyClassification()
    {
        final String guid            = "b239d832-50bd-471b-b17a-15a335fc7f40";
        final String name            = "PrimaryKey";
        final String description     = "A uniquely identifying relational column.";
        final String descriptionGUID = null;

        final String linkedToEntity = "RelationalColumn";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(
                                                                                         linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.KEY_PATTERN.name;
        final String attribute1Description     = OpenMetadataProperty.KEY_PATTERN.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.KEY_PATTERN.descriptionGUID;


        final String attribute2Name            = OpenMetadataProperty.NAME.name;
        final String attribute2Description     = OpenMetadataProperty.NAME.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;

        property = archiveHelper.getEnumTypeDefAttribute("KeyPattern",
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);

        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    private ClassificationDef getRelationalViewClassification()
    {
        final String guid            = "4814bec8-482d-463d-8376-160b0358e129";
        final String name            = "RelationalView";
        final String description     = "A view within a relational database schema type.";
        final String descriptionGUID = null;

        final String linkedToEntity = "RelationalTable";

        ClassificationDef classificationDef =  archiveHelper.getClassificationDef(guid,
                                                                                  name,
                                                                                  null,
                                                                                  description,
                                                                                  descriptionGUID,
                                                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                                                  false);
        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "expression";
        final String attribute1Description     = "Expression of the view.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;          
    }


    private RelationshipDef getForeignKeyRelationship()
    {
        final String guid            = "3cd4e0e7-fdbf-47a6-ae88-d4b3205e0c07";
        final String name            = "ForeignKey";
        final String description     = "The primary key for another column is stored in a relational column from another table to enable them to be joined.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "RelationalColumn";
        final String                     end1AttributeName            = "primaryKey";
        final String                     end1AttributeDescription     = "Relational column holding the primary key.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);

        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "RelationalColumn";
        final String                     end2AttributeName            = "foreignKey";
        final String                     end2AttributeDescription     = "Use of primary key from another table to enable table joins.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.CONFIDENCE.name;
        final String attribute2Description     = OpenMetadataProperty.CONFIDENCE.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.CONFIDENCE.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute3Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute4Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute4Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.NAME.name;
        final String attribute5Description     = OpenMetadataProperty.NAME.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute2Name,
                                                        attribute2Description,
                                                        attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0535 Event Schemas describes event/message structures
     */
    private void add0535EventSchemas()
    {
        this.archiveBuilder.addEntityDef(getEventSetEntity());
        this.archiveBuilder.addEntityDef(getEventTypeEntity());
    }


    private EntityDef getEventSetEntity()
    {
        final String guid            = "bead9aa4-214a-4596-8036-aa78395bbfb1";
        final String name            = "EventSet";
        final String description     = "A collection of related event types.";
        final String descriptionGUID = null;

        final String superTypeName = "Collection";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getEventTypeEntity()
    {
        final String guid            = "8bc88aba-d7e4-4334-957f-cfe8e8eadc32";
        final String name            = "EventType";
        final String description     = "A description of an event (message)";
        final String descriptionGUID = null;

        final String superTypeName = "ComplexSchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0536 API schemas define the structure of an API
     */
    private void add0536APISchemas()
    {
        this.archiveBuilder.addEntityDef(getAPISchemaTypeEntity());
        this.archiveBuilder.addEntityDef(getAPIOperationSchemaEntity());

        this.archiveBuilder.addRelationshipDef(getAPIOperationsRelationship());
        this.archiveBuilder.addRelationshipDef(getAPIHeaderRelationship());
        this.archiveBuilder.addRelationshipDef(getAPIRequestRelationship());
        this.archiveBuilder.addRelationshipDef(getAPIResponseRelationship());
    }


    private EntityDef getAPISchemaTypeEntity()
    {
        final String guid            = "b46cddb3-9864-4c5d-8a49-266b3fc95cb8";
        final String name            = "APISchemaType";
        final String description     = "Description of an API.";
        final String descriptionGUID = null;

        final String superTypeName = "SchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getAPIOperationSchemaEntity()
    {
        final String guid            = "f1c0af19-2729-4fac-996e-a7badff3c21c";
        final String name            = "APIOperation";
        final String description     = "Description of an API operation.";
        final String descriptionGUID = null;

        final String superTypeName = "SchemaType";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getAPIOperationsRelationship()
    {
        final String guid            = "03737169-ceb5-45f0-84f0-21c5929945af";
        final String name            = "APIOperations";
        final String description     = "Link between an API and its operations.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "APISchemaType";
        final String                     end1AttributeName            = "usedInAPI";
        final String                     end1AttributeDescription     = "API that this operation belongs to.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "APIOperation";
        final String                     end2AttributeName            = "containsOperations";
        final String                     end2AttributeDescription     = "Operations for this API type.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getAPIHeaderRelationship()
    {
        final String guid            = "e8fb46d1-5f75-481b-aa66-f43ad44e2cc6";
        final String name            = "APIHeader";
        final String description     = "Link between an API operation and its header.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "APIOperation";
        final String                     end1AttributeName            = "usedAsAPIHeader";
        final String                     end1AttributeDescription     = "API operations using this structure as the header.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SchemaType";
        final String                     end2AttributeName            = "apiHeader";
        final String                     end2AttributeDescription     = "Header structure for this API operation.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getAPIRequestRelationship()
    {
        final String guid            = "4ab3b466-31bd-48ea-8aa2-75623476f2e2";
        final String name            = "APIRequest";
        final String description     = "Link between an API operation and its request structure.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "APIOperation";
        final String                     end1AttributeName            = "usedAsAPIRequest";
        final String                     end1AttributeDescription     = "API operations using this structure as the request body.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SchemaType";
        final String                     end2AttributeName            = "apiRequest";
        final String                     end2AttributeDescription     = "Request structure for this API operation.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getAPIResponseRelationship()
    {
        final String guid            = "e8001de2-1bb1-442b-a66f-9addc3641eae";
        final String name            = "APIResponse";
        final String description     = "Link between an API operation and its response structure.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "APIOperation";
        final String                     end1AttributeName            = "usedAsAPIResponse";
        final String                     end1AttributeDescription     = "API operations using this structure as the response.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SchemaType";
        final String                     end2AttributeName            = "apiResponse";
        final String                     end2AttributeDescription     = "Response structure for this API operation.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0540 Data Class definitions
     */
    private void add0540DataClasses()
    {
        this.archiveBuilder.addEnumDef(getDataClassAssignmentStatusEnum());

        this.archiveBuilder.addEntityDef(getDataClassEntity());

        this.archiveBuilder.addRelationshipDef(getDataClassHierarchyRelationship());
        this.archiveBuilder.addRelationshipDef(getDataClassCompositionRelationship());
        this.archiveBuilder.addRelationshipDef(getDataClassAssignmentRelationship());

    }


    private EnumDef getDataClassAssignmentStatusEnum()
    {
        EnumDef enumDef = archiveHelper.getEmptyEnumDef(DataClassAssignmentStatus.getOpenTypeGUID(),
                                                        DataClassAssignmentStatus.getOpenTypeName(),
                                                        DataClassAssignmentStatus.getOpenTypeDescription(),
                                                        DataClassAssignmentStatus.getOpenTypeDescriptionGUID(),
                                                        DataClassAssignmentStatus.getOpenTypeDescriptionWiki());

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        for (CriticalityLevel enumValues : CriticalityLevel.values())
        {
            elementDef = archiveHelper.getEnumElementDef(enumValues.getOrdinal(),
                                                         enumValues.getName(),
                                                         enumValues.getDescription(),
                                                         enumValues.getDescriptionGUID());

            elementDefs.add(elementDef);

            if (enumValues.isDefault())
            {
                enumDef.setDefaultValue(elementDef);
            }
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getDataClassEntity()
    {
       EntityDef entityDef = archiveHelper.getDefaultEntityDef(OpenMetadataType.DATA_CLASS.typeGUID,
                                                                OpenMetadataType.DATA_CLASS.typeName,
                                                                this.archiveBuilder.getEntityDef(OpenMetadataType.REFERENCEABLE.typeName),
                                                                OpenMetadataType.DATA_CLASS.description,
                                                                OpenMetadataType.DATA_CLASS.descriptionGUID,
                                                                OpenMetadataType.DATA_CLASS.wikiURL);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NAME.name;
        final String attribute1Description     = OpenMetadataProperty.NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "classCode";
        final String attribute3Description     = "Name of processing class that can identify the data class.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "userDefined";
        final String attribute4Description     = "Defined by owning organization rather than vendor.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "namespace";
        final String attribute5Description     = "Logical group for this data class.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "specification";
        final String attribute6Description     = "Parsing string used to identify values of this data class.";
        final String attribute6DescriptionGUID = null;
        final String attribute7Name            = "specificationDetails";
        final String attribute7Description     = "Additional properties used in the specification.";
        final String attribute7DescriptionGUID = null;
        final String attribute8Name            = "dataType";
        final String attribute8Description     = "Typical data type used to store this value.";
        final String attribute8DescriptionGUID = null;
        final String attribute9Name            = "defaultThreshold";
        final String attribute9Description     = "Match threshold that a data field is expected to achieve to be assigned this data class.";
        final String attribute9DescriptionGUID = null;
        final String attribute10Name            = "example";
        final String attribute10Description     = "Example of a data value that matches this data class.";
        final String attribute10DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getBooleanTypeDefAttribute(attribute4Name,
                                                            attribute4Description,
                                                            attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute7Name,
                                                           attribute7Description,
                                                           attribute7DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute8Name,
                                                           attribute8Description,
                                                           attribute8DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getFloatTypeDefAttribute(attribute9Name,
                                                          attribute9Description,
                                                          attribute9DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute10Name,
                                                           attribute10Description,
                                                           attribute10DescriptionGUID);
        properties.add(property);



        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getDataClassHierarchyRelationship()
    {
        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(OpenMetadataType.DATA_CLASS_HIERARCHY.typeGUID,
                                                                                OpenMetadataType.DATA_CLASS_HIERARCHY.typeName,
                                                                                null,
                                                                                OpenMetadataType.DATA_CLASS_HIERARCHY.description,
                                                                                OpenMetadataType.DATA_CLASS_HIERARCHY.descriptionGUID,
                                                                                OpenMetadataType.DATA_CLASS_HIERARCHY.wikiURL,
                                                                                ClassificationPropagationRule.NONE);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.DATA_CLASS.typeName;
        final String                     end1AttributeName            = "superDataClass";
        final String                     end1AttributeDescription     = "Data class that is the more abstract.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.DATA_CLASS.typeName;
        final String                     end2AttributeName            = "subDataClasses";
        final String                     end2AttributeDescription     = "Data classes that are more concrete.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getDataClassCompositionRelationship()
    {
        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(OpenMetadataType.DATA_CLASS_COMPOSITION.typeGUID,
                                                                                OpenMetadataType.DATA_CLASS_COMPOSITION.typeName,
                                                                                null,
                                                                                OpenMetadataType.DATA_CLASS_COMPOSITION.description,
                                                                                OpenMetadataType.DATA_CLASS_COMPOSITION.descriptionGUID,
                                                                                OpenMetadataType.DATA_CLASS_COMPOSITION.wikiURL,
                                                                                ClassificationPropagationRule.NONE);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.DATA_CLASS.typeName;
        final String                     end1AttributeName            = "partOfDataClasses";
        final String                     end1AttributeDescription     = "Data classes that includes other data classes in its definition.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.DATA_CLASS.typeName;
        final String                     end2AttributeName            = "madeOfDataClasses";
        final String                     end2AttributeDescription     = "Data classes that provide part of another data class's definitions.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getDataClassAssignmentRelationship()
    {
        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(OpenMetadataType.DATA_CLASS_ASSIGNMENT.typeGUID,
                                                                                OpenMetadataType.DATA_CLASS_ASSIGNMENT.typeName,
                                                                                null,
                                                                                OpenMetadataType.DATA_CLASS_ASSIGNMENT.description,
                                                                                OpenMetadataType.DATA_CLASS_ASSIGNMENT.descriptionGUID,
                                                                                OpenMetadataType.DATA_CLASS_ASSIGNMENT.wikiURL,
                                                                                ClassificationPropagationRule.NONE);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "elementsAssignedToDataClass";
        final String                     end1AttributeDescription     = "Elements identified as managing data values that match the specification of a data class.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.DATA_CLASS.typeName;
        final String                     end2AttributeName            = "dataClassesAssignedToElement";
        final String                     end2AttributeDescription     = "Logical data type for this element.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "method";
        final String attribute1Description     = "Method used to identify data class.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "status";
        final String attribute2Description     = "The status of the relationship.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "partialMatch";
        final String attribute3Description     = "Are there data values outside of the data class specification?";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.CONFIDENCE.name;
        final String attribute4Description     = OpenMetadataProperty.CONFIDENCE.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.CONFIDENCE.descriptionGUID;
        final String attribute5Name            = "threshold";
        final String attribute5Description     = "What was the threshold result used to determine that the data class matched.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "valueFrequency";
        final String attribute6Description     = "How often does the data class specification match the data values.";
        final String attribute6DescriptionGUID = null;
        final String attribute7Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute7Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute7DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute8Name            = OpenMetadataProperty.SOURCE.name;
        final String attribute8Description     = OpenMetadataProperty.SOURCE.description;
        final String attribute8DescriptionGUID = OpenMetadataProperty.SOURCE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("DataClassAssignmentStatus",
                                                         attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getBooleanTypeDefAttribute(attribute3Name,
                                                            attribute3Description,
                                                            attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute4Name,
                                                        attribute4Description,
                                                        attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getFloatTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getLongTypeDefAttribute(attribute6Name,
                                                         attribute6Description,
                                                         attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute7Name,
                                                           attribute7Description,
                                                           attribute7DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute8Name,
                                                           attribute8Description,
                                                           attribute8DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0545ReferenceData()
    {
        this.archiveBuilder.addClassificationDef(getReferenceDataClassification());

        this.archiveBuilder.addEntityDef(getValidValueDefinitionEntity());
        this.archiveBuilder.addEntityDef(getValidValuesSetEntity());

        this.archiveBuilder.addRelationshipDef(getValidValuesAssignmentRelationship());
        this.archiveBuilder.addRelationshipDef(getValidValuesMemberRelationship());
        this.archiveBuilder.addRelationshipDef(getValidValuesImplementationRelationship());
    }


    private ClassificationDef getReferenceDataClassification()
    {
        final String guid            = OpenMetadataType.REFERENCE_DATA_CLASSIFICATION.typeGUID;
        final String name            = OpenMetadataType.REFERENCE_DATA_CLASSIFICATION.typeName;
        final String description     = OpenMetadataType.REFERENCE_DATA_CLASSIFICATION.description;
        final String descriptionGUID = OpenMetadataType.REFERENCE_DATA_CLASSIFICATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.REFERENCE_DATA_CLASSIFICATION.wikiURL;

        final String linkedToEntity = OpenMetadataType.ASSET.typeName;

        return archiveHelper.getClassificationDef(guid,
                                                  name,
                                                  null,
                                                  description,
                                                  descriptionGUID,
                                                  descriptionWiki,
                                                  this.archiveBuilder.getEntityDef(linkedToEntity),
                                                  false);
    }


    private EntityDef getValidValueDefinitionEntity()
    {
        final String guid            = OpenMetadataType.VALID_VALUE_DEFINITION.typeGUID;
        final String name            = OpenMetadataType.VALID_VALUE_DEFINITION.typeName;
        final String description     = OpenMetadataType.VALID_VALUE_DEFINITION.description;
        final String descriptionGUID = OpenMetadataType.VALID_VALUE_DEFINITION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.VALID_VALUE_DEFINITION.wikiURL;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.NAME.name;
        final String attribute1Description     = OpenMetadataProperty.NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.USAGE.name;
        final String attribute3Description     = OpenMetadataProperty.USAGE.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.USAGE.descriptionGUID;
        final String attribute4Name            = "scope";
        final String attribute4Description     = "Situations where this value can be used.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "preferredValue";
        final String attribute5Description     = "Preferred implementation value.";
        final String attribute5DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getValidValuesSetEntity()
    {
        final String guid            = OpenMetadataType.VALID_VALUE_SET.typeGUID;
        final String name            = OpenMetadataType.VALID_VALUE_SET.typeName;
        final String description     = OpenMetadataType.VALID_VALUE_SET.description;
        final String descriptionGUID = OpenMetadataType.VALID_VALUE_SET.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.VALID_VALUE_SET.wikiURL;

        final String superTypeName = OpenMetadataType.VALID_VALUE_DEFINITION.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    private RelationshipDef getValidValuesAssignmentRelationship()
    {
        final String guid            = OpenMetadataType.VALID_VALUES_ASSIGNMENT_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.VALID_VALUES_ASSIGNMENT_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.VALID_VALUES_ASSIGNMENT_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.VALID_VALUES_ASSIGNMENT_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.VALID_VALUES_ASSIGNMENT_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "validValuesConsumer";
        final String                     end1AttributeDescription     = "The valid values set that this element belongs to.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.VALID_VALUE_DEFINITION.typeName;
        final String                     end2AttributeName            = "validValues";
        final String                     end2AttributeDescription     = "A definition of the valid values for this element.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "strictRequirement";
        final String attribute1Description     = "Only values from the ValidValues set/definition are allowed.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getBooleanTypeDefAttribute(attribute1Name,
                                                            attribute1Description,
                                                            attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getValidValuesMemberRelationship()
    {
        final String guid            = OpenMetadataType.VALID_VALUE_MEMBER_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.VALID_VALUE_MEMBER_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.VALID_VALUE_MEMBER_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.VALID_VALUE_MEMBER_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.VALID_VALUE_MEMBER_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.VALID_VALUE_SET.typeName;
        final String                     end1AttributeName            = "validValuesSet";
        final String                     end1AttributeDescription     = "The valid values set that this element belongs to.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.VALID_VALUE_DEFINITION.typeName;
        final String                     end2AttributeName            = "memberOfValidValuesSet";
        final String                     end2AttributeDescription     = "Description of a single valid value.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getValidValuesImplementationRelationship()
    {
        final String guid = OpenMetadataType.VALID_VALUES_IMPL_RELATIONSHIP.typeGUID;

        final String name            = OpenMetadataType.VALID_VALUES_IMPL_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.VALID_VALUES_IMPL_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.VALID_VALUES_IMPL_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.VALID_VALUES_IMPL_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.VALID_VALUE_DEFINITION.typeName;
        final String                     end1AttributeName            = "validValues";
        final String                     end1AttributeDescription     = "The valid values set that this element implements.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.ASSET.typeName;
        final String                     end2AttributeName            = "validValuesImplementation";
        final String                     end2AttributeDescription     = "The asset where the valid values are implemented.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "implementationValue";
        final String attribute1Description     = "Value in the asset that maps to this valid value if different from the preferred value.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }




    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0565DesignModelElements()
    {
        this.archiveBuilder.addEntityDef(getDesignModelElementEntity());
    }


    private EntityDef getDesignModelElementEntity()
    {
        final String guid            = "492e343f-2516-43b8-94b0-5bae0760dda6";
        final String name            = "DesignModelElement";
        final String description     = "An abstract, but well-formed representation of a concept, activity, architecture or other design element.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute2Name            = "technicalName";
        final String attribute2Description     = "Technical name (no spaces) that can be used in artifact generation.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute3Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute4Name            = "versionNumber";
        final String attribute4Description     = "Version number of the model element.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "author";
        final String attribute5Description     = "Name of the creator of the model (person or organization).";
        final String attribute5DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0566DesignModelOrganization()
    {
        this.archiveBuilder.addEntityDef(getDesignModelEntity());
        this.archiveBuilder.addEntityDef(getDesignModelGroupEntity());

        this.archiveBuilder.addRelationshipDef(getDesignModelElementOwnershipRelationship());
        this.archiveBuilder.addRelationshipDef(getDesignModelGroupOwnershipRelationship());
        this.archiveBuilder.addRelationshipDef(getDesignModelGroupHierarchyRelationship());
        this.archiveBuilder.addRelationshipDef(getDesignModelGroupMembershipRelationship());
    }


    private EntityDef getDesignModelEntity()
    {
        final String guid            = "bf17143d-8605-48c2-ba80-64c2ac8f8379";
        final String name            = "DesignModel";
        final String description     = "A curated collection of design model elements.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.ASSET.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;


        final String attribute1Name            = "technicalName";
        final String attribute1Description     = "Technical name (no spaces) that can be used in artifact generation.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "versionNumber";
        final String attribute2Description     = "Version number of the model.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "author";
        final String attribute3Description     = "Name of the creator of the model (person or organization).";
        final String attribute3DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getDesignModelGroupEntity()
    {
        final String guid            = "b144ee2a-fa71-4897-b51a-dd5239c26910";
        final String name            = "DesignModelGroup";
        final String description     = "A collection of related design model elements within a model.";
        final String descriptionGUID = null;

        final String superTypeName = "DesignModelElement";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private RelationshipDef getDesignModelElementOwnershipRelationship()
    {
        final String guid            = "f3b18ac7-3357-4a0c-8988-77a98adad5b5";
        final String name            = "DesignModelElementOwnership";
        final String description     = "Links design model elements to their owning model.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "DesignModel";
        final String                     end1AttributeName            = "owningModel";
        final String                     end1AttributeDescription     = "Owning model.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "DesignModelElement";
        final String                     end2AttributeName            = "elementsInModel";
        final String                     end2AttributeDescription     = "List of elements that belong to this model.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getDesignModelGroupOwnershipRelationship()
    {
        final String guid            = "4a985162-8130-4559-b68e-6e6a5dc19c2a";
        final String name            = "DesignModelGroupOwnership";
        final String description     = "Links a model to a design model group.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "DesignModel";
        final String                     end1AttributeName            = "anchorModel";
        final String                     end1AttributeDescription     = "Model that owns this group.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "DesignModelGroup";
        final String                     end2AttributeName            = "groupsInModel";
        final String                     end2AttributeDescription     = "List of groups that belong to this model.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getDesignModelGroupHierarchyRelationship()
    {
        final String guid            = "809b7c6c-69f9-4dbf-a5dd-085664499438";
        final String name            = "DesignModelGroupHierarchy";
        final String description     = "Links a model's groups into a hierarchy.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "DesignModelGroup";
        final String                     end1AttributeName            = "parentModelGroup";
        final String                     end1AttributeDescription     = "Link parent group.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "DesignModelGroup";
        final String                     end2AttributeName            = "childModelGroups";
        final String                     end2AttributeDescription     = "The groups nested in this group.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getDesignModelGroupMembershipRelationship()
    {
        final String guid            = "2dcfe62b-341c-4c3d-b336-a94a52c20556";
        final String name            = "DesignModelGroupMembership";
        final String description     = "Links a design model element to a group.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "DesignModelGroup";
        final String                     end1AttributeName            = "memberOfModelGroups";
        final String                     end1AttributeDescription     = "Link to a list of groups this element is a member of.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "DesignModelElement";
        final String                     end2AttributeName            = "elementsInGroup";
        final String                     end2AttributeDescription     = "List of elements that belong to this group.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0568DesignModelScoping()
    {
        this.archiveBuilder.addEntityDef(getDesignModelScopeEntity());
        this.archiveBuilder.addRelationshipDef(getDesignModelElementsInScopeRelationship());
    }


    private EntityDef getDesignModelScopeEntity()
    {
        final String guid            = "788957f7-a203-45bd-994d-0ab018275821";
        final String name            = "DesignModelScope";
        final String description     = "A selection of design model element needed for a project.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute2Name            = "technicalName";
        final String attribute2Description     = "Technical name (no spaces) that can be used in artifact generation.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute3Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute4Name            = "versionNumber";
        final String attribute4Description     = "Version number of the model element.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "author";
        final String attribute5Description     = "UserId of the creator of the model element.";
        final String attribute5DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getDesignModelElementsInScopeRelationship()
    {
        final String guid            = "4ff6d91b-3836-4ba2-9ca9-87da91081faa";
        final String name            = "DesignModelElementsInScope";
        final String description     = "Links a model to an implementation.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "DesignModelScope";
        final String                     end1AttributeName            = "usedInScope";
        final String                     end1AttributeDescription     = "Link to a scope where this element is used.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "DesignModelElement";
        final String                     end2AttributeName            = "inScopeModelElements";
        final String                     end2AttributeDescription     = "List of elements that belong to this scope.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0569DesignModelImplementation()
    {
        this.archiveBuilder.addRelationshipDef(geDesignModelImplementationRelationship());
    }


    private RelationshipDef geDesignModelImplementationRelationship()
    {
        final String guid            = "c5cb1362-07f6-486b-b80b-ba7922cacee9";
        final String name            = "DesignModelImplementation";
        final String description     = "Links a concept model to an implementation.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "implementationFollowingModel";
        final String                     end1AttributeDescription     = "Definition of an implementation of the model.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "DesignModelElement";
        final String                     end2AttributeName            = "modelDescribingBehavior";
        final String                     end2AttributeDescription     = "Descriptive abstraction.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0570MetaModel()
    {
        this.archiveBuilder.addClassificationDef(getMetamodelClassification());
    }


    private ClassificationDef getMetamodelClassification()
    {
        final String guid            = "07bd0820-6b14-43b0-a625-2c89f2beb93a";
        final String name            = "MetamodelInstance";
        final String description     = "Identifies the element from a metadata model that this element embodies.";
        final String descriptionGUID = null;

        final String linkedToEntity = "DesignModelElement";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "metamodelElementGUID";
        final String attribute1Description     = "Element in the metadata model that the attached element embodies.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0571ConceptModels()
    {
        this.archiveBuilder.addEnumDef(getConceptModelAttributeCoverageCategoryEnum());
        this.archiveBuilder.addEnumDef(getConceptModelDecorationEnum());

        this.archiveBuilder.addEntityDef(getConceptModelElementEntity());
        this.archiveBuilder.addEntityDef(getConceptBeadEntity());
        this.archiveBuilder.addEntityDef(getConceptBeadLinkEntity());
        this.archiveBuilder.addEntityDef(getConceptBeadAttributeEntity());

        this.archiveBuilder.addRelationshipDef(getConceptBeadRelationshipEndRelationship());
        this.archiveBuilder.addRelationshipDef(getConceptBeadAttributeLinkRelationship());

        this.archiveBuilder.addClassificationDef(getConceptBeadAttributeCoverageClassification());
    }


    private EnumDef getConceptModelAttributeCoverageCategoryEnum()
    {
        final String guid            = "2c0ac237-e02e-431a-89fd-3107d94d4007";
        final String name            = "ConceptModelAttributeCoverageCategory";
        final String description     = "Describes the type of attribute - this is used in scoping the model.";
        final String descriptionGUID = null;

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID);

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        final int    element1Ordinal         = 0;
        final String element1Value           = "Unknown";
        final String element1Description     = "The attribute's coverage category is unknown - this is the default.";
        final String element1DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element1Ordinal,
                                                     element1Value,
                                                     element1Description,
                                                     element1DescriptionGUID);
        elementDefs.add(elementDef);
        enumDef.setDefaultValue(elementDef);

        final int    element2Ordinal         = 1;
        final String element2Value           = "UniqueIdentifier";
        final String element2Description     = "The attribute uniquely identifies the concept bead.";
        final String element2DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element2Ordinal,
                                                     element2Value,
                                                     element2Description,
                                                     element2DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element3Ordinal         = 2;
        final String element3Value           = "Identifier";
        final String element3Description     = "The attribute is a good indicator of the identity of the concept bead but not guaranteed to be unique.";
        final String element3DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element3Ordinal,
                                                     element3Value,
                                                     element3Description,
                                                     element3DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element4Ordinal         = 3;
        final String element4Value           = "CoreDetail";
        final String element4Description     = "The attribute provides information that is typically required by all of the consumers of the concept bead.";
        final String element4DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element4Ordinal,
                                                     element4Value,
                                                     element4Description,
                                                     element4DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element5Ordinal         = 4;
        final String element5Value           = "ExtendedDetail";
        final String element5Description     = "The attribute contains supplementary information that is of interest to specific consumers of the concept bead.";
        final String element5DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element5Ordinal,
                                                     element5Value,
                                                     element5Description,
                                                     element5DescriptionGUID);
        elementDefs.add(elementDef);

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EnumDef getConceptModelDecorationEnum()
    {
        final String guid            = "a97d9167-7dd6-4dea-a8cf-c73c57a0f470";
        final String name            = "ConceptModelDecoration";
        final String description     = "Describes the type of relationship end.";
        final String descriptionGUID = null;

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID);

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        final int    element1Ordinal         = 0;
        final String element1Value           = "None";
        final String element1Description     = "The relationship links two concept beads together.";
        final String element1DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element1Ordinal,
                                                     element1Value,
                                                     element1Description,
                                                     element1DescriptionGUID);
        elementDefs.add(elementDef);
        enumDef.setDefaultValue(elementDef);

        final int    element2Ordinal         = 1;
        final String element2Value           = "Aggregation";
        final String element2Description     = "The relationship links an independent concept bead to a collection concept bead.";
        final String element2DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element2Ordinal,
                                                     element2Value,
                                                     element2Description,
                                                     element2DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element3Ordinal         = 2;
        final String element3Value           = "Composition";
        final String element3Description     = "The relationship links a sub-part to a composite.";
        final String element3DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element3Ordinal,
                                                     element3Value,
                                                     element3Description,
                                                     element3DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element4Ordinal         = 3;
        final String element4Value           = "Extension";
        final String element4Description     = "The relationship links an extension to a base concept bead.";
        final String element4DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element4Ordinal,
                                                     element4Value,
                                                     element4Description,
                                                     element4DescriptionGUID);
        elementDefs.add(elementDef);

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getConceptModelElementEntity()
    {
        final String guid            = "06659195-3111-4c91-8931-a65f655378d9";
        final String name            = "ConceptModelElement";
        final String description     = "An abstract, but well-formed representation of a concept.";
        final String descriptionGUID = null;

        final String superTypeName = "DesignModelElement";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getConceptBeadEntity()
    {
        final String guid            = "f7feb509-bce6-4989-a340-5dc7e3eec313";
        final String name            = "ConceptBead";
        final String description     = "An abstract, but well-formed representation of a person, place or object.";
        final String descriptionGUID = null;

        final String superTypeName = "ConceptModelElement";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getConceptBeadLinkEntity()
    {
        final String guid            = "13defd95-6452-4398-8382-e47f1a271eff";
        final String name            = "ConceptBeadLink";
        final String description     = "A relationship between concept beads.";
        final String descriptionGUID = null;

        final String superTypeName = "ConceptModelElement";

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID);
    }


    private EntityDef getConceptBeadAttributeEntity()
    {
        final String guid            = "d804d406-ac74-4f92-9bde-2ba0793680ea";
        final String name            = "ConceptBeadAttribute";
        final String description     = "An abstract, but well-formed fact about a concept bead.";
        final String descriptionGUID = null;

        final String superTypeName = "ConceptModelElement";

        return  archiveHelper.getDefaultEntityDef(guid,
                                                  name,
                                                  this.archiveBuilder.getEntityDef(superTypeName),
                                                  description,
                                                  descriptionGUID);
    }


    private RelationshipDef getConceptBeadRelationshipEndRelationship()
    {
        final String guid            = "1a379e55-a4c0-4289-a1a4-b89d257611d1";
        final String name            = "ConceptBeadRelationshipEnd";
        final String description     = "Links one end of a concept bead link relationship to a concept bead.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "ConceptBeadLink";
        final String                     end1AttributeName            = "relationships";
        final String                     end1AttributeDescription     = "The relationships that the concept bead can be a part of.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ConceptBead";
        final String                     end2AttributeName            = "endBeads";
        final String                     end2AttributeDescription     = "The concept beads that are linked via this relationship.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "attributeName";
        final String attribute1Description     = "Name for the relationship end.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "decoration";
        final String attribute2Description     = "Usage and lifecycle for this connection between the concept bead and the link.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "position";
        final String attribute3Description     = "Position of this relationship in the concept bead's list of relationships.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "minCardinality";
        final String attribute4Description     = "Minimum number of occurrences of this attribute allowed (0 = optional, 1+ = mandatory).";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "maxCardinality";
        final String attribute5Description     = "Maximum number of occurrences of this attribute allowed (-1 = infinite).";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "uniqueValues";
        final String attribute6Description     = "When multiple occurrences are allowed, indicates whether duplicates of the same value are allowed or not.";
        final String attribute6DescriptionGUID = null;
        final String attribute7Name            = "orderedValues";
        final String attribute7Description     = "When multiple occurrences are allowed, indicates whether the values are ordered or not.";
        final String attribute7DescriptionGUID = null;
        final String attribute8Name            = "navigable";
        final String attribute8Description     = "Is it possible to follow the link in this direction.";
        final String attribute8DescriptionGUID = null;


        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("ConceptModelDecoration",
                                                         attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute3Name,
                                                        attribute3Description,
                                                        attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute4Name,
                                                        attribute4Description,
                                                        attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute5Name,
                                                        attribute5Description,
                                                        attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getBooleanTypeDefAttribute(attribute6Name,
                                                            attribute6Description,
                                                            attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getBooleanTypeDefAttribute(attribute7Name,
                                                            attribute7Description,
                                                            attribute7DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getBooleanTypeDefAttribute(attribute8Name,
                                                            attribute8Description,
                                                            attribute8DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getConceptBeadAttributeLinkRelationship()
    {
        final String guid            = "5bad1df2-664b-407b-8036-2855e2ede92f";
        final String name            = "ConceptBeadAttributeLink";
        final String description     = "Links a concept bead to its attributes.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "ConceptBead";
        final String                     end1AttributeName            = "parentBead";
        final String                     end1AttributeDescription     = "Concept bead that this attribute belongs to.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "ConceptBeadAttribute";
        final String                     end2AttributeName            = "attributes";
        final String                     end2AttributeDescription     = "Attribute detail for the concept bead.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "position";
        final String attribute1Description     = "Position of this relationship in the concept bead's list of relationships.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "minCardinality";
        final String attribute2Description     = "Minimum number of occurrences of this attribute allowed (0 = optional, 1+ = mandatory).";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "maxCardinality";
        final String attribute3Description     = "Maximum number of occurrences of this attribute allowed (-1 = infinite).";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "uniqueValues";
        final String attribute4Description     = "When multiple occurrences are allowed, indicates whether duplicates of the same value are allowed or not.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "orderedValues";
        final String attribute5Description     = "When multiple occurrences are allowed, indicates whether the values are ordered or not.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "navigable";
        final String attribute6Description     = "Is it possible to follow the link in this direction.";
        final String attribute6DescriptionGUID = null;


        property = archiveHelper.getIntTypeDefAttribute(attribute1Name,
                                                        attribute1Description,
                                                        attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute2Name,
                                                        attribute2Description,
                                                        attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute3Name,
                                                        attribute3Description,
                                                        attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getBooleanTypeDefAttribute(attribute4Name,
                                                            attribute4Description,
                                                            attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getBooleanTypeDefAttribute(attribute5Name,
                                                            attribute5Description,
                                                            attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getBooleanTypeDefAttribute(attribute6Name,
                                                            attribute6Description,
                                                            attribute6DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private ClassificationDef getConceptBeadAttributeCoverageClassification()
    {
        final String guid            = "f8b60afe-ddef-4b6f-9628-82ebfff34d65";
        final String name            = "ConceptBeadAttributeCoverage";
        final String description     = "Identifies the coverage category of a concept bead attribute.";
        final String descriptionGUID = null;

        final String linkedToEntity = "ConceptBeadAttribute";

        ClassificationDef classificationDef = archiveHelper.getClassificationDef(guid,
                                                                                 name,
                                                                                 null,
                                                                                 description,
                                                                                 descriptionGUID,
                                                                                 this.archiveBuilder.getEntityDef(linkedToEntity),
                                                                                 false);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "coverageCategory";
        final String attribute1Description     = "Type of role that the attribute plays as part of the concept bead.";
        final String attribute1DescriptionGUID = null;


        property = archiveHelper.getEnumTypeDefAttribute("ConceptModelAttributeCoverageCategory",
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);

        classificationDef.setPropertiesDefinition(properties);

        return classificationDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    private void add0595DesignPatterns()
    {
        this.archiveBuilder.addEntityDef(getDesignPatternEntity());

        this.archiveBuilder.addRelationshipDef(getRelatedPatternRelationship());

    }


    private EntityDef getDesignPatternEntity()
    {
        final String guid            = "6b60a73e-47bc-4096-9073-f94cab975958";
        final String name            = "DesignPattern";
        final String description     = "A description of a common solution with details of the problems it solves and its pros and cons.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "context";
        final String attribute1Description     = "Description of the situation where this pattern may be useful.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "forces";
        final String attribute2Description     = "Description of the aspects of the situation that make the problem hard to solve.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "problemStatement";
        final String attribute3Description     = "Description of the types of problem that this design pattern provides a solution to.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "problemExample";
        final String attribute4Description     = "One or more examples of the problem and its consequences.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "solutionDescription";
        final String attribute5Description     = "Description of how the solution works.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "solutionExample";
        final String attribute6Description     = "Illustrations of how the solution resolves the problem examples.";
        final String attribute6DescriptionGUID = null;
        final String attribute7Name            = "benefits";
        final String attribute7Description     = "The positive outcomes from using this pattern.";
        final String attribute7DescriptionGUID = null;
        final String attribute8Name            = "liabilities";
        final String attribute8Description     = "The additional issues that need to be considered when using this pattern.";
        final String attribute8DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute7Name,
                                                                attribute7Description,
                                                                attribute7DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute8Name,
                                                                attribute8Description,
                                                                attribute8DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getRelatedPatternRelationship()
    {
        final String guid            = "6447c9cd-8e5a-461b-97f9-5151bcb97a9e";
        final String name            = "RelatedDesignPattern";
        final String description     = "Links design patterns together.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "DesignPattern";
        final String                     end1AttributeName            = "relatedDesignPattern";
        final String                     end1AttributeDescription     = "Another design pattern that operates in similar contexts.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "DesignPattern";
        final String                     end2AttributeName            = "relatedDesignPattern";
        final String                     end2AttributeDescription     = "Another design pattern that operates in similar contexts.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute1Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private void add0598LineageRelationships() {

        this.archiveBuilder.addRelationshipDef(getPortSchemaRelationship());
        this.archiveBuilder.addRelationshipDef(getLineageMappingRelationship());
    }

    /**
     * The PortSchema relationship describes the link between a Port and the SchemaType linked to the Port
     * @return PortSchema RelationshipDef
     */
    private RelationshipDef getPortSchemaRelationship() {
        /*
         * Build the relationship
         */
        final String guid            = "B216fA00-8281-F9CC-9911-Ae6377f2b457";
        final String name            = "PortSchema";
        final String description     = "A link between a Port and a SchemaType";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "Port";
        final String                     end1AttributeName            = "port";
        final String                     end1AttributeDescription     = "Port";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                end1AttributeName,
                end1AttributeDescription,
                end1AttributeDescriptionGUID,
                end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SchemaType";
        final String                     end2AttributeName            = "SchemaType";
        final String                     end2AttributeDescription     = "SchemaType";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                end2AttributeName,
                end2AttributeDescription,
                end2AttributeDescriptionGUID,
                end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);


        return relationshipDef;
    }

    /**
     *  The LineageMapping relationship describes the directional mapping between SchemaTypes
     * @return LineageMapping RelationshipDef
     */
    private RelationshipDef getLineageMappingRelationship() {
        /*
         * Build the relationship
         */
        final String guid            = "a5991bB2-660D-A3a1-2955-fAcDA2d5F4Ff";
        final String name            = "LineageMapping";
        final String description     = "A link between two schema attributes.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;
        final String                     end1AttributeName            = "sourceElement";
        final String                     end1AttributeDescription     = "Source Attribute.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                end1AttributeName,
                end1AttributeDescription,
                end1AttributeDescriptionGUID,
                end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;
        final String                     end2AttributeName            = "targetElement";
        final String                     end2AttributeDescription     = "Target Attribute.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                end2AttributeName,
                end2AttributeDescription,
                end2AttributeDescriptionGUID,
                end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);


        return relationshipDef;
    }


    /*
     * ========================================
     * AREA 6: metadata discovery
     */

    /**
     * Area 6 are the types for the discovery server (and discovery pipelines
     * supporting the Open Discovery Framework (ODF)).
     */
    private void addArea6Types()
    {
        this.add0601OpenDiscoveryEngine();
        this.add0602OpenDiscoveryServices();
        this.add0603OpenDiscoveryPipelines();
        this.add0605OpenDiscoveryAnalysisReports();
        this.add0610Annotations();
        this.add0612AnnotationReviews();
        this.add0615SchemaExtraction();
        this.add0617DataFieldAnalysis();
        this.add0620DataProfiling();
        this.add0625DataClassDiscovery();
        this.add0630SemanticDiscovery();
        this.add0635ClassificationDiscovery();
        this.add0640QualityScores();
        this.add0650RelationshipDiscovery();
        this.add0660ResourceMeasures();
        this.add0690RequestForAction();

    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0601 Metadata Discovery Server defines how a metadata discovery server is represented in the metadata repository.
     */
    private void add0601OpenDiscoveryEngine()
    {
        this.archiveBuilder.addEntityDef(getOpenDiscoveryEngineEntity());
    }


    private EntityDef getOpenDiscoveryEngineEntity()
    {
        final String guid            = OpenMetadataType.OPEN_DISCOVERY_ENGINE.typeGUID;
        final String name            = OpenMetadataType.OPEN_DISCOVERY_ENGINE.typeName;
        final String description     = OpenMetadataType.OPEN_DISCOVERY_ENGINE.description;
        final String descriptionGUID = OpenMetadataType.OPEN_DISCOVERY_ENGINE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.OPEN_DISCOVERY_ENGINE.wikiURL;

        final String superTypeName = OpenMetadataType.SOFTWARE_SERVER_CAPABILITY.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0602 Open Discovery Services defines information about a specific implementation of an Open Discovery
     * Service.  This is a pluggable component of the Open Discovery Framework (ODF).
     */
    private void add0602OpenDiscoveryServices()
    {
        this.archiveBuilder.addEntityDef(getOpenDiscoveryServiceEntity());
    }


    private EntityDef getOpenDiscoveryServiceEntity()
    {
        final String guid            = OpenMetadataType.OPEN_DISCOVERY_SERVICE.typeGUID;
        final String name            = OpenMetadataType.OPEN_DISCOVERY_SERVICE.typeName;
        final String description     = OpenMetadataType.OPEN_DISCOVERY_SERVICE.description;
        final String descriptionGUID = OpenMetadataType.OPEN_DISCOVERY_SERVICE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.OPEN_DISCOVERY_SERVICE.wikiURL;

        final String superTypeName = OpenMetadataType.DEPLOYED_SOFTWARE_COMPONENT_TYPE_NAME;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0603 Open Discovery Pipelines defines information about a specific implementation of an Open Discovery
     * Pipeline.  This is a specialized open discovery service that executes other open discovery services.
     */
    private void add0603OpenDiscoveryPipelines()
    {
        this.archiveBuilder.addEntityDef(getOpenDiscoveryPipelineEntity());
    }


    private EntityDef getOpenDiscoveryPipelineEntity()
    {
        final String guid            = OpenMetadataType.OPEN_DISCOVERY_PIPELINE.typeGUID;
        final String name            = OpenMetadataType.OPEN_DISCOVERY_PIPELINE.typeName;
        final String description     = OpenMetadataType.OPEN_DISCOVERY_PIPELINE.description;
        final String descriptionGUID = OpenMetadataType.OPEN_DISCOVERY_PIPELINE.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.OPEN_DISCOVERY_PIPELINE.wikiURL;

        final String superTypeName = OpenMetadataType.OPEN_DISCOVERY_SERVICE.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0605 Discovery Analysis Reports defines hoe the annotations from a discovery service are grouped
     * together and connected to the Assets they refer to.
     */
    private void add0605OpenDiscoveryAnalysisReports()
    {
        this.archiveBuilder.addEnumDef(getDiscoveryRequestStatusEnum());

        this.archiveBuilder.addEntityDef(getOpenDiscoveryAnalysisReportEntity());

        this.archiveBuilder.addRelationshipDef(getDiscoveryEngineReportRelationship());
        this.archiveBuilder.addRelationshipDef(getDiscoveryInvocationReportRelationship());
        this.archiveBuilder.addRelationshipDef(getAssetDiscoveryReportRelationship());
    }


    private EnumDef getDiscoveryRequestStatusEnum()
    {
        final String guid            = "ecb48ca2-4d29-4de9-99a1-bc4db9816d68";
        final String name            = "DiscoveryRequestStatus";
        final String description     = "Defines the progress or completion of a discovery request.";
        final String descriptionGUID = null;

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID);

        ArrayList<EnumElementDef> elementDefs = new ArrayList<>();
        EnumElementDef            elementDef;

        final int    element1Ordinal         = 0;
        final String element1Value           = "Waiting";
        final String element1Description     = "Discovery request is waiting to execute.";
        final String element1DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element1Ordinal,
                                                     element1Value,
                                                     element1Description,
                                                     element1DescriptionGUID);
        elementDefs.add(elementDef);
        enumDef.setDefaultValue(elementDef);

        final int    element2Ordinal         = 1;
        final String element2Value           = "InProgress";
        final String element2Description     = "Discovery request is executing.";
        final String element2DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element2Ordinal,
                                                     element2Value,
                                                     element2Description,
                                                     element2DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element3Ordinal         = 2;
        final String element3Value           = "Failed";
        final String element3Description     = "Discovery request has failed.";
        final String element3DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element3Ordinal,
                                                     element3Value,
                                                     element3Description,
                                                     element3DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element4Ordinal         = 3;
        final String element4Value           = "Completed";
        final String element4Description     = "Discovery request has completed successfully.";
        final String element4DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element4Ordinal,
                                                     element4Value,
                                                     element4Description,
                                                     element4DescriptionGUID);
        elementDefs.add(elementDef);

        final int    element99Ordinal         = 99;
        final String element99Value           = "Unknown";
        final String element99Description     = "Discovery request status is unknown.";
        final String element99DescriptionGUID = null;

        elementDef = archiveHelper.getEnumElementDef(element99Ordinal,
                                                     element99Value,
                                                     element99Description,
                                                     element99DescriptionGUID);
        elementDefs.add(elementDef);

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getOpenDiscoveryAnalysisReportEntity()
    {
        final String guid = "acc7cbc8-09c3-472b-87dd-f78459323dcb";

        final String name            = "OpenDiscoveryAnalysisReport";
        final String description     = "A set of results from an open discovery service.";
        final String descriptionGUID = null;

        final String superTypeName = OpenMetadataType.REFERENCEABLE.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.DISPLAY_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.DISPLAY_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.DISPLAY_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.DESCRIPTION.name;
        final String attribute2Description     = OpenMetadataProperty.DESCRIPTION.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.DESCRIPTION.descriptionGUID;
        final String attribute3Name            = "executionDate";
        final String attribute3Description     = "Date that the analysis was run.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = "analysisParameters";
        final String attribute4Description     = "Additional parameters used to drive the analysis.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "discoveryRequestStatus";
        final String attribute5Description     = "Status of the discovery analysis show in the report.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "anchorGUID";
        final String attribute6Description     = "Optional identifier of the asset that this report is connected to.";
        final String attribute6DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute4Name,
                                                                    attribute4Description,
                                                                    attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("DiscoveryRequestStatus",
                                                         attribute5Name,
                                                         attribute5Description,
                                                         attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }

    private RelationshipDef getDiscoveryEngineReportRelationship()
    {
        final String guid = "2c318c3a-5dc2-42cd-a933-0087d852f67f";

        final String name            = "DiscoveryEngineReport";
        final String description     = "A discovery analysis report created by a discovery engine.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "OpenDiscoveryEngine";
        final String                     end1AttributeName            = "sourceDiscoveryEngine";
        final String                     end1AttributeDescription     = "The discovery engine that produced the report.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "OpenDiscoveryAnalysisReport";
        final String                     end2AttributeName            = "discoveryEngineAnalysisReports";
        final String                     end2AttributeDescription     = "The reports produced by this discovery engine.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getDiscoveryInvocationReportRelationship()
    {
        final String guid = "1744d72b-903d-4273-9229-de20372a17e2";

        final String name            = "DiscoveryInvocationReport";
        final String description     = "An analysis report from a discovery service.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "OpenDiscoveryService";
        final String                     end1AttributeName            = "sourceDiscoveryService";
        final String                     end1AttributeDescription     = "The discovery service that produced the report.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "OpenDiscoveryAnalysisReport";
        final String                     end2AttributeName            = "serviceDiscoveryAnalysisReports";
        final String                     end2AttributeDescription     = "The reports produced by this discovery service.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getAssetDiscoveryReportRelationship()
    {
        final String guid = "7eded424-f176-4258-9ae6-138a46b2845f";

        final String name            = "AssetDiscoveryReport";
        final String description     = "An analysis report from a discovery service.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.ASSET.typeName;
        final String                     end1AttributeName            = "discoveryReportTarget";
        final String                     end1AttributeDescription     = "The asset that is analyzed in the report.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "OpenDiscoveryAnalysisReport";
        final String                     end2AttributeName            = "assetDiscoveryAnalysisReports";
        final String                     end2AttributeDescription     = "The reports produced about this asset.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0610 Annotations describes the basic structure of an annotation created by a discovery service.
     */
    private void add0610Annotations()
    {
        this.archiveBuilder.addEntityDef(getAnnotationEntity());

        this.archiveBuilder.addRelationshipDef(getDiscoveredAnnotationRelationship());
        this.archiveBuilder.addRelationshipDef(getAnnotationExtensionRelationship());
    }


    private EntityDef getAnnotationEntity()
    {
        final String guid            = OpenMetadataType.ANNOTATION.typeGUID;
        final String name            = OpenMetadataType.ANNOTATION.typeName;
        final String description     = OpenMetadataType.ANNOTATION.description;
        final String descriptionGUID = OpenMetadataType.ANNOTATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.ANNOTATION.wikiURL;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                null,
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.ANNOTATION_TYPE.name;
        final String attribute1Description     = OpenMetadataProperty.ANNOTATION_TYPE.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.ANNOTATION_TYPE.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.SUMMARY.name;
        final String attribute2Description     = OpenMetadataProperty.SUMMARY.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.SUMMARY.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.CONFIDENCE_LEVEL.name;
        final String attribute3Description     = OpenMetadataProperty.CONFIDENCE_LEVEL.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.CONFIDENCE_LEVEL.descriptionGUID;
        final String attribute4Name            = OpenMetadataProperty.EXPRESSION.name;
        final String attribute4Description     = OpenMetadataProperty.EXPRESSION.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.EXPRESSION.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.EXPLANATION.name;
        final String attribute5Description     = OpenMetadataProperty.EXPLANATION.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.EXPLANATION.descriptionGUID;
        final String attribute6Name            = OpenMetadataProperty.ANALYSIS_STEP.name;
        final String attribute6Description     = OpenMetadataProperty.ANALYSIS_STEP.description;
        final String attribute6DescriptionGUID = OpenMetadataProperty.ANALYSIS_STEP.descriptionGUID;
        final String attribute7Name            = OpenMetadataProperty.JSON_PROPERTIES.name;
        final String attribute7Description     = OpenMetadataProperty.JSON_PROPERTIES.description;
        final String attribute7DescriptionGUID = OpenMetadataProperty.JSON_PROPERTIES.descriptionGUID;
        final String attribute8Name            = OpenMetadataProperty.ADDITIONAL_PROPERTIES.name;
        final String attribute8Description     = OpenMetadataProperty.ADDITIONAL_PROPERTIES.description;
        final String attribute8DescriptionGUID = OpenMetadataProperty.ADDITIONAL_PROPERTIES.descriptionGUID;
        final String attribute9Name            = OpenMetadataProperty.ANCHOR_GUID.name;
        final String attribute9Description     = OpenMetadataProperty.ANCHOR_GUID.description;
        final String attribute9DescriptionGUID = OpenMetadataProperty.ANCHOR_GUID.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute3Name,
                                                        attribute3Description,
                                                        attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute7Name,
                                                           attribute7Description,
                                                           attribute7DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute8Name,
                                                                    attribute8Description,
                                                                    attribute8DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute9Name,
                                                           attribute9Description,
                                                           attribute9DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getDiscoveredAnnotationRelationship()
    {
        final String guid            = "51d386a3-3857-42e3-a3df-14a6cad08b93";
        final String name            = "DiscoveredAnnotation";
        final String description     = "The annotations that make up a discovery analysis report.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.ANNOTATION.typeName;
        final String                     end1AttributeName            = "reportedAnnotations";
        final String                     end1AttributeDescription     = "The annotations providing the contents for the report.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "OpenDiscoveryAnalysisReport";
        final String                     end2AttributeName            = "fromAnalysisReport";
        final String                     end2AttributeDescription     = "The report that the annotations belong to.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getAnnotationExtensionRelationship()
    {
        final String guid            = OpenMetadataType.ANNOTATION_EXTENSION_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.ANNOTATION_EXTENSION_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.ANNOTATION_EXTENSION_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.ANNOTATION_EXTENSION_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.ANNOTATION_EXTENSION_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.ANNOTATION.typeName;
        final String                     end1AttributeName            = "extendedAnnotations";
        final String                     end1AttributeDescription     = "The annotations being extended.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.ANNOTATION.typeName;
        final String                     end2AttributeName            = "annotationExtensions";
        final String                     end2AttributeDescription     = "The annotations providing additional information.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0612 Annotation Review defines the outcome of a stewardship review of the annotations in a discovery
     * analysis report.
     */
    private void add0612AnnotationReviews()
    {
        this.archiveBuilder.addEnumDef(getAnnotationStatusEnum());

        this.archiveBuilder.addEntityDef(getAnnotationReviewEntity());

        this.archiveBuilder.addRelationshipDef(getAnnotationReviewLinkRelationship());
    }


    private EnumDef getAnnotationStatusEnum()
    {
        final String guid            = AnnotationStatus.getOpenTypeGUID();
        final String name            = AnnotationStatus.getOpenTypeName();
        final String description     = AnnotationStatus.getOpenTypeDescription();
        final String descriptionGUID = AnnotationStatus.getOpenTypeDescriptionGUID();
        final String descriptionWiki = AnnotationStatus.getOpenTypeWikiURL();

        EnumDef enumDef = archiveHelper.getEmptyEnumDef(guid, name, description, descriptionGUID, descriptionWiki);

        List<EnumElementDef> elementDefs = new ArrayList<>();

        for (AnnotationStatus annotationStatus : AnnotationStatus.values())
        {
            elementDefs.add(archiveHelper.getEnumElementDef(annotationStatus.getOrdinal(),
                                                            annotationStatus.getName(),
                                                            annotationStatus.getDescription(),
                                                            annotationStatus.getDescriptionGUID()));
        }

        enumDef.setElementDefs(elementDefs);

        return enumDef;
    }


    private EntityDef getAnnotationReviewEntity()
    {
        final String guid            = OpenMetadataType.ANNOTATION_REVIEW.typeGUID;
        final String name            = OpenMetadataType.ANNOTATION_REVIEW.typeName;
        final String description     = OpenMetadataType.ANNOTATION_REVIEW.description;
        final String descriptionGUID = OpenMetadataType.ANNOTATION_REVIEW.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.ANNOTATION_REVIEW.wikiURL;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                null,
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.REVIEW_DATE.name;
        final String attribute1Description     = OpenMetadataProperty.REVIEW_DATE.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.REVIEW_DATE.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.STEWARD.name;
        final String attribute2Description     = OpenMetadataProperty.STEWARD.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.STEWARD.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.COMMENT.name;
        final String attribute3Description     = OpenMetadataProperty.COMMENT.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.COMMENT.descriptionGUID;

        property = archiveHelper.getDateTypeDefAttribute(attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getAnnotationReviewLinkRelationship()
    {
        final String guid            = OpenMetadataType.ANNOTATION_REVIEW_LINK_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.ANNOTATION_REVIEW_LINK_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.ANNOTATION_REVIEW_LINK_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.ANNOTATION_REVIEW_LINK_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.ANNOTATION_REVIEW_LINK_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.ANNOTATION.typeName;
        final String                     end1AttributeName            = "reviewedAnnotations";
        final String                     end1AttributeDescription     = "The annotations being reviewed.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.ANNOTATION_REVIEW.typeName;
        final String                     end2AttributeName            = "annotationReviews";
        final String                     end2AttributeDescription     = "The feedback about the annotations.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.ANNOTATION_STATUS.name;
        final String attribute1Description     = OpenMetadataProperty.ANNOTATION_STATUS.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.ANNOTATION_STATUS.descriptionGUID;

        property = archiveHelper.getEnumTypeDefAttribute("AnnotationStatus",
                                                         attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0615 Schema Extraction creates an annotation to describe a nested schema.
     */
    private void add0615SchemaExtraction()
    {
        this.archiveBuilder.addEntityDef(getSchemaAnalysisAnnotationEntity());
        this.archiveBuilder.addEntityDef(getDataFieldEntity());

        this.archiveBuilder.addRelationshipDef(getSchemaTypeDefinitionRelationship());
        this.archiveBuilder.addRelationshipDef(getDiscoveredDataFieldRelationship());
        this.archiveBuilder.addRelationshipDef(getSchemaAttributeDefinitionRelationship());
        this.archiveBuilder.addRelationshipDef(getDiscoveredNestedDataFieldRelationship());
        this.archiveBuilder.addRelationshipDef(getDataClassDefinitionRelationship());
    }

    private EntityDef getSchemaAnalysisAnnotationEntity()
    {
        final String guid            = OpenMetadataType.SCHEMA_ANALYSIS_ANNOTATION.typeGUID;
        final String name            = OpenMetadataType.SCHEMA_ANALYSIS_ANNOTATION.typeName;
        final String description     = OpenMetadataType.SCHEMA_ANALYSIS_ANNOTATION.description;
        final String descriptionGUID = OpenMetadataType.SCHEMA_ANALYSIS_ANNOTATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.SCHEMA_ANALYSIS_ANNOTATION.wikiURL;

        final String superTypeName = OpenMetadataType.ANNOTATION.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.SCHEMA_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.SCHEMA_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.SCHEMA_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.SCHEMA_TYPE.name;
        final String attribute2Description     = OpenMetadataProperty.SCHEMA_TYPE.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.SCHEMA_TYPE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getDataFieldEntity()
    {
        EntityDef entityDef = archiveHelper.getDefaultEntityDef(OpenMetadataType.DATA_FIELD.typeGUID,
                                                                OpenMetadataType.DATA_FIELD.typeName,
                                                                this.archiveBuilder.getEntityDef(OpenMetadataType.REFERENCEABLE.typeName),
                                                                OpenMetadataType.DATA_FIELD.description,
                                                                OpenMetadataType.DATA_FIELD.descriptionGUID,
                                                                OpenMetadataType.DATA_FIELD.wikiURL);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "dataFieldName";
        final String attribute1Description     = "Display name the data field.";
        final String attribute1DescriptionGUID = null;
        final String attribute2Name            = "dataFieldType";
        final String attribute2Description     = "Type name for the data field.";
        final String attribute2DescriptionGUID = null;
        final String attribute3Name            = "defaultValue";
        final String attribute3Description     = "Default value that is added to the field if no value is specified.";
        final String attribute3DescriptionGUID = null;
        final String attribute4Name            = OpenMetadataProperty.ADDITIONAL_PROPERTIES.name;
        final String attribute4Description     = "Additional properties discovered during the analysis.";
        final String attribute4DescriptionGUID = null;
        final String attribute5Name            = "anchorGUID";
        final String attribute5Description     = "Optional identifier of the asset that this data field is indirectly connected to.";
        final String attribute5DescriptionGUID = null;
        final String attribute6Name            = "dataFieldDescription";
        final String attribute6Description     = "Optional descriptive information about a data field.";
        final String attribute6DescriptionGUID = null;
        final String attribute7Name            = "dataFieldAliases";
        final String attribute7Description     = "Optional list of aliases for the data field.";
        final String attribute7DescriptionGUID = null;
        final String attribute8Name            = "dataFieldSortOrder";
        final String attribute8Description     = "Sort order for the values of the data field.";
        final String attribute8DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute4Name,
                                                                    attribute4Description,
                                                                    attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute7Name,
                                                                attribute7Description,
                                                                attribute7DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("DataItemSortOrder",
                                                         attribute8Name,
                                                         attribute8Description,
                                                         attribute8DescriptionGUID);
        properties.add(property);
        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getSchemaTypeDefinitionRelationship()
    {
        final String guid            = "60f2d263-e24d-4f20-8c0d-b5e24648cd54";
        final String name            = "SchemaTypeDefinition";
        final String description     = "Link between schema analysis annotation and the identified schema type definition.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = "SchemaAnalysisAnnotation";
        final String                     end1AttributeName            = "deployedSchemaTypes";
        final String                     end1AttributeDescription     = "The analysis of the schema type for deployed assets.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "SchemaType";
        final String                     end2AttributeName            = "schemaTypeDefinition";
        final String                     end2AttributeDescription     = "Official schema type definition.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getDiscoveredDataFieldRelationship()
    {
        final String guid            = OpenMetadataType.DISCOVERED_DATA_FIELD_TYPE_GUID;
        final String name            = OpenMetadataType.DISCOVERED_DATA_FIELD_TYPE_NAME;
        final String description     = "Data field detected in asset during schema analysis.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.SCHEMA_ANALYSIS_ANNOTATION.typeName;
        final String                     end1AttributeName            = "schemaAnalysisAnnotation";
        final String                     end1AttributeDescription     = "The annotation collecting the results of the schema analysis.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.DATA_FIELD.typeName;
        final String                     end2AttributeName            = "discoveredDataFields";
        final String                     end2AttributeDescription     = "The data fields discovered during schema analysis.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "dataFieldPosition";
        final String attribute1Description     = "Location of the data field in the parent annotation's list of data fields.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getIntTypeDefAttribute(attribute1Name,
                                                        attribute1Description,
                                                        attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getSchemaAttributeDefinitionRelationship()
    {
        final String guid            = "60f1e263-e24d-4f20-8c0d-b5e21232cd54";
        final String name            = "SchemaAttributeDefinition";
        final String description     = "Link between data field analysis and the identified schema attribute definition.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.DATA_FILE.typeName;
        final String                     end1AttributeName            = "deployedSchemaAttributes";
        final String                     end1AttributeDescription     = "The analysis of the equivalent data fields from deployed assets.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.SCHEMA_ATTRIBUTE.typeName;
        final String                     end2AttributeName            = "schemaAttributeDefinition";
        final String                     end2AttributeDescription     = "Official schema attribute definition.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    private RelationshipDef getDiscoveredNestedDataFieldRelationship()
    {
        final String guid            = OpenMetadataType.NESTED_DATA_FIELD_TYPE_GUID;
        final String name            = OpenMetadataType.NESTED_DATA_FIELD_TYPE_NAME;
        final String description     = "Nested data fields under a single parent node.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.DATA_FIELD.typeName;
        final String                     end1AttributeName            = "parentDataField";
        final String                     end1AttributeDescription     = "Parent node.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.DATA_FIELD.typeName;
        final String                     end2AttributeName            = "nestedDataFields";
        final String                     end2AttributeDescription     = "Nested data fields.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "dataFieldPosition";
        final String attribute1Description     = "Positional order of the data field with its parent data field.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getIntTypeDefAttribute(attribute1Name,
                                                        attribute1Description,
                                                        attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    private RelationshipDef getDataClassDefinitionRelationship()
    {
        final String guid = "51a2d263-e24d-4f20-8c0d-b5e12356cd54";

        final String name            = "DataClassDefinition";
        final String description     = "Link between schema analysis and the identified data class for a data field.";
        final String descriptionGUID = null;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.DATA_FIELD.typeName;
        final String                     end1AttributeName            = "deployedDataClasses";
        final String                     end1AttributeDescription     = "The mapped data fields from deployed assets.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = "DataClass";
        final String                     end2AttributeName            = "dataClassDefinition";
        final String                     end2AttributeDescription     = "Official data class definition.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = "assetGUID";
        final String attribute1Description     = "Unique identifier for the analyzed asset.";
        final String attribute1DescriptionGUID = null;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0617 Data Field Analysis defines the structures to record the individual data fields found in an asset.
     */
    private void add0617DataFieldAnalysis()
    {
        this.archiveBuilder.addEntityDef(getDataFieldAnnotationEntity());

        this.archiveBuilder.addRelationshipDef(getDataFieldAnalysisRelationship());
    }


    private EntityDef getDataFieldAnnotationEntity()
    {
        final String guid            = OpenMetadataType.DATA_FIELD_ANNOTATION.typeGUID;
        final String name            = OpenMetadataType.DATA_FIELD_ANNOTATION.typeName;
        final String description     = OpenMetadataType.DATA_FIELD_ANNOTATION.description;
        final String descriptionGUID = OpenMetadataType.DATA_FIELD_ANNOTATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.DATA_FIELD_ANNOTATION.wikiURL;

        final String superTypeName = OpenMetadataType.ANNOTATION.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    private RelationshipDef getDataFieldAnalysisRelationship()
    {
        final String guid            = OpenMetadataType.DATA_FIELD_ANALYSIS_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.DATA_FIELD_ANALYSIS_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.DATA_FIELD_ANALYSIS_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.DATA_FIELD_ANALYSIS_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.DATA_FIELD_ANALYSIS_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.DATA_FIELD_ANNOTATION.typeName;
        final String                     end1AttributeName            = "dataFieldAnnotations";
        final String                     end1AttributeDescription     = "The annotations for this data field.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.DATA_FIELD.typeName;
        final String                     end2AttributeName            = "annotatedDataFields";
        final String                     end2AttributeDescription     = "Data fields with addition properties attached.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0620 Profiling describes an annotation that can be attached to a column to describe the profile of its
     * values.
     */
    private void add0620DataProfiling()
    {
        this.archiveBuilder.addEntityDef(getDataProfileAnnotationEntity());
        this.archiveBuilder.addEntityDef(getResourceProfileLogAnnotationEntity());

        this.archiveBuilder.addRelationshipDef(getResourceProfileLogFileRelationship());

    }


    private EntityDef getDataProfileAnnotationEntity()
    {
        final String guid            = OpenMetadataType.RESOURCE_PROFILE_ANNOTATION.typeGUID;
        final String name            = OpenMetadataType.RESOURCE_PROFILE_ANNOTATION.typeName;
        final String description     = OpenMetadataType.RESOURCE_PROFILE_ANNOTATION.description;
        final String descriptionGUID = OpenMetadataType.RESOURCE_PROFILE_ANNOTATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.RESOURCE_PROFILE_ANNOTATION.wikiURL;

        final String superTypeName = OpenMetadataType.DATA_FIELD_ANNOTATION.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);
        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.LENGTH.name;
        final String attribute1Description     = OpenMetadataProperty.LENGTH.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.LENGTH.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.INFERRED_DATA_TYPE.name;
        final String attribute2Description     = OpenMetadataProperty.INFERRED_DATA_TYPE.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.INFERRED_DATA_TYPE.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.INFERRED_FORMAT.name;
        final String attribute3Description     = OpenMetadataProperty.INFERRED_FORMAT.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.INFERRED_FORMAT.descriptionGUID;
        final String attribute4Name            = OpenMetadataProperty.INFERRED_LENGTH.name;
        final String attribute4Description     = OpenMetadataProperty.INFERRED_LENGTH.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.INFERRED_LENGTH.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.INFERRED_PRECISION.name;
        final String attribute5Description     = OpenMetadataProperty.INFERRED_PRECISION.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.INFERRED_PRECISION.descriptionGUID;
        final String attribute6Name            = OpenMetadataProperty.INFERRED_SCALE.name;
        final String attribute6Description     = OpenMetadataProperty.INFERRED_SCALE.description;
        final String attribute6DescriptionGUID = OpenMetadataProperty.INFERRED_SCALE.descriptionGUID;
        final String attribute7Name            = OpenMetadataProperty.PROFILE_PROPERTIES.name;
        final String attribute7Description     = OpenMetadataProperty.PROFILE_PROPERTIES.description;
        final String attribute7DescriptionGUID = OpenMetadataProperty.PROFILE_PROPERTIES.descriptionGUID;
        final String attribute8Name            = OpenMetadataProperty.PROFILE_FLAGS.name;
        final String attribute8Description     = OpenMetadataProperty.PROFILE_FLAGS.description;
        final String attribute8DescriptionGUID = OpenMetadataProperty.PROFILE_FLAGS.descriptionGUID;
        final String attribute9Name            = OpenMetadataProperty.PROFILE_COUNTS.name;
        final String attribute9Description     = OpenMetadataProperty.PROFILE_COUNTS.description;
        final String attribute9DescriptionGUID = OpenMetadataProperty.PROFILE_COUNTS.descriptionGUID;
        final String attribute10Name            = OpenMetadataProperty.VALUE_LIST.name;
        final String attribute10Description     = OpenMetadataProperty.VALUE_LIST.description;
        final String attribute10DescriptionGUID = OpenMetadataProperty.VALUE_LIST.descriptionGUID;
        final String attribute11Name            = OpenMetadataProperty.VALUE_COUNT.name;
        final String attribute11Description     = OpenMetadataProperty.VALUE_COUNT.description;
        final String attribute11DescriptionGUID = OpenMetadataProperty.VALUE_COUNT.descriptionGUID;
        final String attribute12Name            = OpenMetadataProperty.VALUE_RANGE_FROM.name;
        final String attribute12Description     = OpenMetadataProperty.VALUE_RANGE_FROM.description;
        final String attribute12DescriptionGUID = OpenMetadataProperty.VALUE_RANGE_FROM.descriptionGUID;
        final String attribute13Name            = OpenMetadataProperty.VALUE_RANGE_TO.name;
        final String attribute13Description     = OpenMetadataProperty.VALUE_RANGE_TO.description;
        final String attribute13DescriptionGUID = OpenMetadataProperty.VALUE_RANGE_TO.descriptionGUID;
        final String attribute14Name            = OpenMetadataProperty.AVERAGE_VALUE.name;
        final String attribute14Description     = OpenMetadataProperty.AVERAGE_VALUE.description;
        final String attribute14DescriptionGUID = OpenMetadataProperty.AVERAGE_VALUE.descriptionGUID;

        property = archiveHelper.getIntTypeDefAttribute(attribute1Name,
                                                        attribute1Description,
                                                        attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute4Name,
                                                        attribute4Description,
                                                        attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute5Name,
                                                        attribute5Description,
                                                        attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute6Name,
                                                        attribute6Description,
                                                        attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute7Name,
                                                                    attribute7Description,
                                                                    attribute7DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringBooleanTypeDefAttribute(attribute8Name,
                                                                     attribute8Description,
                                                                     attribute8DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringLongTypeDefAttribute(attribute9Name,
                                                                  attribute9Description,
                                                                  attribute9DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute10Name,
                                                                attribute10Description,
                                                                attribute10DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringIntTypeDefAttribute(attribute11Name,
                                                                 attribute11Description,
                                                                 attribute11DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute12Name,
                                                           attribute12Description,
                                                           attribute12DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute13Name,
                                                           attribute13Description,
                                                           attribute13DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute14Name,
                                                           attribute14Description,
                                                           attribute14DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getResourceProfileLogAnnotationEntity()
    {
        final String guid            = OpenMetadataType.RESOURCE_PROFILE_LOG_ANNOTATION.typeGUID;
        final String name            = OpenMetadataType.RESOURCE_PROFILE_LOG_ANNOTATION.typeName;
        final String description     = OpenMetadataType.RESOURCE_PROFILE_LOG_ANNOTATION.description;
        final String descriptionGUID = OpenMetadataType.RESOURCE_PROFILE_LOG_ANNOTATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.RESOURCE_PROFILE_LOG_ANNOTATION.wikiURL;

        final String superTypeName = OpenMetadataType.DATA_FIELD_ANNOTATION.typeName;

        return archiveHelper.getDefaultEntityDef(guid,
                                                 name,
                                                 this.archiveBuilder.getEntityDef(superTypeName),
                                                 description,
                                                 descriptionGUID,
                                                 descriptionWiki);
    }


    private RelationshipDef getResourceProfileLogFileRelationship()
    {
        final String guid            = OpenMetadataType.RESOURCE_PROFILE_DATA_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.RESOURCE_PROFILE_DATA_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.RESOURCE_PROFILE_DATA_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.RESOURCE_PROFILE_DATA_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.RESOURCE_PROFILE_DATA_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.RESOURCE_PROFILE_LOG_ANNOTATION.typeName;
        final String                     end1AttributeName            = "resourceProfileAnnotations";
        final String                     end1AttributeDescription     = "The annotations that refer to this log file.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.ASSET.typeName;
        final String                     end2AttributeName            = "resourceProfileLogs";
        final String                     end2AttributeDescription     = "Location of the profile information.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0625 Data Class Discovery records potential data classes that match the data field, and the level of
     * error in the match.
     */
    private void add0625DataClassDiscovery()
    {
        this.archiveBuilder.addEntityDef(getDataClassAnnotationEntity());
    }


    private EntityDef getDataClassAnnotationEntity()
    {
        final String guid            = OpenMetadataType.DATA_CLASS_ANNOTATION.typeGUID;
        final String name            = OpenMetadataType.DATA_CLASS_ANNOTATION.typeName;
        final String description     = OpenMetadataType.DATA_CLASS_ANNOTATION.description;
        final String descriptionGUID = OpenMetadataType.DATA_CLASS_ANNOTATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.DATA_CLASS_ANNOTATION.wikiURL;

        final String superTypeName = OpenMetadataType.DATA_FIELD_ANNOTATION.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.CANDIDATE_DATA_CLASS_GUIDS.name;
        final String attribute1Description     = OpenMetadataProperty.CANDIDATE_DATA_CLASS_GUIDS.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.CANDIDATE_DATA_CLASS_GUIDS.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.MATCHING_VALUES.name;
        final String attribute2Description     = OpenMetadataProperty.MATCHING_VALUES.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.MATCHING_VALUES.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.NON_MATCHING_VALUES.name;
        final String attribute3Description     = OpenMetadataProperty.NON_MATCHING_VALUES.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.NON_MATCHING_VALUES.descriptionGUID;

        property = archiveHelper.getArrayStringTypeDefAttribute(attribute1Name,
                                                                attribute1Description,
                                                                attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getLongTypeDefAttribute(attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getLongTypeDefAttribute(attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0630 Semantic Discovery describes an annotation for a candidate glossary term assignment for a data element.
     */
    private void add0630SemanticDiscovery()
    {
        this.archiveBuilder.addEntityDef(getSemanticAnnotationEntity());
    }

    private EntityDef getSemanticAnnotationEntity()
    {
        final String guid            = OpenMetadataType.SEMANTIC_ANNOTATION.typeGUID;
        final String name            = OpenMetadataType.SEMANTIC_ANNOTATION.typeName;
        final String description     = OpenMetadataType.SEMANTIC_ANNOTATION.description;
        final String descriptionGUID = OpenMetadataType.SEMANTIC_ANNOTATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.SEMANTIC_ANNOTATION.wikiURL;

        final String superTypeName = OpenMetadataType.DATA_FIELD_ANNOTATION.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.INFORMAL_TERM.name;
        final String attribute1Description     = OpenMetadataProperty.INFORMAL_TERM.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.INFORMAL_TERM.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.CANDIDATE_GLOSSARY_TERM_GUIDS.name;
        final String attribute2Description     = OpenMetadataProperty.CANDIDATE_GLOSSARY_TERM_GUIDS.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.CANDIDATE_GLOSSARY_TERM_GUIDS.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.INFORMAL_CATEGORY.name;
        final String attribute3Description     = OpenMetadataProperty.INFORMAL_CATEGORY.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.INFORMAL_CATEGORY.descriptionGUID;
        final String attribute4Name            = OpenMetadataProperty.CANDIDATE_GLOSSARY_CATEGORY_GUIDS.name;
        final String attribute4Description     = OpenMetadataProperty.CANDIDATE_GLOSSARY_CATEGORY_GUIDS.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.CANDIDATE_GLOSSARY_CATEGORY_GUIDS.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute2Name,
                                                                attribute2Description,
                                                                attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getArrayStringTypeDefAttribute(attribute4Name,
                                                                attribute4Description,
                                                                attribute4DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0635 Classification Discovery creates a base class for a classification annotation.
     */
    private void add0635ClassificationDiscovery()
    {
        this.archiveBuilder.addEntityDef(getClassificationAnnotationEntity());
    }


    private EntityDef getClassificationAnnotationEntity()
    {
        final String guid            = OpenMetadataType.CLASSIFICATION_ANNOTATION.typeGUID;
        final String name            = OpenMetadataType.CLASSIFICATION_ANNOTATION.typeName;
        final String description     = OpenMetadataType.CLASSIFICATION_ANNOTATION.description;
        final String descriptionGUID = OpenMetadataType.CLASSIFICATION_ANNOTATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.CLASSIFICATION_ANNOTATION.wikiURL;

        final String superTypeName = OpenMetadataType.DATA_FIELD_ANNOTATION.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.CANDIDATE_CLASSIFICATIONS.name;
        final String attribute1Description     = OpenMetadataProperty.CANDIDATE_CLASSIFICATIONS.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.CANDIDATE_CLASSIFICATIONS.descriptionGUID;

        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute1Name,
                                                                    attribute1Description,
                                                                    attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */

    /**
     * 0640 Quality Scores records quality analysis results.  Quality has many different dimensions and
     * the dimension that this score reflects is also recorded with the score.
     */
    private void add0640QualityScores()
    {
        this.archiveBuilder.addEntityDef(getQualityAnnotationEntity());
    }


    private EntityDef getQualityAnnotationEntity()
    {
        final String guid            = OpenMetadataType.QUALITY_ANNOTATION.typeGUID;
        final String name            = OpenMetadataType.QUALITY_ANNOTATION.typeName;
        final String description     = OpenMetadataType.QUALITY_ANNOTATION.description;
        final String descriptionGUID = OpenMetadataType.QUALITY_ANNOTATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.QUALITY_ANNOTATION.wikiURL;

        final String superTypeName = OpenMetadataType.DATA_FIELD_ANNOTATION.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.QUALITY_DIMENSION.name;
        final String attribute1Description     = OpenMetadataProperty.QUALITY_DIMENSION.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.QUALITY_DIMENSION.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.QUALITY_SCORE.name;
        final String attribute2Description     = OpenMetadataProperty.QUALITY_SCORE.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.QUALITY_SCORE.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute2Name,
                                                        attribute2Description,
                                                        attribute2DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0650 Relationship Discovery provides a base annotation for describing a relationship between two referenceables.
     */
    private void add0650RelationshipDiscovery()
    {
        this.archiveBuilder.addEntityDef(getRelationshipAdviceAnnotationEntity());

        this.archiveBuilder.addRelationshipDef(getRelationshipAnnotationRelationship());
    }


    private EntityDef getRelationshipAdviceAnnotationEntity()
    {
        final String guid            = OpenMetadataType.RELATIONSHIP_ADVICE_ANNOTATION.typeGUID;
        final String name            = OpenMetadataType.RELATIONSHIP_ADVICE_ANNOTATION.typeName;
        final String description     = OpenMetadataType.RELATIONSHIP_ADVICE_ANNOTATION.description;
        final String descriptionGUID = OpenMetadataType.RELATIONSHIP_ADVICE_ANNOTATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.RELATIONSHIP_ADVICE_ANNOTATION.wikiURL;

        final String superTypeName = OpenMetadataType.DATA_FIELD_ANNOTATION.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);
        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.RELATIONSHIP_TYPE_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.RELATIONSHIP_TYPE_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.RELATIONSHIP_TYPE_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.RELATIONSHIP_PROPERTIES.name;
        final String attribute2Description     = OpenMetadataProperty.RELATIONSHIP_PROPERTIES.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.RELATIONSHIP_PROPERTIES.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.RELATED_ENTITY_GUID.name;
        final String attribute3Description     = OpenMetadataProperty.RELATED_ENTITY_GUID.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.RELATED_ENTITY_GUID.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute2Name,
                                                                    attribute2Description,
                                                                    attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute3Name,
                                                           attribute3Description,
                                                           attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private RelationshipDef getRelationshipAnnotationRelationship()
    {
        final String guid            = OpenMetadataType.RELATIONSHIP_ANNOTATION_RELATIONSHIP.typeGUID;
        final String name            = OpenMetadataType.RELATIONSHIP_ANNOTATION_RELATIONSHIP.typeName;
        final String description     = OpenMetadataType.RELATIONSHIP_ANNOTATION_RELATIONSHIP.description;
        final String descriptionGUID = OpenMetadataType.RELATIONSHIP_ANNOTATION_RELATIONSHIP.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.RELATIONSHIP_ANNOTATION_RELATIONSHIP.wikiURL;

        final ClassificationPropagationRule classificationPropagationRule = ClassificationPropagationRule.NONE;

        RelationshipDef relationshipDef = archiveHelper.getBasicRelationshipDef(guid,
                                                                                name,
                                                                                null,
                                                                                description,
                                                                                descriptionGUID,
                                                                                descriptionWiki,
                                                                                classificationPropagationRule);

        RelationshipEndDef relationshipEndDef;

        /*
         * Set up end 1.
         */
        final String                     end1EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end1AttributeName            = "relatedFromObjectAnnotations";
        final String                     end1AttributeDescription     = "The referenceables linked from.";
        final String                     end1AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end1Cardinality              = RelationshipEndCardinality.ANY_NUMBER;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end1EntityType),
                                                                 end1AttributeName,
                                                                 end1AttributeDescription,
                                                                 end1AttributeDescriptionGUID,
                                                                 end1Cardinality);
        relationshipDef.setEndDef1(relationshipEndDef);


        /*
         * Set up end 2.
         */
        final String                     end2EntityType               = OpenMetadataType.REFERENCEABLE.typeName;
        final String                     end2AttributeName            = "relatedToObjectAnnotations";
        final String                     end2AttributeDescription     = "The referenceables linked to.";
        final String                     end2AttributeDescriptionGUID = null;
        final RelationshipEndCardinality end2Cardinality              = RelationshipEndCardinality.AT_MOST_ONE;

        relationshipEndDef = archiveHelper.getRelationshipEndDef(this.archiveBuilder.getEntityDef(end2EntityType),
                                                                 end2AttributeName,
                                                                 end2AttributeDescription,
                                                                 end2AttributeDescriptionGUID,
                                                                 end2Cardinality);
        relationshipDef.setEndDef2(relationshipEndDef);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.ANNOTATION_TYPE.name;
        final String attribute1Description     = OpenMetadataProperty.ANNOTATION_TYPE.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.ANNOTATION_TYPE.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.SUMMARY.name;
        final String attribute2Description     = OpenMetadataProperty.SUMMARY.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.SUMMARY.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.CONFIDENCE_LEVEL.name;
        final String attribute3Description     = OpenMetadataProperty.CONFIDENCE_LEVEL.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.CONFIDENCE_LEVEL.descriptionGUID;
        final String attribute4Name            = OpenMetadataProperty.EXPRESSION.name;
        final String attribute4Description     = OpenMetadataProperty.EXPRESSION.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.EXPRESSION.descriptionGUID;
        final String attribute5Name            = OpenMetadataProperty.EXPLANATION.name;
        final String attribute5Description     = OpenMetadataProperty.EXPLANATION.description;
        final String attribute5DescriptionGUID = OpenMetadataProperty.EXPLANATION.descriptionGUID;
        final String attribute6Name            = OpenMetadataProperty.ANALYSIS_STEP.name;
        final String attribute6Description     = OpenMetadataProperty.ANALYSIS_STEP.description;
        final String attribute6DescriptionGUID = OpenMetadataProperty.ANALYSIS_STEP.descriptionGUID;
        final String attribute7Name            = OpenMetadataProperty.JSON_PROPERTIES.name;
        final String attribute7Description     = OpenMetadataProperty.JSON_PROPERTIES.description;
        final String attribute7DescriptionGUID = OpenMetadataProperty.JSON_PROPERTIES.descriptionGUID;
        final String attribute8Name            = OpenMetadataProperty.ADDITIONAL_PROPERTIES.name;
        final String attribute8Description     = OpenMetadataProperty.ADDITIONAL_PROPERTIES.description;
        final String attribute8DescriptionGUID = OpenMetadataProperty.ADDITIONAL_PROPERTIES.descriptionGUID;
        final String attribute9Name            = OpenMetadataProperty.ANNOTATION_STATUS.name;
        final String attribute9Description     = OpenMetadataProperty.ANNOTATION_STATUS.description;
        final String attribute9DescriptionGUID = OpenMetadataProperty.ANNOTATION_STATUS.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getIntTypeDefAttribute(attribute3Name,
                                                        attribute3Description,
                                                        attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute5Name,
                                                           attribute5Description,
                                                           attribute5DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute6Name,
                                                           attribute6Description,
                                                           attribute6DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute7Name,
                                                           attribute7Description,
                                                           attribute7DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute8Name,
                                                                    attribute8Description,
                                                                    attribute8DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getEnumTypeDefAttribute("AnnotationStatus",
                                                         attribute9Name,
                                                         attribute9Description,
                                                         attribute9DescriptionGUID);
        properties.add(property);

        relationshipDef.setPropertiesDefinition(properties);

        return relationshipDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0660 Data Source Measurements describe annotations that are measuring characteristics of a data source.
     * This may be a DataSet, DataFeed or DataSource.
     */
    private void add0660ResourceMeasures()
    {
        this.archiveBuilder.addEntityDef(getResourceMeasureAnnotationEntity());
        this.archiveBuilder.addEntityDef(getDataSourcePhysicalStatusAnnotationEntity());
    }


    private EntityDef getResourceMeasureAnnotationEntity()
    {
        final String guid = OpenMetadataType.RESOURCE_MEASURE_ANNOTATION.typeGUID;

        final String name            = OpenMetadataType.RESOURCE_MEASURE_ANNOTATION.typeName;
        final String description     = OpenMetadataType.RESOURCE_MEASURE_ANNOTATION.description;
        final String descriptionGUID = OpenMetadataType.RESOURCE_MEASURE_ANNOTATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.RESOURCE_MEASURE_ANNOTATION.wikiURL;

        final String superTypeName = OpenMetadataType.ANNOTATION.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.RESOURCE_PROPERTIES.name;
        final String attribute1Description     = OpenMetadataProperty.RESOURCE_PROPERTIES.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.RESOURCE_PROPERTIES.descriptionGUID;

        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute1Name,
                                                                    attribute1Description,
                                                                    attribute1DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    private EntityDef getDataSourcePhysicalStatusAnnotationEntity()
    {
        final String guid = OpenMetadataType.RESOURCE_PHYSICAL_STATUS_ANNOTATION.typeGUID;

        final String name            = OpenMetadataType.RESOURCE_PHYSICAL_STATUS_ANNOTATION.typeName;
        final String description     = OpenMetadataType.RESOURCE_PHYSICAL_STATUS_ANNOTATION.description;
        final String descriptionGUID = OpenMetadataType.RESOURCE_PHYSICAL_STATUS_ANNOTATION.descriptionGUID;
        final String descriptionWiki = OpenMetadataType.RESOURCE_PHYSICAL_STATUS_ANNOTATION.wikiURL;

        final String superTypeName = OpenMetadataType.RESOURCE_MEASURE_ANNOTATION.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                descriptionWiki);
        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.RESOURCE_CREATE_TIME.name;
        final String attribute1Description     = OpenMetadataProperty.RESOURCE_CREATE_TIME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.RESOURCE_CREATE_TIME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.RESOURCE_UPDATE_TIME.name;
        final String attribute2Description     = OpenMetadataProperty.RESOURCE_UPDATE_TIME.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.RESOURCE_UPDATE_TIME.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.SIZE.name;
        final String attribute3Description     = OpenMetadataProperty.SIZE.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.SIZE.descriptionGUID;
        final String attribute4Name            = OpenMetadataProperty.ENCODING.name;
        final String attribute4Description     = OpenMetadataProperty.ENCODING.description;
        final String attribute4DescriptionGUID = OpenMetadataProperty.ENCODING.descriptionGUID;

        property = archiveHelper.getDateTypeDefAttribute(attribute1Name,
                                                         attribute1Description,
                                                         attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getDateTypeDefAttribute(attribute2Name,
                                                         attribute2Description,
                                                         attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getLongTypeDefAttribute(attribute3Name,
                                                         attribute3Description,
                                                         attribute3DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute4Name,
                                                           attribute4Description,
                                                           attribute4DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }


    /*
     * -------------------------------------------------------------------------------------------------------
     */


    /**
     * 0690 Request for Action creates an annotation for requesting a stewardship or governance action.
     */
    private void add0690RequestForAction()
    {
        this.archiveBuilder.addEntityDef(getRequestForActionAnnotationEntity());
    }


    private EntityDef getRequestForActionAnnotationEntity()
    {
        final String guid = OpenMetadataType.REQUEST_FOR_ACTION_ANNOTATION.typeGUID;

        final String name            = OpenMetadataType.REQUEST_FOR_ACTION_ANNOTATION.typeName;
        final String description     = OpenMetadataType.REQUEST_FOR_ACTION_ANNOTATION.description;
        final String descriptionGUID = OpenMetadataType.REQUEST_FOR_ACTION_ANNOTATION.descriptionGUID;
        final String wikiURL         = OpenMetadataType.REQUEST_FOR_ACTION_ANNOTATION.wikiURL;

        final String superTypeName = OpenMetadataType.DATA_FIELD_ANNOTATION.typeName;

        EntityDef entityDef = archiveHelper.getDefaultEntityDef(guid,
                                                                name,
                                                                this.archiveBuilder.getEntityDef(superTypeName),
                                                                description,
                                                                descriptionGUID,
                                                                wikiURL);

        /*
         * Build the attributes
         */
        List<TypeDefAttribute> properties = new ArrayList<>();
        TypeDefAttribute       property;

        final String attribute1Name            = OpenMetadataProperty.ACTION_SOURCE_NAME.name;
        final String attribute1Description     = OpenMetadataProperty.ACTION_SOURCE_NAME.description;
        final String attribute1DescriptionGUID = OpenMetadataProperty.ACTION_SOURCE_NAME.descriptionGUID;
        final String attribute2Name            = OpenMetadataProperty.ACTION_REQUESTED.name;
        final String attribute2Description     = OpenMetadataProperty.ACTION_REQUESTED.description;
        final String attribute2DescriptionGUID = OpenMetadataProperty.ACTION_REQUESTED.descriptionGUID;
        final String attribute3Name            = OpenMetadataProperty.ACTION_PROPERTIES.name;
        final String attribute3Description     = OpenMetadataProperty.ACTION_PROPERTIES.description;
        final String attribute3DescriptionGUID = OpenMetadataProperty.ACTION_PROPERTIES.descriptionGUID;

        property = archiveHelper.getStringTypeDefAttribute(attribute1Name,
                                                           attribute1Description,
                                                           attribute1DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getStringTypeDefAttribute(attribute2Name,
                                                           attribute2Description,
                                                           attribute2DescriptionGUID);
        properties.add(property);
        property = archiveHelper.getMapStringStringTypeDefAttribute(attribute3Name,
                                                                    attribute3Description,
                                                                    attribute3DescriptionGUID);
        properties.add(property);

        entityDef.setPropertiesDefinition(properties);

        return entityDef;
    }
}
